#include "pcheader.h"

// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

#include "server.h"

AccountList al;

// eventually, provide a mechanism for accounts to be added at run-time
void GetAccounts(char *accountfile)
{
	FILE *ifile;
	char buf[1000], account[30], pw[30];
	int count, j, pcount, acount;
	bool found;

	if((ifile = fopen(accountfile, "rb")) == NULL) {
		//printf("Account file not found.\n");
		return;
	}

	while(fgets(buf, 1000, ifile) != NULL) {
		count = strlen(buf);
		found = false;
		acount = pcount = 0;
		for(j = 0; j < count; j++) {
			if(buf[j] == '|') {
				if(!found) {
					found = true;
				}
				else {
					//printf("Error, bad account\n");
				}
			}
			// carriage return actually appears first w/ MS-DOS
			else if(buf[j] == 0x0d || buf[j] == 0x0a/*'\n'*/) {
				if(!found) {
					//printf("Error, bad account\n");
				}
				else {
					break;
				}
			}
			else {
				if(found) {
					if(pcount < 30) {
						pw[pcount] = buf[j];
						pcount += 1;
					}
				}
				else {
					if(acount < 30) {
						account[acount] = buf[j];
						acount += 1;
					}
				}
			}
		}

		// fill in zeros
		for(j = pcount; j < 30; j++)
			pw[j] = 0;

		// in case string too long
		pw[29] = 0;

		// fill in zeros
		for(j = acount; j < 30; j++)
			account[j] = 0;

		// in case string too long
		account[29] = 0;

		al.Register(account, pw);
	}

	fclose(ifile);
}

int ValidateAccount(char *acct, char *pass)
{
	return al.Validate(acct, pass);
}

const PCList *GetPlayers(int id)
{
	return al.GetPlayers(id);
}

bool AddPlayer(int id, int npcid)
{
	return al.AddPlayer(id, npcid);
}

bool DeletePlayer(int id, int npcid)
{
	return al.DeletePlayer(id, npcid);
}

bool SaveAccounts(FILE *sfil)
{
	int i, j, temp, count;
	unsigned short ver;
	char c;

	ver = htons(0x0100);

	if(fwrite(&ver, 2, 1, sfil) != 1) {
		return false;
	}

#ifdef WORLD_BUILD
	count = 0;
#else
	count = al.count;
#endif

	temp = htonl(count);

	if(fwrite(&temp, 4, 1, sfil) != 1) {
		return false;
	}

	for(i = 0; i < count; i++) {
		if(fwrite(al.list[i] -> name, STRSIZE, 1, sfil) != 1) {
			return false;
		}

		if(fwrite(al.list[i] -> pass, STRSIZE, 1, sfil) != 1) {
			return false;
		}

		if(!al.list[i] -> blocked) {
			c = 0;
		}
		else {
			c = 1;
		}

		if(fwrite(&c, 1, 1, sfil) != 1) {
			return false;
		}

		temp = htonl(al.list[i] -> pcs.count);

		if(fwrite(&temp, 4, 1, sfil) != 1) {
			return false;
		}

		for(j = 0; j < al.list[i] -> pcs.count; j++) {
			temp = htonl(al.list[i] -> pcs.list[j]);

			if(fwrite(&temp, 4, 1, sfil) != 1) {
				return false;
			}
		}
	}

	return true;
}

bool LoadAccounts(FILE *sfil)
{
	int i, j, temp, count;
	char c, name[STRSIZE], pass[STRSIZE];
	unsigned short ver;

	if(fread(&ver, 2, 1, sfil) != 1) {
		return false;
	}

	ver = ntohs(ver);

	if(ver == 0x0100) {
		if(fread(&temp, 4, 1, sfil) != 1) {
			return false;
		}

		count = ntohl(temp);

		for(i = 0; i < count; i++) {
			if(fread(name, STRSIZE, 1, sfil) != 1) {
				return false;
			}

			if(fread(pass, STRSIZE, 1, sfil) != 1) {
				return false;
			}

			// we shouldn't use register. instead, we should
			// write the name and pass directly to a new struct
			al.Register(name, pass);

			if(fread(&c, 1, 1, sfil) != 1) {
				return false;
			}

			if(c == 0) {
				al.list[i] -> blocked = false;
			}
			else {
				al.list[i] -> blocked = true;
			}

			if(fread(&temp, 4, 1, sfil) != 1) {
				return false;
			}

			al.list[i] -> pcs.count = ntohl(temp);

			for(j = 0; j < al.list[i] -> pcs.count; j++) {
				if(fread(&temp, 4, 1, sfil) != 1) {
					return false;
				}

				al.list[i] -> pcs.list[j] = ntohl(temp);
			}
		}
	}
	else {
		return false;
	}

	return true;
}
