// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

#define PCLIST_SIZE 5

struct PCList {
	int count, list[PCLIST_SIZE];

	PCList(void)
	{
		count = 0;
	}
} ;

struct Account10 {
	char name[30], pass[30];
	PCList pcs;
	bool blocked;

	Account10(void)
	{
		name[0] = pass[0] = 0;
		blocked = false;
	}
} ;

struct AccountList {
	int count;
	Account10 **list;

	AccountList(void)
	{
		count = 0;
		list = NULL;
	}

	~AccountList(void)
	{
		Cleanup();
	}

	void Cleanup(void)
	{
		int i;

		if(list != NULL) {
			for(i = 0; i < count; i++) {
				delete list[i];
			}

			delete []list;
		}

		list = NULL;
		count = 0;
	}

	void Register(char *name, char *pword)
	{
		int i;
		Account10 **list2;
		char *pass = pword;

		for(i = 0; i < count; i++) {
			if(strcmp(list[i] -> name, name) == 0) {
				if(*pass == '*') {
					list[i] -> blocked =  true;
					pass++;
				}
				else {
					list[i] -> blocked =  false;
				}

				if(strcmp(list[i] -> pass, pass) != 0) {
					strcpy(list[i] -> pass, pass);
				}

				break;
			}
		}

		// add the account
		if(i == count) {
			list2 = new Account10 *[count + 1];

			for(i = 0; i < count; i++) {
				list2[i] = list[i];
			}

			list2[i] = new Account10;

			strcpy(list2[i] -> name, name);

			if(*pass == '*') {
				list2[i] -> blocked =  true;
				pass++;
			}
			else {
				list2[i] -> blocked = false;
			}

			strcpy(list2[i] -> pass, pass);
			list2[i] -> pcs.count = 0;

			delete []list;
			list = list2;
			count += 1;
		}
	}

	int Validate(char *name, char *pass)
	{
		int i;

		for(i = 0; i < count; i++) {
			if(strcmp(list[i] -> name, name) == 0) {
				break;
			}
		}

		if(i < count) {
			if(strcmp(list[i] -> pass, pass) != 0)
				return -2;
			else if(list[i] -> blocked)
				return -3;

			return i;
		}
		else {
			return -1;
		}
	}

	const PCList *GetPlayers(int id)
	{
		if(id < count) {
			return &(list[id] -> pcs);
		}
		else {
			return NULL;
		}
	}

	bool AddPlayer(int id, int npcid)
	{
		bool val = false;
		int index;

		if(id < count) {
			if(list[id] -> pcs.count < PCLIST_SIZE) {
				index = list[id] -> pcs.count;
				list[id] -> pcs.count += 1;
				list[id] -> pcs.list[index] = npcid;
				val = true;
			}
		}

		return val;
	}

	bool DeletePlayer(int id, int npcid)
	{
		bool val = false;
		int i;

		if(id < count) {
			for(i = 0; i < list[id] -> pcs.count; i++) {
				if(list[id] -> pcs.list[i] == npcid)
					break;
			}

			if(i < list[id] -> pcs.count) {
				list[id] -> pcs.count -= 1;
				list[id] -> pcs.list[i] = list[id] -> pcs.list[list[id] -> pcs.count];
				val = true;
			}
		}

		return val;
	}
} ;

void GetAccounts(char *);
int ValidateAccount(char *, char *);
const PCList *GetPlayers(int);
bool AddPlayer(int, int);
bool DeletePlayer(int, int);
bool SaveAccounts(FILE *);
bool LoadAccounts(FILE *);
