#include "pcheader.h"

// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

#include "server.h"

connectlist connections;

void initwrite(SOCKET s, int count, int offset);

int addconnection(SOCKET s, int acct, unsigned npcid, int ecpos, const char *encrypt, const char *data)
{
	return connections.add(s, acct, npcid, ecpos, encrypt, data);
}

void processconnections(void)
{
	connections.process();
}

void readyconnection(int id, bool ready)
{
	connections.ready(id, ready);
}

bool wantdpa = false;

bool GetDPAInput(void)
{
	return wantdpa;
}

FUSELIBFNDEF void SetDPAInput(int flag)
{
	if(flag == 0)
		wantdpa = false;
	else
		wantdpa = true;
}

void connection::close(int index)
{
	dynamic *data;

	ready = false;

	PCLogout(index);

	//click.type = 0;
	closesocket(s);
	available = true;

	data = getdata(npcid);

	if(data != NULL)
		data -> connection(-1);
}

void connection::dpaprocess(int index)
{
	int size;
	char buf[80000];

	size = recv(s, buf, 80000, 0);

	if(size != SOCKET_ERROR && size != 0) {
		ClientDecode(buf, size, ecpos, encrypt);
		DPAProcessInput(index, size, buf);
	}
}

// have a getmessage procedure which retrieves messages
// each message has a specified size.  if it is -1, it
// means that the size is specified as a short int that
// follows the message type byte
void connection::process(int index)
{
	char buf[2000], buf2[500], data1[1000];
	unsigned short size;
	short stemp1, stemp2, stemp3;
	int offset, count, c, i, temp1, temp2;
	bool val;

	if(GetBytes(buf, 1) > 0) {
		switch(buf[0]) {
		case 0x02:
			// send move requests
			if(GetBytes(buf + 1, 2) > 0) {
				buf[0] = FUSECMSG_MOVE1;
				ProcessInput(index, 3, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x03:
			// send console entry
			// console entry
			if(GetBytes(buf + 1, 2) > 0) {
				size = UnBufferShort(buf, 1);
				if((GetBytes(buf + 3, size - 3)) > 0) {
					buf[0] = FUSECMSG_CIN1;
					ProcessInput(index, size, buf);
				}
				else {
					close(index);
				}
			}
			else {
				close(index);
			}

			break;

		case 0x05:
			// attack npc
			if(GetBytes(buf + 1, 4) > 0) {
				buf[0] = FUSECMSG_ATTACK1;
				ProcessInput(index, 5, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x06:
			// item use and npc paperdoll
			if(GetBytes(buf + 1, 4) > 0) {
				buf[0] = FUSECMSG_DBLCLICK1;
				ProcessInput(index, 5, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x07:
			// pick up item
			if(GetBytes(buf + 1, 6) > 0) {
				buf[0] = FUSECMSG_PICKUP1;
				ProcessInput(index, 7, buf);
			}
			else {
				close(index);
			}

			break;

		// why is this message generated erroneously, sometimes?
		case 0x08:
			// drop item
			// notice that at this time, item count isn't being used...
			if(GetBytes(buf + 1, 13) > 0) {
				buf[0] = FUSECMSG_DROP1;
				ProcessInput(index, 14, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x09:
			// item/npc name
			if(GetBytes(buf + 1, 4) > 0) {
				buf[0] = FUSECMSG_CLICK1;
				ProcessInput(index, 5, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x12:
			// extended command
			if(GetBytes(buf + 1, 2) > 0) {
				size = UnBufferShort(buf, 1);
				if((GetBytes(buf + 3, size - 3)) > 0) {
					switch(buf[3]) {
					case 0x24:
						// what is second parameter? why is it always 0?
						if(sscanf(buf + 4, "%d %d", &temp1, &temp2) == 2) {
							buf[0] = FUSECMSG_SKILLUSE1;
							BufferShort(buf, 1, temp1);
							BufferShort(buf, 3, temp2);
							ProcessInput(index, 5, buf);
						}

						break;

					case 0x27:
						if(sscanf(buf + 4, "%d %d", &temp1, &temp2) == 2) {
							buf[0] = FUSECMSG_SPELLUSE2;
							BufferShort(buf, 1, temp1); // spell number (1-64)
							BufferInt(buf, 3, temp2); // spellbook id
							ProcessInput(index, 7, buf);
						}

						break;

					case 0x56:
						if(sscanf(buf + 4, "%d", &temp1) == 1) {
							buf[0] = FUSECMSG_SPELLUSE1;
							BufferShort(buf, 1, temp1);
							ProcessInput(index, 3, buf);
						}

						break;

					case (char)0xc7:
						if(sscanf(buf + 4, "%s", buf2) == 1) {
							buf[0] = FUSECMSG_ACTION1;
							strcpy(buf + 1, buf2);
							ProcessInput(index, 1 + strlen(buf2) + 1, buf);
						}

						break;
					}
				}
				else {
					close(index);
				}
			}
			else {
				close(index);
			}

			break;

		case 0x13:
			// equip item
			if(GetBytes(buf + 1, 9) > 0) {
				buf[0] = FUSECMSG_EQUIP1;
				ProcessInput(index, 10, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x22:
			// send move requests
			if(GetBytes(buf + 1, 2) > 0) {
				buf[0] = FUSECMSG_SET1;
				ProcessInput(index, 3, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x34:
			// send data select
			// data select
			if(GetBytes(buf + 1, 9) > 0) {
				buf[0] = FUSECMSG_DATA1;
				ProcessInput(index, 10, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x3b:
			// send purchase
			// purchase
			if(GetBytes(buf + 1, 2) > 0) {
				size = UnBufferShort(buf, 1);
				if((GetBytes(buf + 3, size - 3)) > 0) {
					data1[0] = FUSECMSG_PURCHASE1;
					BufferInt(data1, 1, UnBufferInt(buf, 3));
					BufferByte(data1, 5, UnBufferByte(buf, 7));

					temp1 = 8; // offset into buf
					temp2 = 6; // offset into data1

					for(stemp1 = 0; stemp1 < (unsigned char)UnBufferByte(buf, 7); stemp1++) {
						BufferInt(data1, temp2, UnBufferInt(buf, temp1 + 1));
						BufferShort(data1, temp2 + 4, UnBufferShort(buf, temp1 + 5));

						temp1 += 7;
						temp2 += 6;
					}

					ProcessInput(index, temp2, data1);
				}
				else {
					close(index);
				}
			}
			else {
				close(index);
			}

			break;

		case 0x66:
			// send book page
			// book page
			if(GetBytes(buf + 1, 2) > 0) {
				size = UnBufferShort(buf, 1);
				if((GetBytes(buf + 3, size - 3)) > 0) {
					if(UnBufferShort(buf, 7) == 0x0001 && UnBufferShort(buf, 11) == -1) {
						data1[0] = FUSECMSG_BOOKGETPAGE1;
						BufferInt(data1, 1, UnBufferInt(buf, 3));
						BufferShort(data1, 5, UnBufferShort(buf, 9));
						ProcessInput(index, 7, data1);
					}
					else {
						data1[0] = FUSECMSG_BOOKPAGES1;
						BufferInt(data1, 1, UnBufferInt(buf, 3));

						BufferShort(data1, 5, UnBufferShort(buf, 7));

						temp1 = 9; // offset into buf
						temp2 = 7; // offset into data1

						val = true;

						for(stemp1 = 0; stemp1 < UnBufferShort(buf, 7); stemp1++) {
							if(temp1 > size || temp2 > 700) {
								val = false;
								break;
							}

							BufferShort(data1, temp2, UnBufferShort(buf, temp1));
							temp1 += 2;
							temp2 += 2;

							stemp2 = UnBufferShort(buf, temp1);
							temp1 += 2;

							BufferShort(data1, temp2, stemp2);
							temp2 += 2;

							for(stemp3 = 0; stemp3 < stemp2; stemp3++) {
								count = strlen(buf + temp1);

								if(count > 127) {
									val = false;
									break;
								}

								BufferByte(data1, temp2, count);
								temp2 += 1;

								BufferString(data1, temp2, buf + temp1, count);
								temp2 += count;

								temp1 += count + 1;
							}

							if(!val)
								break;
						}

						if(val)
							ProcessInput(index, temp2/*size*/, data1);
					}
				}
				else {
					close(index);
				}
			}
			else {
				close(index);
			}

			break;

		case 0x6c:
			// send target response
			// client target
			if(GetBytes(buf + 1, 18) > 0) {
				buf[0] = FUSECMSG_TARGET1;
				ProcessInput(index, 19, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x6f:
			// secure trade
			if(GetBytes(buf + 1, 2) > 0) {
				size = UnBufferShort(buf, 1);
				if((GetBytes(buf + 3, size - 3)) > 0) {
					buf[0] = FUSECMSG_SECURETRADE1;
					ProcessInput(index, size, buf);
				}
				else {
					close(index);
				}
			}
			else {
				close(index);
			}

			break;

		case 0x72:
			// send combat requests
			if(GetBytes(buf + 1, 4) > 0) {
				buf[0] = FUSECMSG_MODE1;
				ProcessInput(index, 5, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x73:
			// we will process idle messages here
			// idle
			if(GetBytes(buf + 1, 1) > 0) {
				buf[0] = FUSECMSG_IDLE1;
				ProcessInput(index, 2, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x75:
			// npc rename
			if(GetBytes(buf + 1, 34) > 0) {
				buf[0] = FUSECMSG_NPCRENAME1;
				ProcessInput(index, 35, buf);
			}
			else {
				close(index);
			}

			break;

		case (char)0x7d:
			// gm page
			if(GetBytes(buf + 1, 12) > 0) {
				buf[0] = FUSECMSG_DIALOGBOX1;
				ProcessInput(index, 13, buf);
			}
			else {
				close(index);
			}

			break;

		case (char)0x95:
			// send dye chart response
			if(GetBytes(buf + 1, 8) > 0) {
				buf[0] = FUSECMSG_DYESELECT1;
				ProcessInput(index, 9, buf);
			}
			else {
				close(index);
			}

			break;

		case (char)0x9a:
			// send console prompt entry
			// console prompt entry
			if(GetBytes(buf + 1, 2) > 0) {
				size = UnBufferShort(buf, 1);
				if((GetBytes(buf + 3, size - 3)) > 0) {
					buf[0] = FUSECMSG_CPROMPT1;
					ProcessInput(index, size, buf);
				}
				else {
					close(index);
				}
			}
			else {
				close(index);
			}

			break;

		case (char)0x9b:
			// gm page
			if(GetBytes(buf + 1, 257) > 0) {
				buf[0] = FUSECMSG_GMPAGE1;
				ProcessInput(index, 1, buf);
			}
			else {
				close(index);
			}

			break;

		case 0x9f:
			// send sale
			// sale
			if(GetBytes(buf + 1, 2) > 0) {
				size = UnBufferShort(buf, 1);
				if((GetBytes(buf + 3, size - 3)) > 0) {
					data1[0] = FUSECMSG_SALE1;
					BufferInt(data1, 1, UnBufferInt(buf, 3));
					BufferShort(data1, 5, UnBufferShort(buf, 7));

					temp1 = 9; // offset into buf
					temp2 = 7; // offset into data1

					for(stemp1 = 0; stemp1 < UnBufferShort(buf, 7); stemp1++) {
						BufferInt(data1, temp2, UnBufferInt(buf, temp1));
						BufferShort(data1, temp2 + 4, UnBufferShort(buf, temp1 + 4));

						temp1 += 6;
						temp2 += 6;
					}

					ProcessInput(index, temp2, data1);
				}
				else {
					close(index);
				}
			}
			else {
				close(index);
			}

			break;

		case (char)0xa7:
			if(GetBytes(buf + 1, 3) > 0) {
				buf[0] = FUSECMSG_TIPS1;
				ProcessInput(index, 4, buf);
			}
			else {
				close(index);
			}

			break;

		default:
			size = recv(s, buf + 1, 999, 0);
			if(size > 0) {
				ClientDecode(buf + 1, size, ecpos, encrypt);

				count = size + 1;

				strcpy(buf2, "Client Message:");
				offset = strlen(buf2);

				for(i = 0; i < count; i++) {
					buf2[offset++] = ' ';

					c = (buf[i] >> 4) & 0x0000000F;
					if(c < 0xa)
						buf2[offset++] = c + '0';
					else
						buf2[offset++] = c - 0xa + 'a';

					c = buf[i] & 0x0000000F;
					if(c < 0xa)
						buf2[offset++] = c + '0';
					else
						buf2[offset++] = c - 0xa + 'a';
				}

				buf2[offset++] = ' ';
				buf2[offset++] = '(';

				for(i = 0; i < count; i++) {
					if(buf[i] < 0x20 || buf[i] > 0x7f)
						buf2[offset++] = '.';
					else
						buf2[offset++] = buf[i];
				}

				buf2[offset++] = ')';
				buf2[offset] = 0;

				offset = 0;
				BufferInt(data1, 0, 0xffffffff);
				BufferShort(data1, 4, (short)0xffff);
				CopyBuf(data1 + 6, STRSIZE, "");
				BufferByte(data1, STRSIZE + 6, 0x06);
				BufferShort(data1, STRSIZE + 7, 0x03b2);
				BufferShort(data1, STRSIZE + 9, 0x0003);
				BufferString(data1, STRSIZE + 11, buf2, strlen(buf2) + 1);
				PrepareMessage(FUSEMSG_CLIENTBARK1, buf, offset, data1);

				// enable for message display
				//Globalcast(-1, buf, offset);
			}
		}
	}
	else {
		close(index);
	}
}

int connectlist::add(SOCKET s, int acct, unsigned npcid, int ecpos, const char *encrypt, const char *data)
{
	int i, j;

	for(i = 0; i < MAXCONNECT; i++) {
		if(connects[i].available) {
			connects[i].available = false;
			// ready should already be false
			connects[i].setready(false);
			connects[i].acct = acct;
			connects[i].s = s;
			connects[i].ecpos = ecpos;
			//connects[i].click.type = 0;
			connects[i].pool.id = 0;
			connects[i].pool.count = 0;
			//WSAEventSelect(s, events[i], FD_READ);

			for(j = 0; j < 64; j++)
				connects[i].encrypt[j] = encrypt[j];

			if(npcid == 0) {
				// create npc
				connects[i].npcid = CreateNPC();

				CreatePlayer(connects[i].npcid, data);

				AddPlayer(connects[i].acct, connects[i].npcid);
			}
			else {
				// assign npc
				connects[i].npcid = npcid;
			}

			npcconnect(connects[i].npcid, i);
			//connects[i].setnpc();
			PCLogin(i);

			break;
		}
	}

	if(i < MAXCONNECT) {
		return i;
	}
	else {
		return -1;
	}
}

/*void connectlist::broadcast(fieldlist &list, int id, char *buf, int size)
{
	dynamic *data;
	int i, field;

	for(i = 0; i < MAXCONNECT; i++) {
		if(i != id && connects[i].ready) {
			data = getdata(connects[i].npcid);

			field = GetField(connects[i].npcid);

			if(FieldInList(field, list)) {
				senddata(i, buf, size);
			}
		}
	}
}*/

FUSELIBFNDEF int DPAPrepareMessage(char *output, int &oindex, const char *buf, int offset)
{
	int size;

	// encoding section
	if(offset > 0) {
		size = PutMessage(buf, offset, output + oindex);
	}
	else {
		size = 0;
	}

	oindex += size;

	return size;
}

// buf had better be long enough
FUSELIBFNDEF int PrepareMessage(unsigned type, char *output, int &oindex, const char *data)
{
	char buf[10000];
	int offset, size, temp, itemp;
	short stemp, stemp2, stemp3;
	bool b1, b2, b3, b4, b5;

	offset = 0;

	switch(type) {
	// status
	case FUSEMSG_PLAYERSTATUS1:
		BufferByte(buf, offset + 0, 0x11);

		BufferShort(buf, offset + 1, 66); // size
		BufferInt(buf, offset + 3, UnBufferInt(data, 0)); // id
		CopyBuf(buf + offset + 7, STRSIZE, data + 4); // name
		BufferShort(buf, offset + 37, UnBufferShort(data, STRSIZE + 4)); // hp1
		BufferShort(buf, offset + 39, UnBufferShort(data, STRSIZE + 6)); // hp2
		BufferByte(buf, offset + 41, UnBufferByte(data, STRSIZE + 8)); // renameable
		BufferByte(buf, offset + 42, UnBufferByte(data, STRSIZE + 9)); // ??
		BufferByte(buf, offset + 43, UnBufferByte(data, STRSIZE + 10)); // gender
		BufferShort(buf, offset + 44, UnBufferShort(data, STRSIZE + 11)); // str
		BufferShort(buf, offset + 46, UnBufferShort(data, STRSIZE + 13)); // dex
		BufferShort(buf, offset + 48, UnBufferShort(data, STRSIZE + 15)); // int
		BufferShort(buf, offset + 50, UnBufferShort(data, STRSIZE + 17)); // stam1
		BufferShort(buf, offset + 52, UnBufferShort(data, STRSIZE + 19)); // stam2
		BufferShort(buf, offset + 54, UnBufferShort(data, STRSIZE + 21)); // mana1
		BufferShort(buf, offset + 56, UnBufferShort(data, STRSIZE + 23)); // mana2
		BufferInt(buf, offset + 58, UnBufferInt(data, STRSIZE + 25)); // gold
		BufferShort(buf, offset + 62, UnBufferShort(data, STRSIZE + 29)); // ar
		BufferShort(buf, offset + 64, UnBufferShort(data, STRSIZE + 31)); // weight

		offset += 66;

		break;

	case FUSEMSG_NPCSTATUS1:
		BufferByte(buf, offset + 0, 0x11);

		BufferShort(buf, offset + 1, 43);
		BufferInt(buf, offset + 3, UnBufferInt(data, 0)); // id
		CopyBuf(buf + offset + 7, STRSIZE, data + 4); // name
		BufferShort(buf, offset + 37, UnBufferShort(data, STRSIZE + 4)); // hp1
		BufferShort(buf, offset + 39, UnBufferShort(data, STRSIZE + 6)); // hp2
		BufferByte(buf, offset + 41, UnBufferByte(data, STRSIZE + 8)); // renameable
		BufferByte(buf, offset + 42, UnBufferByte(data, STRSIZE + 9)); // ??

		offset += 43;

		break;

	// item placement (ground)
	case FUSEMSG_ITEMPLACE1:
		// extra modifies id
		if(UnBufferShort(data, 4) != 0) {
			temp = UnBufferInt(data, 0) | 0x80000000;
			b1 = true;
		}
		else {
			temp = UnBufferInt(data, 0);
			b1 = false;
		}

		// color modifies y
		if(UnBufferShort(data, 6) != 0) {
			b2 = true;
			stemp = UnBufferShort(data, 8) | 0x8000;
		}
		else {
			b2 = false;
			stemp = UnBufferShort(data, 8);
		}

		// flag modifies y
		if(UnBufferByte(data, 10) != 0) {
			b3 = true;
			stemp |= 0x4000;
		}
		else {
			b3 = false;
		}

		// flag2 modifies x
		if(UnBufferByte(data, 11) != 0) {
			b4 = true;
			stemp2 = UnBufferShort(data, 12) | 0x8000;
		}
		else {
			b4 = false;
			stemp2 = UnBufferShort(data, 12);
		}

		// flag3 modifies model
		if(UnBufferShort(data, 14) != 0) {
			b5 = true;
			stemp3 = UnBufferShort(data, 15) | 0x8000;
		}
		else {
			b5 = false;
			stemp3 = UnBufferShort(data, 15);
		}

		itemp = 0;

		BufferByte(buf, offset + itemp, 0x1a);
		itemp += 3; // fill in size later

		BufferInt(buf, offset + itemp, temp); // id
		itemp += 4;

		BufferShort(buf, offset + itemp, stemp3); // model
		itemp += 2;

		if(b5) {
			BufferByte(buf, offset + itemp, UnBufferByte(data, 14)); // flag3
			itemp += 1;
		}

		if(b1) {
			BufferShort(buf, offset + itemp, UnBufferShort(data, 4)); // extra
			itemp += 2;
		}

		BufferShort(buf, offset + itemp, stemp2); // x
		itemp += 2;

		BufferShort(buf, offset + itemp, stemp); // y
		itemp += 2;

		if(b4) {
			BufferByte(buf, offset + itemp, UnBufferByte(data, 11)); // flag2
			itemp += 1;
		}

		BufferByte(buf, offset + itemp, UnBufferByte(data, 17)); // z
		itemp += 1;

		if(b2) {
			BufferShort(buf, offset + itemp, UnBufferShort(data, 6)); // color
			itemp += 2;
		}

		if(b3) {
			BufferByte(buf, offset + itemp, UnBufferByte(data, 10)); // flag
			itemp += 1;
		}

		BufferShort(buf, offset + 1, itemp); // size

		offset += itemp;

		break;

	// startup - whoafully incomplete
	case FUSEMSG_PLAYERLOGIN1:
		memset(buf + offset, 0, 37);

		BufferByte(buf, offset + 0, 0x1b);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // id
		BufferShort(buf, offset + 9, UnBufferShort(data, 4)); // model
		BufferShort(buf, offset + 11, UnBufferShort(data, 6)); // x
		BufferShort(buf, offset + 13, UnBufferShort(data, 8)); // y
		BufferShort(buf, offset + 15, UnBufferByte(data, 10)); // z
		BufferByte(buf, offset + 17, UnBufferByte(data, 11)); // facing
		BufferShort(buf, offset + 20, (short)0xff7f);
		BufferShort(buf, offset + 25, 0x07f0);
		BufferShort(buf, offset + 27, 0x0a80);
		BufferShort(buf, offset + 29, 0x0810);

		offset += 37;

		break;

	// speech
	case FUSEMSG_CLIENTBARK1:
		temp = strlen(data + STRSIZE + 11);

		BufferByte(buf, offset + 0, 0x1c);
		BufferShort(buf, offset + 1, 44 + temp + 1);
		BufferInt(buf, offset + 3, UnBufferInt(data, 0)); // id
		BufferShort(buf, offset + 7, UnBufferShort(data, 4)); // model
		BufferByte(buf, offset + 9, UnBufferByte(data, STRSIZE + 6)); // type
		BufferShort(buf, offset + 10, UnBufferShort(data, STRSIZE + 7)); // color
		BufferShort(buf, offset + 12, UnBufferShort(data, STRSIZE + 9)); // font
		CopyBuf(buf + offset + 14, STRSIZE, data + 6); // name
		CopyBuf(buf + offset + 44, temp + 1, data + STRSIZE + 11); // text

		offset += 44 + temp + 1;

		break;

	// object remove
	case FUSEMSG_OBJECTREMOVE1:
		BufferByte(buf, offset + 0, 0x1d);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // id

		offset += 5;

		break;

	// place local npc
	case FUSEMSG_PLAYERPLACE1:
		memset(buf + offset, 0, 19);

		BufferByte(buf, offset + 0, 0x20);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // id
		BufferShort(buf, offset + 5, UnBufferShort(data, 4)); // model
		BufferShort(buf, offset + 8, UnBufferShort(data, 6)); // color
		BufferByte(buf, offset + 10, UnBufferByte(data, 8)); // show
		BufferShort(buf, offset + 11, UnBufferShort(data, 9)); // x
		BufferShort(buf, offset + 13, UnBufferShort(data, 11)); // y
		BufferByte(buf, offset + 17, UnBufferByte(data, 13)); // facing
		BufferByte(buf, offset + 18, UnBufferByte(data, 14)); // z

		BufferByte(buf, offset + 7, (char)0xff); // reputation??
		BufferShort(buf, offset + 15, (short)0xffff); // ??

		offset += 19;

		break;

	// reset local npc (response)
	case FUSEMSG_PLAYERRESET1:
		memset(buf + offset, 0, 8);

		BufferByte(buf, offset + 0, 0x21);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // serial
		BufferShort(buf, offset + 2, UnBufferShort(data, 1)); // x
		BufferShort(buf, offset + 4, UnBufferShort(data, 3)); // y
		BufferByte(buf, offset + 6, UnBufferByte(data, 5)); // facing
		BufferByte(buf, offset + 7, UnBufferByte(data, 6)); // z

		offset += 8;

		break;

	// move local npc (response)
	case FUSEMSG_PLAYERSET1:
		memset(buf + offset, 0, 3);

		BufferByte(buf, offset + 0, 0x22);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // serial
		BufferByte(buf, offset + 2, UnBufferByte(data, 1)); // reputation

		offset += 3;

		break;

	// drag item
	case FUSEMSG_ITEMDRAG1:
		BufferByte(buf, offset + 0, 0x23);

		BufferShort(buf, offset + 1, UnBufferShort(data, 0)); // model
		BufferByte(buf, offset + 3, 0x00/*UnBufferByte(data, 22)*/); // flag3
		BufferShort(buf, offset + 4, 0x0000); // NOT color/extra
		BufferShort(buf, offset + 6, 0x0000); // NOT color/extra
		BufferInt(buf, offset + 8, UnBufferInt(data, 2)); // source id
		BufferShort(buf, offset + 12, UnBufferShort(data, 6)); // source x
		BufferShort(buf, offset + 14, UnBufferShort(data, 8)); // source y
		BufferByte(buf, offset + 16, UnBufferByte(data, 10)); // source z
		BufferInt(buf, offset + 17, UnBufferInt(data, 11)); // dest id
		BufferShort(buf, offset + 21, UnBufferShort(data, 15)); // dest x
		BufferShort(buf, offset + 23, UnBufferShort(data, 17)); // dest y
		BufferByte(buf, offset + 25, UnBufferByte(data, 19)); // dest z

		offset += 26;

		break;

	// open container window
	case FUSEMSG_ITEMGUMP1:
		BufferByte(buf, offset + 0, 0x24);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // container id
		BufferShort(buf, offset + 5, UnBufferShort(data, 4)); // gumpid

		offset += 7;

		break;

	// item placement (container)
	case FUSEMSG_ITEMCONTAIN1:
		itemp = 0;

		BufferByte(buf, offset + itemp, 0x25);
		itemp += 1;

		BufferInt(buf, offset + itemp, UnBufferInt(data, 0)); // id
		itemp += 4;

		BufferShort(buf, offset + itemp, UnBufferShort(data, 4)); // model
		itemp += 2;

		BufferByte(buf, offset + itemp, UnBufferByte(data, 6)); // flag3
		itemp += 1;

		BufferShort(buf, offset + itemp, UnBufferShort(data, 7)); // extra
		itemp += 2;

		BufferShort(buf, offset + itemp, UnBufferShort(data, 9)); // x
		itemp += 2;

		BufferShort(buf, offset + itemp, UnBufferShort(data, 11)); // y
		itemp += 2;

		BufferInt(buf, offset + itemp, UnBufferInt(data, 13)); // container id
		itemp += 4;

		BufferShort(buf, offset + itemp, UnBufferShort(data, 17)); // color
		itemp += 2;

		offset += itemp;

		break;

	// player kick
	case FUSEMSG_PLAYERKICK1:
		BufferByte(buf, offset + 0, 0x26);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // ??

		offset += 5;

		break;

	// item pickup
	case FUSEMSG_PICKUPCANCEL1:
		BufferByte(buf, offset + 0, 0x27);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // nodragid

		offset += 2;

		break;

	// ??
	case 0x29:
		BufferByte(buf, offset + 0, 0x29);

		offset += 1;

		break;

	// death dialog
	case FUSEMSG_DEATHOPTION1:
		BufferByte(buf, offset + 0, 0x2c);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); //deathid

		offset += 2;

		break;

	// item placement (npc)
	case FUSEMSG_ITEMEQUIP1:
		BufferByte(buf, offset + 0, 0x2e);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // id
		BufferShort(buf, offset + 5, UnBufferShort(data, 4)); // model
		BufferByte(buf, offset + 7, UnBufferByte(data, 6)); // flag2
		BufferByte(buf, offset + 8, UnBufferByte(data, 7)); // equip
		BufferInt(buf, offset + 9, UnBufferInt(data, 8)); // container id
		BufferShort(buf, offset + 13, UnBufferShort(data, 12)); // color

		offset += 15;

		break;

	// client freeze
	case FUSEMSG_CLIENTFREEZE1:
		memset(buf + offset, 0, 2);

		BufferByte(buf, offset + 0, 0x33);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // freezeid

		offset += 2;

		break;

	// skills
	case FUSEMSG_PLAYERSKILLS1:
		BufferByte(buf, offset + 0, 0x3a);
		temp = offset;

		if(UnBufferShort(data, 0) == 1) {
			// this must mean no reset
			// plus list not null terminated
			BufferByte(buf, offset + 3, (char)0xff); // type - single
			BufferShort(buf, offset + 4, UnBufferShort(data, 2)); // skillid
			BufferShort(buf, offset + 6, UnBufferShort(data, 4)); // skillval

			offset += 8;
		}
		else {
			// this must mean reset all
			// plus list null terminated
			BufferByte(buf, offset + 3, 0x00); // type - list
			offset += 4;

			for(stemp = 0/*mi -> word2*/; stemp < UnBufferShort(data, 0)/*mi -> word2 + mi -> word1*/; stemp++) {
				BufferShort(buf, offset, UnBufferShort(data, 2 + 4 * stemp + 0)); // skillid
				offset += 2;
				BufferShort(buf, offset, UnBufferShort(data, 2 + 4 * stemp + 2)); // skillval
				offset += 2;
			}

			BufferShort(buf, offset, 0); // list terminator
			offset += 2;
		}

		BufferShort(buf, temp + 1, offset - temp); // size

		break;

	// multiple item placement (container)
	case FUSEMSG_CONTAINERSETUP1:
		itemp = 0;

		BufferByte(buf, offset + itemp, 0x3c);
		itemp += 1;

		BufferShort(buf, offset + itemp, UnBufferShort(data, 0) * 19 + 5); // size
		itemp += 2;

		BufferShort(buf, offset + itemp, UnBufferShort(data, 0)); // count
		itemp += 2;

		for(stemp = 0; stemp < UnBufferShort(data, 0); stemp++) {
			BufferInt(buf, offset + itemp, UnBufferInt(data, 2 + 19 * stemp + 0)); // id
			itemp += 4;

			BufferShort(buf, offset + itemp, UnBufferShort(data, 2 + 19 * stemp + 4)); // model
			itemp += 2;

			BufferByte(buf, offset + itemp, UnBufferByte(data, 2 + 19 * stemp + 6)); // flag3
			itemp += 1;

			BufferShort(buf, offset + itemp, UnBufferShort(data, 2 + 19 * stemp + 7)); // extra
			itemp += 2;

			BufferShort(buf, offset + itemp, UnBufferShort(data, 2 + 19 * stemp + 9)); // x
			itemp += 2;

			BufferShort(buf, offset + itemp, UnBufferShort(data, 2 + 19 * stemp + 11)); // y
			itemp += 2;

			BufferInt(buf, offset + itemp, UnBufferInt(data, 2 + 19 * stemp + 13)); // countainer id
			itemp += 4;

			BufferShort(buf, offset + itemp, UnBufferShort(data, 2 + 19 * stemp + 17)); // color
			itemp += 2;
		}

		offset += itemp;

		break;

	// npc-specific lighting
	case FUSEMSG_NPCLIGHT1:
		memset(buf + offset, 0, 6);

		BufferByte(buf, offset + 0, 0x4e);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // id
		BufferByte(buf, offset + 5, UnBufferByte(data, 4)); // lightmod

		offset += 6;

		break;

	// global lighting
	case FUSEMSG_GLOBALLIGHT1:
		//memset(buf + offset, 0, 2);

		BufferByte(buf, offset + 0, 0x4f);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // light

		offset += 2;

		break;

	// idle warning
	case FUSEMSG_IDLEWARN1:
		BufferByte(buf, offset + 0, 0x53);
		BufferByte(buf, offset + 1, 0x07);

		offset += 2;

		break;

	// play sound
	case FUSEMSG_SOUND1:
		BufferByte(buf, offset + 0, 0x54);

		BufferByte(buf, offset + 1, 0x01); // quiet, infinite repeat or single shot (0/1)
		BufferShort(buf, offset + 2, UnBufferShort(data, 0)); // soundid
		BufferByte(buf, offset + 4, 0x00); // ??
		BufferByte(buf, offset + 5, 0x00); // ??
		//BufferShort(buf, offset + 4, 0); // maybe line of sight stuff in here?
		BufferShort(buf, offset + 6, UnBufferShort(data, 2)); // x
		BufferShort(buf, offset + 8, UnBufferShort(data, 4)); // y
		BufferShort(buf, offset + 10, UnBufferByte(data, 6)); // z

		offset += 12;

		break;

	// ??
	case FUSEMSG_UNKNOWN1:
		BufferByte(buf, offset + 0, 0x55);

		offset += 1;

		break;

	// weather
	case FUSEMSG_WEATHER1:
		BufferByte(buf, offset + 0, 0x65);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // weatherid
		BufferShort(buf, offset + 2, UnBufferShort(data, 1)); // weather density

		offset += 4;

		break;

	// book pages
	case FUSEMSG_BOOKPAGES1:
		BufferByte(buf, offset + 0, (char)0x66);
		BufferInt(buf, offset + 3, UnBufferInt(data, 0));
		BufferShort(buf, offset + 7, UnBufferShort(data, 4));

		temp = 9; // offset into buf
		itemp = 6; // offset into data

		for(stemp = 0; stemp < UnBufferShort(data, 4); stemp++) {
			// page id
			BufferShort(buf, offset + temp, UnBufferShort(data, itemp));
			temp += 2;
			itemp += 2;

			// page line count
			stemp2 = UnBufferShort(data, itemp);
			itemp += 2;

			BufferShort(buf, offset + temp, stemp2);
			temp += 2;

			for(stemp3 = 0; stemp3 < stemp2; stemp3++) {
				BufferString(buf, offset + temp, data + itemp + 1, (unsigned char)(UnBufferByte(data, itemp)));
				temp += (unsigned char)UnBufferByte(data, itemp);
				buf[offset + temp] = 0;
				temp += 1;

				itemp += 1 + (unsigned char)UnBufferByte(data, itemp);
			}
		}

		BufferShort(buf, offset + 1, temp);

		offset += temp;

		break;

	// target select
	case FUSEMSG_TARGET1:
		memset(buf + offset, 0, 19);

		BufferByte(buf, offset + 0, 0x6c);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // type
		BufferByte(buf, offset + 2, UnBufferInt(data, 4)); // id

		offset += 19;

		break;

	// play music
	case FUSEMSG_MUSIC1:
		BufferByte(buf, offset + 0, 0x6d);

		BufferShort(buf, offset + 1, UnBufferShort(data, 0)); // musicid

		offset += 3;

		break;

	// animate
	case FUSEMSG_NPCANIMATE1:
		memset(buf + offset, 0, 14);

		BufferByte(buf, offset + 0, 0x6e);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // id
		BufferShort(buf, offset + 5, UnBufferShort(data, 4)); // animid
		BufferByte(buf, offset + 7, 0x05); // ??
		BufferByte(buf, offset + 8, 0x07); // direction?? (nwo, methinks tis incorrect)
		BufferShort(buf, offset + 9, 0x0001); // num. consec. anims (0 is all/infinite?)
		BufferByte(buf, offset + 11, 0x00); // backwards (0/1)
		BufferByte(buf, offset + 12, 0x00); // forward and back (0/1)
		BufferByte(buf, offset + 13, 0x00); // frame delay

		offset += 14;

		break;

	// secure trade
	case FUSEMSG_SECURETRADE1:
		BufferByte(buf, offset + 0, 0x6f);

		BufferByte(buf, offset + 3, UnBufferByte(data, 0)); // type
		BufferInt(buf, offset + 4, UnBufferInt(data, 1)); // id
		BufferInt(buf, offset + 8, UnBufferInt(data, 5)); // id/flag
		BufferInt(buf, offset + 12, UnBufferInt(data, 9)); // id/flag
		BufferByte(buf, offset + 16, UnBufferByte(data, 13)); // nameflag

		if(UnBufferByte(data, 13) == 0) {
			BufferShort(buf, offset + 1, 17); // size
			offset += 17;
		}
		else {
			BufferShort(buf, offset + 1, 47); // size
			CopyBuf(buf + offset + 17, 30, data + 14); // name
			offset += 47;
		}

		break;

	// effect
	case FUSEMSG_EFFECT1:
		BufferByte(buf, offset + 0, 0x70);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // type
		BufferInt(buf, offset + 2, UnBufferInt(data, 1)); // source id
		BufferInt(buf, offset + 6, UnBufferInt(data, 5)); // dest id
		BufferShort(buf, offset + 10, UnBufferShort(data, 9)); // model
		BufferShort(buf, offset + 12, UnBufferShort(data, 11)); // source x
		BufferShort(buf, offset + 14, UnBufferShort(data, 13)); // source y
		BufferByte(buf, offset + 16, UnBufferByte(data, 15)); // source z
		BufferShort(buf, offset + 17, UnBufferShort(data, 16)); // dest x
		BufferShort(buf, offset + 19, UnBufferShort(data, 18)); // dest y
		BufferByte(buf, offset + 21, UnBufferByte(data, 20)); // dest z
		// flag3, speed, duration, repeat count...
		BufferByte(buf, offset + 22, 0x0a);
		BufferByte(buf, offset + 23, 0x0a);
		BufferShort(buf, offset + 24, 0); // color?
		BufferShort(buf, offset + 26, 0x0100); // ???

		offset += 28;

		break;

	// local npc mode
	case FUSEMSG_PLAYERMODE1:
		//memset(buf + offset, 0, 5);
		BufferByte(buf, offset + 0, 0x72);

		BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // are 1 and 0 the only valid values?  what about 2?
		BufferByte(buf, offset + 2, 0x00);
		BufferByte(buf, offset + 3, 0x32);
		BufferByte(buf, offset + 4, 0x00);

		offset += 5;

		break;

	// local npc mode
	case FUSEMSG_PLAYERMODE2:
		//memset(buf + offset, 0, 5);
		BufferByte(buf, offset + 0, 0x72);
		BufferInt(buf, offset + 1, UnBufferInt(data, 0));

		/*BufferByte(buf, offset + 1, UnBufferByte(data, 0)); // are 1 and 0 the only valid values?  what about 2?
		BufferByte(buf, offset + 2, 0x00);
		BufferByte(buf, offset + 3, 0x32);
		BufferByte(buf, offset + 4, 0x00);*/

		offset += 5;

		break;

	// buy window
	case FUSEMSG_BUYWINDOW1:
		BufferByte(buf, offset + 0, (char)0x74);
		BufferInt(buf, offset + 3, UnBufferInt(data, 0));
		BufferByte(buf, offset + 7, UnBufferByte(data, 4));

		temp = 8; // offset into buf
		itemp = 5; // offset into data

		for(stemp = 0; stemp < (unsigned char)UnBufferByte(data, 4); stemp++) {
			// price
			BufferInt(buf, offset + temp, UnBufferInt(data, itemp));
			temp += 4;
			itemp += 4;

			// descriplen
			BufferByte(buf, offset + temp, UnBufferByte(data, itemp));

			// description
			BufferString(buf, offset + temp + 1, data + itemp + 1, (unsigned char)UnBufferByte(data, itemp));

			temp += 1 + (unsigned char)UnBufferByte(data, itemp);
			itemp += 1 + (unsigned char)UnBufferByte(data, itemp);
		}

		BufferShort(buf, offset + 1, temp);

		offset += temp;

		break;

	// npc move
	case FUSEMSG_NPCSET1:
		memset(buf + offset, 0, 17);

		BufferByte(buf, offset + 0, 0x77);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0));
		BufferShort(buf, offset + 5, UnBufferShort(data, 4));
		BufferShort(buf, offset + 7, UnBufferShort(data, 6));
		BufferShort(buf, offset + 9, UnBufferShort(data, 8));
		BufferByte(buf, offset + 11, UnBufferByte(data, 10));
		BufferByte(buf, offset + 12, UnBufferByte(data, 11));
		BufferShort(buf, offset + 13, UnBufferShort(data, 12));
		BufferByte(buf, offset + 15, UnBufferByte(data, 14));
		BufferByte(buf, offset + 16, UnBufferByte(data, 15));

		offset += 17;

		break;

	// npc placement
	case FUSEMSG_NPCSETUP1:
		memset(buf + offset, 0, 19);

		BufferByte(buf, offset + 0, 0x78);

		BufferInt(buf, offset + 3, UnBufferInt(data, 0));
		BufferShort(buf, offset + 7, UnBufferShort(data, 4));
		BufferShort(buf, offset + 9, UnBufferShort(data, 6));
		BufferShort(buf, offset + 11, UnBufferShort(data, 8));
		BufferByte(buf, offset + 13, UnBufferByte(data, 10));
		BufferByte(buf, offset + 14, UnBufferByte(data, 11));
		BufferShort(buf, offset + 15, UnBufferShort(data, 12));
		BufferByte(buf, offset + 17, UnBufferByte(data, 14));
		BufferByte(buf, offset + 18, UnBufferByte(data, 15));

		temp = 19;

		// add other equipped items.
		for(stemp = 0; stemp < UnBufferShort(data, 16); stemp++) {
			BufferInt(buf, offset + temp + 0, UnBufferInt(data, 18 + 9 * stemp + 0));

			if(UnBufferShort(data, 18 + 9 * stemp + 7) != 0) {
				BufferShort(buf, offset + temp + 4, (short)(UnBufferShort(data, 18 + 9 * stemp + 4) | 0x8000));
				BufferByte(buf, offset + temp + 6, UnBufferByte(data, 18 + 9 * stemp + 6));
				BufferShort(buf, offset + temp + 7, UnBufferShort(data, 18 + 9 * stemp + 7));

				temp += 9;
			}
			else {
				BufferShort(buf, offset + temp + 4, UnBufferShort(data, 18 + 9 * stemp + 4));
				BufferByte(buf, offset + temp + 6, UnBufferByte(data, 18 + 9 * stemp + 6));

				temp += 7;
			}
		}

		// trailing null item
		memset(buf + offset + temp, 0, 4);
		temp += 4;

		BufferShort(buf, offset + 1, temp);

		offset += temp;

		break;

	// dialog box
	case FUSEMSG_DIALOGBOX1:
		BufferByte(buf, offset + 0, (char)0x7c);
		BufferShort(buf, offset + 1, UnBufferShort(data, 6) + 9);
		BufferInt(buf, offset + 3, UnBufferInt(data, 0));
		BufferShort(buf, offset + 7, UnBufferShort(data, 4));
		BufferString(buf, offset + 9, data + 8, UnBufferShort(data, 6));

		offset += UnBufferShort(data, 6) + 9;

		break;

	// paper doll display
	case FUSEMSG_NPCPAPERDOLL1:
		memset(buf + offset, 0, 66);

		BufferByte(buf, offset + 0, (char)0x88);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // id
		CopyBuf(buf + offset + 5, 60, data + 4); // long name
		BufferByte(buf, offset + 65, UnBufferByte(data, 64)); // war mode flag

		offset += 66;

		break;

	// equip corpse
	case FUSEMSG_CORPSESETUP1:
		itemp = 0;

		BufferByte(buf, offset + itemp, (char)0x89);
		itemp += 1;

		itemp += 2; // leave room for size

		BufferInt(buf, offset + itemp, UnBufferInt(data, 0));
		itemp += 4;

		for(stemp = 0; stemp < UnBufferShort(data, 4); stemp++) {
			BufferByte(buf, offset + itemp, UnBufferByte(data, 6 + 5 * stemp + 4));
			itemp += 1;

			BufferInt(buf, offset + itemp, UnBufferInt(data, 6 + 5 * stemp + 0));
			itemp += 4;
		}

		BufferByte(buf, offset + itemp, 0);
		itemp += 1;

		BufferShort(buf, offset + 1, itemp);

		offset += itemp;

		break;

	// book cover
	case FUSEMSG_BOOKCOVER1:
		BufferByte(buf, offset + 0, (char)0x93);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0));
		BufferByte(buf, offset + 5, UnBufferByte(data, 4));
		BufferShort(buf, offset + 6, UnBufferShort(data, 5));
		CopyBuf(buf + offset + 8, 60, data + 7);
		CopyBuf(buf + offset + 68, 30, data + 67);

		offset += 98;

		break;

	// dye window
	case FUSEMSG_DYEOPTION1:
		memset(buf + offset, 0, 9);

		BufferByte(buf, offset + 0, (char)0x95);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0)); // id
		BufferShort(buf, offset + 5, UnBufferShort(data, 4)); // model??
		BufferShort(buf, offset + 7, UnBufferShort(data, 4)); // color

		offset += 9;

		break;

	// console prompt
	case FUSEMSG_CPROMPT1:
		BufferByte(buf, offset + 0, (char)0x9a);

		BufferShort(buf, offset + 1, 16);
		BufferInt(buf, offset + 3, UnBufferInt(data, 0));
		BufferInt(buf, offset + 7, UnBufferInt(data, 4));
		BufferInt(buf, offset + 11, UnBufferInt(data, 8));
		BufferByte(buf, offset + 15, 0);

		offset += 16;

		break;

	// sell window
	case FUSEMSG_SELLWINDOW1:
		BufferByte(buf, offset + 0, (char)0x9e);
		BufferInt(buf, offset + 3, UnBufferInt(data, 0));
		BufferShort(buf, offset + 7, UnBufferShort(data, 4));

		temp = 9; // offset into buf
		itemp = 6; // offset into data

		for(stemp = 0; stemp < UnBufferShort(data, 4); stemp++) {
			// item id
			BufferInt(buf, offset + temp, UnBufferInt(data, itemp));
			temp += 4;
			itemp += 4;

			// model
			BufferShort(buf, offset + temp, UnBufferShort(data, itemp));
			temp += 2;
			itemp += 2;

			// color
			BufferShort(buf, offset + temp, UnBufferShort(data, itemp));
			temp += 2;
			itemp += 2;

			// count
			BufferShort(buf, offset + temp, UnBufferShort(data, itemp));
			temp += 2;
			itemp += 2;

			// price
			BufferShort(buf, offset + temp, UnBufferShort(data, itemp));
			temp += 2;
			itemp += 2;

			// descriplen
			BufferShort(buf, offset + temp, UnBufferShort(data, itemp));

			// description
			BufferString(buf, offset + temp + 2, data + itemp + 2, UnBufferShort(data, itemp));

			temp += 2 + UnBufferShort(data, itemp);
			itemp += 2 + UnBufferShort(data, itemp);
		}

		BufferShort(buf, offset + 1, temp);

		offset += temp;

		break;

	// hp update
	case FUSEMSG_UPDATEHP1:
		BufferByte(buf, offset + 0, (char)0xa1);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0));
		BufferShort(buf, offset + 5, UnBufferShort(data, 4));
		BufferShort(buf, offset + 7, UnBufferShort(data, 6));

		offset += 9;

		break;

	// mana update
	case FUSEMSG_UPDATEMANA1:
		BufferByte(buf, offset + 0, (char)0xa2);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0));
		BufferShort(buf, offset + 5, UnBufferShort(data, 4));
		BufferShort(buf, offset + 7, UnBufferShort(data, 6));

		offset += 9;

		break;

	// stamina update
	case FUSEMSG_UPDATESTAMINA1:
		BufferByte(buf, offset + 0, (char)0xa3);

		BufferInt(buf, offset + 1, UnBufferInt(data, 0));
		BufferShort(buf, offset + 5, UnBufferShort(data, 4));
		BufferShort(buf, offset + 7, UnBufferShort(data, 6));

		offset += 9;

		break;

	// tips/updates
	case FUSEMSG_BROWSER1:
		BufferByte(buf, offset + 0, (char)0xa5);
		BufferShort(buf, offset + 1, 3 + UnBufferShort(data, 0) + 1); // size
		BufferString(buf, offset + 3, data + 2, UnBufferShort(data, 0)); // url
		BufferByte(buf, offset + 3 + UnBufferShort(data, 0), 0x00); // null terminator

		offset += 3 + UnBufferShort(data, 0) + 1;

		break;

	// tips/updates
	case FUSEMSG_TIPS1:
		BufferByte(buf, offset + 0, (char)0xa6);
		BufferShort(buf, offset + 1, 10 + UnBufferShort(data, 3));
		BufferByte(buf, offset + 3, UnBufferByte(data, 0));
		BufferShort(buf, offset + 4, 0); // ??
		BufferShort(buf, offset + 6, UnBufferShort(data, 1));
		BufferShort(buf, offset + 8, UnBufferShort(data, 3));
		BufferString(buf, offset + 10, data + 5, UnBufferShort(data, 3));

		offset += 10 + UnBufferShort(data, 3);

		break;

	/*default:
		offset = 0;*/
	}

	// encoding section
	if(offset > 0) {
		size = PutMessage(buf, offset, output + oindex);
	}
	else {
		size = 0;
	}

	oindex += size;

	return size;
}

FUSELIBFNDEF void SendData(int id, char *buf, int size)
{
	connections.senddata(id, buf, size);
}

/*void Broadcast(int field, int id, char *buf, int size)
{
	fieldlist list;

	FieldNeighbors(field, list);

	connections.broadcast(list, id, buf, size);
}

void BroadcastList(fieldlist &list, int id, char *buf, int size)
{
	connections.broadcast(list, id, buf, size);
}*/

FUSELIBFNDEF void Globalcast(int id, char *buf, int size)
{
	connections.globalcast(id, buf, size);
}

FUSELIBFNDEF void ConnectionPool(int id, int itemid, short count)
{
	connections.pool(id, itemid, count);
}

FUSELIBFNDEF void ConnectionUnpool(int id)
{
	connections.unpool(id);
}

FUSELIBFNDEF int ConnectionPoolID(int id)
{
	return connections.poolid(id);
}

FUSELIBFNDEF int ConnectionPoolCount(int id)
{
	return connections.poolcount(id);
}

FUSELIBFNDEF int GetConnectionNPCID(int id)
{
	return connections.npcid(id);
}

FUSELIBFNDEF int GetMaxConnectionCount(void)
{
	return MAXCONNECT;
}
