// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

// we should set FD_SETSIZE to MAXCONNECT before including windows.h, if we need
// more than 64 sockets to be used in a call to select, however 64 should be enough
#define MAXCONNECT 64

struct MessageInfo {
	unsigned dword1, dword2;
	unsigned short word1, word2, word3, word4, word5, word6;
	unsigned char byte1, byte2;
	const char *buf;
} ;

/*#define CI_MAXSTRING 500

struct ClickInfo {
	// on connection close, reset this
	unsigned char type, byte1, byte2;
	unsigned short word1, word2;
	unsigned dword1, dword2;
	char name[STRSIZE];
	char string[CI_MAXSTRING];
} ;*/

struct PooledItem {
	int id;
	short count;
} ;

struct connection {
	int acct;
	int npcid;
	SOCKET s;
	// ready means use the connection for normal operations
	// ready is not really necessary because of the crit sect
	// available means it can be used to set up a new conn.
	bool ready, available;
	int ecpos;
	char encrypt[64];
	//ClickInfo click;
	/*int pool; // currently held item
	short count; // currently held item count*/
	PooledItem pool;

	connection(void)
	{
		ready = false;
		available = true;
		//click.type = 0;
	}

	void close(int);
	void process(int);
	void dpaprocess(int);

	void setready(bool iready)
	{
		ready = iready;
	}

	int GetBytes(char *buf, int count)
	{
		int size;

		size = ReceiveBytes(s, buf, count);

		if(size > 0)
			ClientDecode(buf, size, ecpos, encrypt);

		return size;
	}
} ;

class connectlist {
	connection connects[MAXCONNECT];
	//HANDLE events[MAXCONNECT];

public:
	connectlist(void)
	{
		/*int i;

		for(i = 0; i < MAXCONNECT; i++) {
			events[i] = WSACreateEvent();
		}*/
	}

	~connectlist(void)
	{
		/*int i;

		for(i = 0; i < MAXCONNECT; i++) {
			WSACloseEvent(events[i]);
		}*/
	}

	void process(void)
	{
		FD_SET read;
		int i;
		timeval stime;

		stime.tv_sec = 0;
		stime.tv_usec = 250000;

		FD_ZERO(&read);

		for(i = 0; i < MAXCONNECT; i++) {
			if(connects[i].ready)
				FD_SET(connects[i].s, &read);
		}

		select(NULL, &read, NULL, NULL, &stime);

		for(i = 0; i < MAXCONNECT; i++) {
			if(connects[i].ready && FD_ISSET(connects[i].s, &read) != 0) {
				if(GetDPAInput())
					connects[i].dpaprocess(i);
				else
					connects[i].process(i);
			}
		}

		//DWORD event;

		/*event = WSAWaitForMultipleEvents(MAXCONNECT, events, FALSE, WSA_INFINITE, FALSE) - WSA_WAIT_EVENT_0;

		if(event >= 0 && event < MAXCONNECT) {
			connects[event].process();
		}*/
	}

	int add(SOCKET, int, unsigned, int, const char *, const char *);

	// delete property prop
	// sort of academic since properties cannot
	// be deleted at run-time
	void close(int id)
	{
		if(id >= 0 && id < MAXCONNECT) {
			//WSAEventSelect(connects[id].s, events[id], 0);
			connects[id].close(id);
		}
	}

	void senddata(int id, char *buf, int size)
	{
		if(id >= 0 && id < MAXCONNECT) {
			if(connects[id].ready) {
				if(SendBytes(connects[id].s, buf, size) <= 0)
					connects[id].close(id);
			}
		}
	}

	void globalcast(int id, char *buf, int size)
	{
		int i;

		for(i = 0; i < MAXCONNECT; i++) {
			if(i != id) {
				senddata(i, buf, size);
			}
		}
	}

	//void broadcast(fieldlist &, int, char *, int);

	void ready(int id, bool iready)
	{
		if(id >= 0 && id < MAXCONNECT) {
			connects[id].setready(iready);
		}
	}

	void pool(int id, int itemid, short count)
	{
		if(id >= 0 && id < MAXCONNECT) {
			connects[id].pool.id = itemid;
			connects[id].pool.count = count;
		}
	}

	void unpool(int id)
	{
		if(id >= 0 && id < MAXCONNECT) {
			connects[id].pool.id = 0;
		}
	}

	int poolid(int id)
	{
		if(id >= 0 && id < MAXCONNECT) {
			return connects[id].pool.id;
		}
		else {
			return 0;
		}
	}

	short poolcount(int id)
	{
		if(id >= 0 && id < MAXCONNECT) {
			return connects[id].pool.count;
		}
		else {
			return 0;
		}
	}

	int npcid(int id)
	{
		if(id >= 0 && id < MAXCONNECT) {
			return connects[id].npcid;
		}
		else {
			return 0;
		}
	}
} ;

//void DisplayNPC(int, char *, int &); // dll
//void DisplayItem(int, char *, int &); // dll
int addconnection(SOCKET, int, unsigned, int, const char *, const char *);
void processconnections(void);
void readyconnection(int, bool);
//void PrepareMessage(unsigned, char *, int &, const MessageInfo *);
//void Broadcast(int, int, char *, int);
//void BroadcastList(fieldlist &, int, char *, int);
//int glight(void); // dll
//void glight(int); // dll
