#include "pcheader.h"

// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

#include "server.h"

#include "resource.h"

#define DefProc DefWindowProc
#define SERVNAMESIZE 32
#define MAPFILESIZE 500

// network addressing stuff
#define TCP_LOGINPORT 4500
#define TCP_GAMEPORT 5000
#define UDP_TESTPORT 12000
//#define TCP_GAMEPORT 50000
// automate the address
//#define HOST_ADDR "24.3.23.30"

#define MAXLOGIN 8

struct LoginInfo {
	bool logged;
	int acct;
	SOCKET s;
	int ecpos;
	char encrypt[64];

	LoginInfo(void)
	{
		logged = false;
	}
} ;

LoginInfo logins[MAXLOGIN];

int serr, servaddr = 0;
char servname[SERVNAMESIZE];
char mapfile[MAPFILESIZE], tilefile[MAPFILESIZE], staidxfile[MAPFILESIZE], staticsfile[MAPFILESIZE], verfile[MAPFILESIZE], multiidxfile[MAPFILESIZE], multifile[MAPFILESIZE];
const HOSTENT *che;

// minimum number of milliseconds between the starts of consecutive cycles
#define GAME_CYCLE_TIME 250
// minimum number of save cycles per game cycle
#define SAVE_CYCLE_MIN 5

int APIENTRY WinMain(HINSTANCE, HINSTANCE, LPSTR, int);
/*LRESULT APIENTRY WndProc(HWND, UINT, WPARAM, LPARAM);
BOOL OnDestroy(HWND);
void OnCommand(HWND, int, HWND, UINT);*/
BOOL CALLBACK DlgMainProc(HWND, UINT, WPARAM, LPARAM);
BOOL OnInitDialogMain(HWND, HWND, LPARAM);
void OnCommandMain(HWND, int, HWND, UINT);
BOOL CALLBACK DlgStartupProc(HWND, UINT, WPARAM, LPARAM);
BOOL OnInitDialogStartup(HWND, HWND, LPARAM);
void OnCommandStartup(HWND, int, HWND, UINT);
BOOL CALLBACK DlgAboutProc(HWND, UINT, WPARAM, LPARAM);
BOOL OnInitDialogAbout(HWND, HWND, LPARAM);
void OnCommandAbout(HWND, int, HWND, UINT);
BOOL CALLBACK DlgSaveIntervalProc(HWND, UINT, WPARAM, LPARAM);
BOOL OnInitDialogSaveInterval(HWND, HWND, LPARAM);
void OnCommandSaveInterval(HWND, int, HWND, UINT);
/*BOOL CALLBACK DlgStatusProc(HWND, UINT, WPARAM, LPARAM);
BOOL OnInitDialogStatus(HWND, HWND, LPARAM);
void OnCommandStatus(HWND, int, HWND, UINT);*/

void CreateThreadSyncObjects(void);
void CloseThreadSyncObjects(void);

DWORD WINAPI LoginThread(LPVOID);
DWORD WINAPI GameLoginThread(LPVOID);
DWORD WINAPI ConnectionThread(LPVOID);
DWORD WINAPI GameCycleThread(LPVOID);
DWORD WINAPI SaveCycleThread(LPVOID);
DWORD WINAPI ConnTestThread(LPVOID);

// this critical section is for the game data (and savemin)
CRITICAL_SECTION gamedatacsect;

int savemin = SAVE_CYCLE_MIN;
int saveinterval; // specified in minutes
//UINT SaveTimer;
LARGE_INTEGER savestart;

bool logincheck;

void SaveBegun(void);
void SaveComplete(void);
void DisplaySystemTime(void);

/*void CALLBACK SaveTimerProc(HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime)
{
	KillTimer(NULL, SaveTimer);

	StartSaveNow();
}*/

// functions
int login(/*SOCKET, sockaddr_in **//*LoginInfo &*/int);
int gamelogin(SOCKET, sockaddr_in *);
void loginresponse(SOCKET, int);
void redirectresponse(SOCKET, int);
void gameloginresponse(SOCKET, int);

// done!
bool done = false;

// decryption stuff
// simplify later
int stuff1[] = { 8,  9, 10, 11, 12, 13, 14, 15,
                16,  1,  2,  3,  4,  5,  6,  7,
				 8,  9, 10, 11, 12, 13, 14, 15,
				16, 17, 18, 19, 20, 21, 22, 23,
				24, 25, 26, 27, 28, 29, 30, 31,
				32, 17, 18, 19, 20, 21, 22, 23,
				24, 25, 26, 27, 28, 29, 30, 31,
                32,  1,  2,  3,  4,  5,  6,  7} ;

int stuff2[] = {32, 17, 18, 19, 20, 21, 22, 23,
                24, 25, 26, 27, 28, 29, 30, 31,
				32, 17, 18, 19, 20, 21, 22, 23,
				24,  9, 10, 11, 12, 13, 14, 15,
				16,  1,  2,  3,  4,  5,  6,  7,
				 8,  9, 10, 11, 12, 13, 14, 15,
				16,  1,  2,  3,  4,  5,  6,  7,
                 8, 25, 26, 27, 28, 29, 30, 31} ;

// key for 1.25.30
/*int stuff3[] = {1, 0, 1, 0, 1, 0, 1, 0,
				0, 1, 1, 0, 0, 0, 1, 1,
				1, 0, 0, 0, 0, 0, 1, 0,
				0, 0, 0, 1, 0, 0, 1, 1,
				0, 0, 0, 0, 0, 1, 0, 0,
				0, 0, 0, 0, 1, 0, 0, 0,
				1, 1, 1, 1, 1, 0, 0, 0,
				1, 0, 0, 0, 1, 1, 0, 1} ;*/

// key for 1.25.31
/*int stuff3[] = {0, 1, 1, 0, 0, 0, 0, 1,
				0, 1, 1, 0, 1, 0, 1, 1,
				0, 0, 0, 0, 1, 1, 0, 0,
				0, 1, 0, 1, 1, 0, 0, 1,
				1, 0, 1, 1, 0, 0, 0, 0,
				0, 1, 0, 0, 0, 1, 1, 1,
				0, 1, 1, 1, 0, 0, 0, 1,
				0, 1, 0, 0, 1, 1, 1, 0} ;*/

// key for 1.25.32
/*int stuff3[] = {0, 1, 1, 1, 1, 0, 1, 0,
				1, 1, 0, 1, 0, 1, 0, 0,
				1, 0, 0, 0, 1, 1, 0, 0,
				0, 1, 0, 1, 1, 1, 1, 0,
				0, 1, 1, 1, 0, 0, 0, 0,
				1, 0, 1, 1, 0, 1, 1, 0,
				1, 0, 1, 1, 0, 0, 0, 1,
				0, 1, 0, 0, 1, 1, 1, 0} ;*/

// key for 1.25.33
/*int stuff3[] = {0, 1, 1, 0, 0, 0, 0, 0,
				1, 0, 1, 0, 1, 0, 0, 0,
				1, 0, 0, 0, 1, 1, 0, 0,
				0, 1, 0, 1, 1, 0, 1, 0,
				0, 1, 0, 1, 0, 0, 0, 0,
				0, 0, 1, 1, 1, 0, 0, 0,
				1, 0, 1, 1, 0, 0, 0, 1,
				0, 1, 0, 0, 1, 1, 1, 0} ;*/

// key for 1.25.34
//              l  l  l  l  l  l  l  l
/*int stuff3[] = {0, 1, 1, 1, 1, 1, 0, 1,
//              l  l  l  l  l  l  l  l
				1, 1, 1, 0, 1, 0, 1, 0,
//              l  l  l  l  l  l  l  l
				1, 0, 0, 0, 1, 1, 0, 0,
//              l  l  l  l  l  l  p  p
				0, 1, 0, 1, 1, 1, 0, 0,
//              p  p  p  p  p  p  p  p
				0, 1, 1, 1, 1, 0, 0, 0,
//              p  p  p  p  p  p  p  p
				1, 1, 1, 1, 0, 0, 0, 1,
//              p  p  p  p  p  p  p  p
				1, 0, 1, 1, 0, 0, 0, 1,
//              p  p  p  p  ?  ?  ?  ?
				0, 1, 0, 0, 1, 1, 1, 0} ;*/

// key for 1.25.35 (1653181528-1417425230)
//              l  l  l  l  l  l  l  l
/*int stuff3[] = {0, 1, 1, 0, 0, 0, 1, 0, // 62
//              l  l  l  l  l  l  l  l
				1, 0, 0, 0, 1, 0, 0, 1, // 89
//              l  l  l  l  l  l  l  l
				1, 0, 0, 0, 1, 1, 0, 0, // 8c
//              l  l  l  l  l  l  p  p
				0, 1, 0, 1, 1, 0, 0, 0, // 58
//              p  p  p  p  p  p  p  p
				0, 1, 0, 1, 0, 1, 0, 0, // 54
//              p  p  p  p  p  p  p  p
				0, 1, 1, 1, 1, 1, 0, 0, // 7c
//              p  p  p  p  p  p  p  p
				0, 0, 1, 1, 0, 0, 0, 1, // 31
//              p  p  p  p  ?  ?  ?  ?
				0, 1, 0, 0, 1, 1, 1, 0} ; // 4e*/

// UGLY HACK!!!
unsigned long ck1, ck2;

HINSTANCE hThisInstance = NULL;
HWND hwndMain = NULL;

struct Locale {
	char name1[31], name2[31];
	int x, y, z;

	Locale(char *str1, char *str2, int x1, int y1, int z1 = -1)
	{
		CopyBuf(name1, 31, str1);
		CopyBuf(name2, 31, str2);
		x = x1;
		y = y1;
		z = z1;
	}
} ;

Locale **locales = NULL;
int localecount = 0;

void GetLocales(const char *filename);

struct serverdef {
	char name[32], zone;
	int addr, testaddr;
	short port;

	serverdef(char *str, int addr1, int p, int addr2, int z)
	{
		CopyBuf(name, 32, str);
		addr = addr1;
		port = p;
		testaddr = addr2;
		zone = z;
	}
} ;

serverdef **servers = NULL;
int servercount = 0;

void GetServers(const char *filename);

HANDLE SaveThread = NULL;

FUSELIBFNDEF int FUSEInterfaceVersion(void)
{
	return 0x00010000;
}

CYCLE cycleTemp;
CREATEPLAYER CreatePlayerTemp;
GETIDPOINT GetIDPointTemp;
OBJECTLOADED NPCLoadedTemp;
OBJECTLOADED ItemLoadedTemp;
GETPCNAMEANDPASS GetPCNameAndPassTemp;
PCLOG PCLoginTemp;
PCLOG PCLogoutTemp;
PROCESSINPUT DPAProcessInputTemp;
PROCESSINPUT ProcessInputTemp;
DELETEPC DeletePCTemp;

void cycle(void)
{
	cycleTemp();
}

void CreatePlayer(int id, const char *data)
{
	CreatePlayerTemp(id, data);
}

void GetIDPoint(int id, posinfo &p)
{
	GetIDPointTemp(id, p);
}

void NPCLoaded(int index)
{
	NPCLoadedTemp(index);
}

void ItemLoaded(int index)
{
	ItemLoadedTemp(index);
}

void GetPCNameAndPass(int id, char *name, char *pass)
{
	GetPCNameAndPassTemp(id, name, pass);
}

void PCLogin(int index)
{
	PCLoginTemp(index);
}

void PCLogout(int index)
{
	PCLogoutTemp(index);
}

void DPAProcessInput(int index, unsigned short size, const char *buf)
{
	DPAProcessInputTemp(index, size, buf);
}

void ProcessInput(int index, unsigned short size, const char *buf)
{
	ProcessInputTemp(index, size, buf);
}

int DeletePC(int acct, int npcid)
{
	return DeletePCTemp(acct, npcid);
}

int APIENTRY WinMain(HINSTANCE hCurrentInst, HINSTANCE, LPSTR, int)
{
	/*WNDCLASS wndclass;
	HWND hwnd;
	MSG msg;*/

	WSADATA wsaData;
	HANDLE hThreads[6];
	DWORD dwThreadID[6]/*, dwObject*/;
	FILE *sfil;
	HANDLE hmutex;

	hmutex = CreateMutex(NULL, FALSE, "FUSE Instance Mutex");

	if(GetLastError() == ERROR_ALREADY_EXISTS) {
		MessageBox(NULL/*hwndMain*/, "Another instance of FUSE is already running.", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		exit(1);
	}

	// DLL Test I
	HINSTANCE dlib;
	dlib = LoadLibrary("fusegame.dll");
	if(dlib == NULL)
		MessageBox(NULL, "DLL not loaded", "EXE Test Error", MB_OK | MB_ICONERROR);
	cycleTemp = (CYCLE)GetProcAddress(dlib, "cycle");
	if(cycleTemp == NULL)
		MessageBox(NULL, "cycle function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	CreatePlayerTemp = (CREATEPLAYER)GetProcAddress(dlib, "CreatePlayer");
	if(CreatePlayerTemp == NULL)
		MessageBox(NULL, "CreatePlayer function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	GetIDPointTemp = (GETIDPOINT)GetProcAddress(dlib, "GetIDPoint");
	if(GetIDPointTemp == NULL)
		MessageBox(NULL, "GetIDPoint function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	NPCLoadedTemp = (OBJECTLOADED)GetProcAddress(dlib, "NPCLoaded");
	if(NPCLoadedTemp == NULL)
		MessageBox(NULL, "NPCLoaded function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	ItemLoadedTemp = (OBJECTLOADED)GetProcAddress(dlib, "ItemLoaded");
	if(ItemLoadedTemp == NULL)
		MessageBox(NULL, "ItemLoaded function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	GetPCNameAndPassTemp = (GETPCNAMEANDPASS)GetProcAddress(dlib, "GetPCNameAndPass");
	if(GetPCNameAndPassTemp == NULL)
		MessageBox(NULL, "GetPCNameAndPass function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	PCLoginTemp = (PCLOG)GetProcAddress(dlib, "PCLogin");
	if(PCLoginTemp == NULL)
		MessageBox(NULL, "PCLogin function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	PCLogoutTemp = (PCLOG)GetProcAddress(dlib, "PCLogout");
	if(PCLogoutTemp == NULL)
		MessageBox(NULL, "PCLogout function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	DPAProcessInputTemp = (PROCESSINPUT)GetProcAddress(dlib, "DPAProcessInput");
	if(DPAProcessInputTemp == NULL)
		MessageBox(NULL, "DPAProcessInput function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	ProcessInputTemp = (PROCESSINPUT)GetProcAddress(dlib, "ProcessInput");
	if(ProcessInputTemp == NULL)
		MessageBox(NULL, "ProcessInput function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	DeletePCTemp = (DELETEPC)GetProcAddress(dlib, "DeletePC");
	if(DeletePC == NULL)
		MessageBox(NULL, "DeletePC function not found!", "EXE Test Error", MB_OK | MB_ICONERROR);
	//FreeLibrary(dlib);
	// End DLL Test

	// temporary prop data (sadly, it sticks around since it's defined inside main)
	proptemplate iprops, pprops;

	hThisInstance = hCurrentInst;

	InitCommonControls();

	/*wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hCurrentInst;
	wndclass.hIcon = LoadIcon(hThisInstance, MAKEINTRESOURCE(IDI_BOMB2));
	wndclass.hCursor =  LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName = MAKEINTRESOURCE(IDR_MENU);
	wndclass.lpszClassName = "FUSE";

	RegisterClass(&wndclass);

	UINT width = GetSystemMetrics(SM_CXSCREEN);
	UINT height = GetSystemMetrics(SM_CYSCREEN);

	hwndMain = hwnd = CreateWindowEx(
		0,
		"FUSE",
		"FUSE",
		WS_OVERLAPPEDWINDOW,
		width / 4,
		height / 4,
		width / 2,
		height / 2,
		NULL,
		NULL,
		hCurrentInst,
		NULL);

	ShowWindow(hwnd, SW_SHOWDEFAULT);*/

	if((WSAStartup(MAKEWORD(1, 1), &wsaData)) != 0) {
		//printf("Could not initialize Windows Sockets.\n");
		MessageBox(NULL/*hwndMain*/, "Error initializing Windows Sockets 1.1", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		exit(1);
	}

	//HOSTENT *he;
	//int n, addr;
	char buf[500];

	if(gethostname(buf, 500) == SOCKET_ERROR) {
		MessageBox(NULL/*hwndMain*/, "Cannot retrieve host name", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		exit(1);
	}

	che = gethostbyname(buf);

	/*n = 0;

	while(he -> h_addr_list[n] != NULL) {
		addr = ntohl(*((int *)(he -> h_addr_list[n])));
		n += 1;
	}*/

	int n;

	if((n = DialogBox(hThisInstance, MAKEINTRESOURCE(IDD_STARTUP), NULL/*hwndMain*/, (DLGPROC)DlgStartupProc)) == -1) {
		MessageBox(NULL/*hwndMain*/, "Cannot process startup data", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		exit(1);
	}

	else if(n != 0) {
		MessageBox(NULL/*hwndMain*/, "Server Startup Cancelled", "FUSE Startup", MB_OK | MB_ICONERROR);
		exit(1);
	}

	// load program script (must be present for execution)
	// the script should contain all info concerning allocation unit size, etc

	// this corresponds to script template info
	inittemplates();

	sfil = fopen("save.fus", "rb");

	if(sfil != NULL) {
		// load accounts
		if(!LoadAccounts(sfil)) {
			//printf("Cannot load account info: invalid version\n");
			MessageBox(NULL/*hwndMain*/, "Error loading account info: invalid version", "FUSE Startup Error", MB_OK | MB_ICONERROR);
			exit(1);
		}

		// load game template data
		if(!LoadGameTemplateInfo(sfil, pprops, iprops)) {
			//printf("Error loading game template info\n");
			MessageBox(NULL/*hwndMain*/, "Error loading game template info", "FUSE Startup Error", MB_OK | MB_ICONERROR);
			exit(1);
		}

		if(!TemplateMatch(pprops, iprops)) {
			//printf("Mismatched script/save templates.  Run FUSEDATA\n");
			MessageBox(NULL/*hwndMain*/, "Error loading game info: mismatched script/save templates", "FUSE Startup Error", MB_OK | MB_ICONERROR);
			exit(1);
		}

		// load game data
		if(!LoadGameData(sfil)) {
			//printf("Error loading game data\n");
			MessageBox(NULL/*hwndMain*/, "Error loading game data", "FUSE Startup Error", MB_OK | MB_ICONERROR);
			exit(1);
		}

		fclose(sfil);
	}

	// this modifies current account info
	GetAccounts("accounts.txt");

	GetLocales("locales.txt");
	GetServers("servers.txt");

	CreateThreadSyncObjects();

	if(!MapSetup(mapfile, tilefile, staidxfile, staticsfile, verfile, multiidxfile, multifile)) {
		//printf("Couldn't open map file\n");
		MessageBox(NULL/*hwndMain*/, "Map file not found", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		exit(1);
	}

	hThreads[0] = CreateThread(NULL, 2000, LoginThread, NULL, 0, dwThreadID);
	//hThreads[5] = CreateThread(NULL, 2000, GameLoginThread, NULL, 0, dwThreadID + 5);
	//hThreads[5] = NULL;
	hThreads[1] = CreateThread(NULL, 2000, ConnectionThread, NULL, 0, dwThreadID + 1);
	hThreads[2] = CreateThread(NULL, 2000, GameCycleThread, NULL, 0, dwThreadID + 2);
	SaveThread = hThreads[3] = CreateThread(NULL, 2000, SaveCycleThread, NULL, 0, dwThreadID + 3);
	hThreads[4] = CreateThread(NULL, 2000, ConnTestThread, NULL, 0, dwThreadID + 4);

	DialogBox(hThisInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)DlgMainProc);

	done = true; // perhaps use interlockedincrement here, although I don't think we need to

	//MsgWaitForMultipleObjects(5, hThreads, FALSE, INFINITE, QS_ALLINPUT);

	/*DialogBox(hThisInstance, MAKEINTRESOURCE(IDD_STATUS), hwndMain, (DLGPROC)DlgStatusProc);

	while(!done) {
		dwObject = MsgWaitForMultipleObjects(5, hThreads, FALSE, INFINITE, QS_ALLINPUT);

		if(dwObject == WAIT_OBJECT_0 + 5) {
			// message in queue
			while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) != 0) {
				if(msg.message == WM_QUIT) {
					done = true;
				}
				else {
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
		}
		else {
			// unknown input type
			done = true;
		}
	}*/

	ResumeThread(SaveThread);

	WaitForMultipleObjects(5, hThreads, TRUE, INFINITE);

	CloseHandle(hThreads[0]);
	//CloseHandle(hThreads[5]);
	CloseHandle(hThreads[1]);
	CloseHandle(hThreads[2]);
	CloseHandle(hThreads[3]);
	CloseHandle(hThreads[4]);

	MapRelease();
	CloseThreadSyncObjects();

	WSACleanup();

	FreeLibrary(dlib);

	CloseHandle(hmutex);

	return 0; //msg.wParam;
}

BOOL CALLBACK DlgMainProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	int delay;

	switch(msg) {
		HANDLE_MSG(hwnd, WM_INITDIALOG, OnInitDialogMain);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommandMain);

	case WM_TIMER:
		if(wParam == 1) {
			DisplaySystemTime();
		}
		else if(wParam == 2) {
			KillTimer(NULL, 2);
			StartSaveNow();
		}

		break;

	case WM_USER:
		if(wParam == IDI_BOMB2) {
			if(lParam == WM_LBUTTONDOWN || lParam == WM_LBUTTONDBLCLK) {
				ShowWindow(hwnd, SW_SHOW);
			}
		}

		break;

	case (WM_USER + 1):
		delay = (int)lParam;

		if(delay > 0) {
			SuspendThread(SaveThread);
			Static_SetText(GetDlgItem(hwnd, IDC_SAVESTATUS), "Idle");
			SetTimer(hwnd, 2, delay, NULL);
		}
		else if(delay < 0) {
			SuspendThread(SaveThread);
			Static_SetText(GetDlgItem(hwnd, IDC_SAVESTATUS), "Idle");
		}

		break;
	}

	return FALSE;
}

BOOL OnInitDialogMain(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
	in_addr sa;
	char buf[100];
	HICON hIcon;
	NOTIFYICONDATA nid;

	// as it currently stands, the threads are started before the window
	// is, so some messagebox routines might have a null parent window
	hwndMain = hwnd;

	DisplaySystemTime();
	// update system time every second
	SetTimer(hwnd, 1, 1000, NULL);

	if(saveinterval < 0) {
		Edit_SetText(GetDlgItem(hwnd, IDC_SAVEINTERVAL), "Never");
	}
	else if(saveinterval == 0) {
		Edit_SetText(GetDlgItem(hwnd, IDC_SAVEINTERVAL), "Continual");
	}
	else if(saveinterval == 1) {
		Edit_SetText(GetDlgItem(hwnd, IDC_SAVEINTERVAL), "1 min.");
	}
	else {
		sprintf(buf, "%d mins.", saveinterval);
		Edit_SetText(GetDlgItem(hwnd, IDC_SAVEINTERVAL), buf);
	}

	if(!logincheck)
		Button_SetCheck(GetDlgItem(hwnd, IDC_LOGINCHECK), BST_UNCHECKED);
	else
		Button_SetCheck(GetDlgItem(hwnd, IDC_LOGINCHECK), BST_CHECKED);

	// setup server address
	sa.s_addr = servaddr;
	sprintf(buf, "%s,%d", inet_ntoa(sa), TCP_LOGINPORT);
	Edit_SetText(GetDlgItem(hwnd, IDC_SERVADDR), buf);

	SetFocus(GetDlgItem(hwnd, IDC_SERVADDR));
	Edit_SetSel(GetDlgItem(hwnd, IDC_SERVADDR), 0, strlen(inet_ntoa(sa)));

	// get icon handle
	hIcon = LoadIcon(hThisInstance, MAKEINTRESOURCE(IDI_BOMB2));

	// setup dialog icon
	SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LONG)hIcon);
	SendMessage(hwnd, WM_SETICON, ICON_BIG, (LONG)hIcon);

	// setup taskbar icon
	nid.cbSize = sizeof(nid);
	nid.hWnd = hwnd;
	nid.uID = IDI_BOMB2;
	nid.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
	nid.uCallbackMessage = WM_USER; // user message for now
	nid.hIcon = hIcon;
	lstrcpyn(nid.szTip, "FUSE Control", sizeof(nid.szTip));
	Shell_NotifyIcon(NIM_ADD, &nid);

	//SetFocus(GetDlgItem(hwnd, IDCLOSE));
	return FALSE;
}

// id - id of menu item, control, or accelerator
// hwndCtl - identifies control sending message (if not control, this is NULL)
// codeNotify - notification code from control (if menu, 0; if accelerator, 1)
void OnCommandMain(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	in_addr sa;
	NOTIFYICONDATA nid;
	int n;

	switch(id) {
	case IDC_CLOSE:
	case IDCANCEL:
		ShowWindow(hwnd, SW_HIDE);
		break;

	case IDC_QUIT:
		n = MessageBox(hwndMain, "Are you sure you want to shutdown the server?", "FUSE", MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2);

		if(n == IDYES) {
			nid.cbSize = sizeof(nid);
			nid.hWnd = hwnd;
			nid.uID = IDI_BOMB2;
			Shell_NotifyIcon(NIM_DELETE, &nid);

			KillTimer(hwnd, 1);

			EndDialog(hwnd, TRUE);
		}

		break;

	case IDC_ABOUT:
		DialogBox(hThisInstance, MAKEINTRESOURCE(IDD_ABOUT), hwnd, (DLGPROC)DlgAboutProc);
		break;

	case IDC_SETSI:
		DialogBox(hThisInstance, MAKEINTRESOURCE(IDD_SAVEINTERVAL), hwndMain, (DLGPROC)DlgSaveIntervalProc);
		break;

	case IDC_CLIENT:
		// write out the current server settings for fuserun
		sa.s_addr = servaddr;
		WritePrivateProfileString("FuseRun", "ServerIP", inet_ntoa(sa), "fuse.ini");
		WritePrivateProfileString("FuseRun", "ServerPort", "4500", "fuse.ini");

		// run fuserun
		ZeroMemory(&si, sizeof(si));
		si.cb = sizeof(si);

		if(CreateProcess("fuserun.exe", NULL, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi) == 0) {
			// change this to provide help, instead (with the messagebox help button)
			MessageBox(hwndMain, "Could not start fuserun.exe. Make sure it is in the FUSE directory.", "FUSERUN Error", MB_OK | MB_ICONERROR);
		}
		else {
			//MessageBox(hwndMain, "Starting the client from FUSE may take a while.", "FUSERUN", MB_OK | MB_ICONINFORMATION);
		}

		//ShowWindow(hwnd, SW_HIDE);

		break;

	case IDC_SAVENOW:
		StartSaveNow();
		break;

	case IDC_LOGINCHECK:
		if(Button_GetCheck(GetDlgItem(hwnd, IDC_LOGINCHECK)) == BST_UNCHECKED) {
			logincheck = false;
			WritePrivateProfileString("Fuse", "LoginCheck", "false", "fuse.ini");
		}
		else {
			logincheck = true;
			WritePrivateProfileString("Fuse", "LoginCheck", "true", "fuse.ini");
		}

		break;

	/*case IDC_STATUS:
		DialogBox(hThisInstance, MAKEINTRESOURCE(IDD_STATUS), hwnd, (DLGPROC)DlgStatusProc);
		break;*/
	}
}

FUSELIBFNDEF void SetSaveInterval(int interval)
{
	char buf[1000];

	// force save now if we weren't saving before
	if(saveinterval < 0 && interval >= 0)
		StartSaveNow();

	saveinterval = interval;

	sprintf(buf, "%d", saveinterval);
	WritePrivateProfileString("Fuse", "SaveInterval", buf, "fuse.ini");

	if(saveinterval < 0) {
		Edit_SetText(GetDlgItem(hwndMain, IDC_SAVEINTERVAL), "Never");
	}
	else if(saveinterval == 0) {
		Edit_SetText(GetDlgItem(hwndMain, IDC_SAVEINTERVAL), "Continual");
	}
	else if(saveinterval == 1) {
		Edit_SetText(GetDlgItem(hwndMain, IDC_SAVEINTERVAL), "1 min.");
	}
	else {
		sprintf(buf, "%d mins.", saveinterval);
		Edit_SetText(GetDlgItem(hwndMain, IDC_SAVEINTERVAL), buf);
	}
}

FUSELIBFNDEF void StartSaveNow(void)
{
	ResumeThread(SaveThread);
	Static_SetText(GetDlgItem(hwndMain, IDC_SAVESTATUS), "In Progress");
}

void SaveBegun(void)
{
	SYSTEMTIME st;
	FILETIME ft;

	// save current time
	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &ft);
	memcpy(&savestart, &ft, sizeof(LARGE_INTEGER));
}

void SaveComplete(void)
{
	SYSTEMTIME st;
	FILETIME ft;
	LARGE_INTEGER li, dest;
	int delay;
	char buf[1000];

	GetLocalTime(&st);

	if(st.wHour < 12) {
		if(st.wHour == 0) {
			sprintf(buf, "12:%02u AM", st.wMinute);
		}
		else {
			sprintf(buf, "%u:%02u AM", st.wHour, st.wMinute);
		}
	}
	else {
		if(st.wHour == 12) {
			sprintf(buf, "12:%02u PM", st.wMinute);
		}
		else {
			sprintf(buf, "%u:%02u PM", st.wHour - 12, st.wMinute);
		}
	}

	Static_SetText(GetDlgItem(hwndMain, IDC_LASTSAVE), buf);

	if(saveinterval >= 0) {
		GetSystemTime(&st);
		SystemTimeToFileTime(&st, &ft);
		memcpy(&li, &ft, sizeof(LARGE_INTEGER));

		// compute save timeout
		dest.QuadPart = saveinterval * 60000; // minute units to millisecond units
		dest.QuadPart *= 10000; // convert millisecond units to 100 nanosecond units
		dest.QuadPart += savestart.QuadPart; // add start time to delay
		dest.QuadPart -= li.QuadPart; // subtract current time
		dest.QuadPart /= 10000; // convert 100 nanosecond units to millisecond units
		delay = (int)dest.QuadPart;

		if(delay > 0) {
			SendMessage(hwndMain, WM_USER + 1, 0, (LPARAM)delay);
			/*SaveTimer = SetTimer(NULL, NULL, delay, (TIMERPROC)SaveTimerProc);
			SuspendThread(SaveThread);*/
		}
		else {
			SendMessage(hwndMain, WM_USER + 1, 0, (LPARAM)0);
		}
	}
	else {
		SendMessage(hwndMain, WM_USER + 1, 0, (LPARAM)-1);
		//SuspendThread(SaveThread);
	}
}

/*LRESULT APIENTRY WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {
		HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);

	default:
		return DefProc(hwnd, msg, wParam, lParam);
	}
}

BOOL OnDestroy(HWND hwnd)
{
	PostQuitMessage(0);

	return TRUE;
}

// id - id of menu item, control, or accelerator
// hendCtl - identifies control sending message (if not control, this is NULL)
// codeNotify - notification code from control (if menu, 0; if accelerator, 1)
void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	// perhaps not necessary...
	switch(id) {
	case ID_FILEEXIT:
		DestroyWindow(hwnd);
		break;

	case ID_OPTIONSSTATUS:
		DialogBox(hThisInstance, MAKEINTRESOURCE(IDD_STATUS), hwndMain, (DLGPROC)DlgStatusProc);
		break;

	case ID_HELPABOUT:
		DialogBox(hThisInstance, MAKEINTRESOURCE(IDD_ABOUT), hwndMain, (DLGPROC)DlgAboutProc);
		break;
	}
}*/

void CreateThreadSyncObjects(void)
{
	InitializeCriticalSection(&gamedatacsect);
}

void CloseThreadSyncObjects(void)
{
	DeleteCriticalSection(&gamedatacsect);
}

DWORD WINAPI ConnTestThread(LPVOID lpVoid)
{
	SOCKET sock;
	sockaddr_in sa, ca;
	int size = sizeof(sockaddr_in), count;
	char buf[1000];
	FD_SET read;
	timeval stime;

	stime.tv_sec = 0;
	stime.tv_usec = 250000;

	sock = socket(AF_INET, SOCK_DGRAM, 0);

	if(sock == INVALID_SOCKET) {
		//printf("Cannot create udp socket!\n");
		MessageBox(hwndMain, "Error creating UDP test socket", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		serr = WSAGetLastError();
		exit(1);
	}
	else {
		//printf("Udp socket created.\n");
	}

	memset(&sa, 0, sizeof(sockaddr_in));
	sa.sin_family = AF_INET;
	sa.sin_port = htons(UDP_TESTPORT);
	sa.sin_addr.s_addr = servaddr; //inet_addr(HOST_ADDR);

	if(bind(sock, (sockaddr *)&sa, size) == SOCKET_ERROR) {
		//printf("Error binding to local address!\n");
		MessageBox(hwndMain, "Error binding test socket to local address", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		serr = WSAGetLastError();
		exit(1);
	}

	while(!done) {
		FD_ZERO(&read);

		FD_SET(sock, &read);

		select(NULL, &read, NULL, NULL, &stime);

		if(FD_ISSET(sock, &read) != 0) {
			size = sizeof(sockaddr_in);
			if((count = recvfrom(sock, buf, 1000, 0, (sockaddr *)&ca, &size)) != SOCKET_ERROR) {
				if(ca.sin_addr.s_addr != (unsigned long)servaddr /*inet_addr(HOST_ADDR)*/) {
					sendto(sock, buf, count, 0, (sockaddr *)&ca, size);
				}
			}
		}
	}

	closesocket(sock);

	ExitThread(0);

	return 0;
}

// requires access to game data (occasional, minor access)
DWORD WINAPI LoginThread(LPVOID lpVoid)
{
	SOCKET sock;
	FD_SET read;
	int i, j, k, index1, index2, bit1, bit2, bit3, val;
	unsigned long key;
	char encrypt[64];
	timeval stime;

	stime.tv_sec = 0;
	stime.tv_usec = 250000;

	sock = socket(AF_INET, SOCK_STREAM, 0);

	if(sock == INVALID_SOCKET) {
		//printf("Cannot create tcp socket!\n");
		MessageBox(hwndMain, "Error creating login socket", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		serr = WSAGetLastError();
		exit(1);
	}
	else {
		//printf("Tcp socket created.\n");
	}

	sockaddr_in sa, ca;
	int size = sizeof(sockaddr_in);
	SOCKET nsock;

	memset(&sa, 0, sizeof(sockaddr_in));
	sa.sin_family = AF_INET;
	sa.sin_port = htons(TCP_LOGINPORT);
	sa.sin_addr.s_addr = servaddr; //inet_addr(HOST_ADDR);

	if(bind(sock, (sockaddr *)&sa, size) == SOCKET_ERROR) {
		//printf("Error binding to local address!\n");
		MessageBox(hwndMain, "Error binding login socket to local address", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		serr = WSAGetLastError();
		exit(1);
	}

	if(listen(sock, 5) == SOCKET_ERROR) {
		//printf("Error listening for client login!\n");
		MessageBox(hwndMain, "Error listening on login socket", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		serr = WSAGetLastError();
		exit(1);
	}

	while(!done) {
		FD_ZERO(&read);

		for(i = 0; i < MAXLOGIN; i++) {
			if(logins[i].logged)
				FD_SET(logins[i].s, &read);
		}

		FD_SET(sock, &read);

		// maintain idle time for logins...
		// put time limit on select

		select(NULL, &read, NULL, NULL, &stime);

		for(i = 0; i < MAXLOGIN; i++) {
			if(logins[i].logged && FD_ISSET(logins[i].s, &read) != 0) {
				if(login(i) != 0) {
					// the socket was closed or passed, so free it
					logins[i].logged = false;
				}
			}
		}

		if(FD_ISSET(sock, &read) != 0) {
			for(i = 0; i < MAXLOGIN; i++) {
				if(!logins[i].logged) {
					size = sizeof(sockaddr_in);

					if((nsock = accept(sock, (sockaddr *)&ca, &size)) == INVALID_SOCKET) {
						serr = WSAGetLastError();
						//printf("Error accepting client login: %d!\n", serr);
					}
					else {
						// read encryption key
						if((size = ReceiveBytes(nsock, &key, 4)) == SOCKET_ERROR) {
							//printf("Error receiving ip address!\n");
							serr = WSAGetLastError();
							closesocket(nsock);
						}
						else if(size == 0) {
							//printf("Login aborted by client (1).\n");
							closesocket(nsock);
						}
						else {
							logins[i].logged = true;
							logins[i].acct = -1;
							logins[i].s = nsock;
							logins[i].ecpos = 0;

							// attempt
							key = ntohl(key);
							for(j = 0; j < 64; j++) {
								val = 0;
								for(k = j - 7; k <= j; k++) {
									index1 = (k + 64) & 0x3f;
									index2 = (k + 64 + 1) & 0x3f;
									bit1 = (key >> (stuff1[index1] - 1)) & 1;
									bit2 = (key >> (stuff1[index2] - 1)) & 1;
									bit3 = (index1 < 32) ? ((ck1 >> (31 - index1)) & 1) : ((ck2 >> (63 - index1)) & 1); // stuff3[index1];
									val = (val >> 1) | ((bit1 ^ bit2 ^ bit3) << 7);
								}

								encrypt[j] = val;
							}

							val = (unsigned char)key ^ 0xaa;

							for(j = 0; j < 64; j++) {
								logins[i].encrypt[j] = val;
								val ^= (unsigned int)encrypt[(j + 64) & 0x3f];
							}

							//break; // this was misplaced!
						}
						/*if(login(nsock, &ca) != 0) {
							closesocket(nsock);
						}*/
					}

					break;
				}
			}
		}
	}

	closesocket(sock);

	ExitThread(0);

	return 0;
}
/*
// requires access to game data (occasional, minor access)
DWORD WINAPI GameLoginThread(LPVOID lpVoid)
{
	SOCKET sock;

	sock = socket(AF_INET, SOCK_STREAM, 0);

	if(sock == INVALID_SOCKET) {
		//printf("Cannot create tcp socket!\n");
		MessageBox(hwndMain, "Error creating login socket", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		serr = WSAGetLastError();
		exit(1);
	}
	else {
		//printf("Tcp socket created.\n");
	}

	// temp
	MessageBox(hwndMain, "Game socket created", "FUSE Info...", MB_OK | MB_ICONINFORMATION);

	sockaddr_in sa, ca;
	int size = sizeof(sockaddr_in);
	SOCKET nsock;

	memset(&sa, 0, sizeof(sockaddr_in));
	sa.sin_family = AF_INET;
	sa.sin_port = htons(TCP_GAMEPORT);
	sa.sin_addr.s_addr = servaddr; //inet_addr(HOST_ADDR);

	if(bind(sock, (sockaddr *)&sa, size) == SOCKET_ERROR) {
		//printf("Error binding to local address!\n");
		MessageBox(hwndMain, "Error binding login socket to local address", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		serr = WSAGetLastError();
		exit(1);
	}

	if(listen(sock, 5) == SOCKET_ERROR) {
		//printf("Error listening for client login!\n");
		MessageBox(hwndMain, "Error listening on login socket", "FUSE Startup Error", MB_OK | MB_ICONERROR);
		serr = WSAGetLastError();
		exit(1);
	}

	while(!done) {
		size = sizeof(sockaddr_in);

		if((nsock = accept(sock, (sockaddr *)&ca, &size)) == INVALID_SOCKET) {
			serr = WSAGetLastError();
			//printf("Error accepting client login: %d!\n", serr);
		}
		else {
			if(gamelogin(nsock, &ca) != 0) {
				closesocket(nsock);
			}
		}
	}

	closesocket(sock);

	ExitThread(0);

	return 0;
}
*/
// this thread would operate more smoothly if we could
// receive messages or preferably events on input
DWORD WINAPI ConnectionThread(LPVOID lpVoid)
{
	while(!done) {
		processconnections();
	}

	ExitThread(0);

	return 0;
}

// requires access to game data
// requires access to savemin
// this thread must NOT get interrupted while it is executing
DWORD WINAPI GameCycleThread(LPVOID lpVoid)
{
	SYSTEMTIME st;
	FILETIME ft;
	LARGE_INTEGER li, dest;

	while(!done) {
		// if there is an active quitting time, then
		// check the current time; if it is >= the
		// quitting time, then break out of the loop

		// enter critical section
		EnterCriticalSection(&gamedatacsect);

		// set savemin to minimum number
		// of save cycles per game cycle
		savemin = SAVE_CYCLE_MIN;

		// compute next cycle time
		GetSystemTime(&st);
		SystemTimeToFileTime(&st, &ft);
		memcpy(&li, &ft, sizeof(LARGE_INTEGER));
		// convert offset time from millisecond units to 100 nanosecond units
		dest.QuadPart = li.QuadPart + GAME_CYCLE_TIME * 10000;

		// perform cycle
		cycle();

		// exit critical section
		LeaveCriticalSection(&gamedatacsect);

		// if nexttime - thistime > 0, sleep for nexttime - thistime
		GetSystemTime(&st);
		SystemTimeToFileTime(&st, &ft);
		memcpy(&li, &ft, sizeof(LARGE_INTEGER));
		// convert relative time from 100 nanosecond units to millisecond units
		dest.QuadPart = (dest.QuadPart - li.QuadPart) / 10000;

		if(dest.QuadPart > 0) {
			Sleep((DWORD)dest.QuadPart);
		}
		else {
			Sleep(0);
		}
	}

	ExitThread(0);

	return 0;
}

// requires access to game data (mostly the game *backup* data)
// requires access to savemin
DWORD WINAPI SaveCycleThread(LPVOID lpVoid)
{
	FILE *sfile;
	int i, index, count = 0;
	backupdata bdata;
	bool backup = false;
	int itemcount, npccount, temp;
	char c;
	/*SYSTEMTIME st;
	FILETIME ft;
	LARGE_INTEGER li, dest, mult;*/

	/*// initial sleep
	mult.QuadPart = 60000; // minute units to millisecond units
	dest.QuadPart = saveinterval.QuadPart * mult.QuadPart;

	if(dest.QuadPart > 0) {
		Sleep((DWORD)dest.QuadPart);
	}
	else if(saveinterval.QuadPart == 0) {
		Sleep(INFINITE);
	}
	else {
		Sleep(0);
	}*/

	// tell the progress bar that no save is in progress
	SendMessage(GetDlgItem(hwndMain, IDC_SAVEPROGRESS), PBM_SETPOS, 0, 0);

	SaveBegun();
	SaveComplete();

	Static_SetText(GetDlgItem(hwndMain, IDC_SAVESTATUS), "In Progress");

	while(!done) {
		// enter critical section
		EnterCriticalSection(&gamedatacsect);

		for(i = 0; i < savemin; i++) {
			// perform a single save cycle
			getbackupdata(bdata);

			if(bdata.data == NULL) {
				// it should be time to start a new backup;
				// open backup file and initiate backup
				if(initbackup() == false) {
					//printf("Error initiating backup\n");
				}
				else {
					itemcount = backitemcount();
					npccount = backnpccount();
					count = 0;

					SaveBegun();

					if(itemcount + npccount > 0) {
						/*// compute next cycle time
						GetSystemTime(&st);
						SystemTimeToFileTime(&st, &ft);
						memcpy(&li, &ft, sizeof(LARGE_INTEGER));
						mult.QuadPart = 600000000; // minute units to 100 nanosecond units
						dest.QuadPart = li.QuadPart + saveinterval.QuadPart * mult.QuadPart;*/

						//SaveBegun();

						// prepare backup (open file, etc)
						sfile = fopen("save.bak", "wb");

						if(sfile == NULL) {
							//printf("Save Error: could not open backup file\n");
							MessageBox(hwndMain, "Error opening backup file", "FUSE Save Error", MB_OK | MB_ICONERROR);
							exit(1);
						}

						// write out accounts, templates, item/npc counts
						// later on, we can set it up to backup the accounts
						// incrementally before getting to the game data
						if(!SaveAccounts(sfile)) {
							//printf("Save Error: could not save account info\n");
							MessageBox(hwndMain, "Error saving account", "FUSE Save Error", MB_OK | MB_ICONERROR);
							exit(1);
						}

						if(!SaveGameTemplateInfo(sfile)) {
							//printf("Save Error: could not save game template info\n");
							MessageBox(hwndMain, "Error saving game template info", "FUSE Save Error", MB_OK | MB_ICONERROR);
							exit(1);
						}

						temp = htonl(npccount);

						if(fwrite(&temp, 4, 1, sfile) != 1) {
							//printf("Save Error: could not save game data\n");
							MessageBox(hwndMain, "Error saving game data", "FUSE Save Error", MB_OK | MB_ICONERROR);
							exit(1);
						}

						temp = htonl(itemcount);

						if(fwrite(&temp, 4, 1, sfile) != 1) {
							//printf("Save Error: could not save game data\n");
							MessageBox(hwndMain, "Error saving game data", "FUSE Save Error", MB_OK | MB_ICONERROR);
							exit(1);
						}

						// tell the progress bar a save has begun (0%)
						backup = true;
					}
				}
			}
			else {
				if(backup) {
					// complete next backup step
					if(!bdata.use) {
						c = 0;

						if(fwrite(&c, 1, 1, sfile) != 1) {
							//printf("Save Error: could not save game data\n");
							MessageBox(hwndMain, "Error saving game data", "FUSE Save Error", MB_OK | MB_ICONERROR);
							exit(1);
						}
					}
					else {
						c = 1;

						if(fwrite(&c, 1, 1, sfile) != 1) {
							//printf("Save Error: could not save game data\n");
							MessageBox(hwndMain, "Error saving game data", "FUSE Save Error", MB_OK | MB_ICONERROR);
							exit(1);
						}

						if(fwrite(bdata.data, bdata.size, 1, sfile) != 1) {
							//printf("Save Error: could not save game data\n");
							MessageBox(hwndMain, "Error saving game data", "FUSE Save Error", MB_OK | MB_ICONERROR);
							exit(1);
						}
					}

					count += 1;

					// increment save progress bar
					SendMessage(GetDlgItem(hwndMain, IDC_SAVEPROGRESS), PBM_SETPOS, (count * 100) / (itemcount + npccount), 0);

					index = nextbackup();

					if(index == -1) {
						// backup should be complete; close backup file and
						// copy backup file to game data file
						fclose(sfile);

						if(CopyFile("save.bak", "save.fus", FALSE) == 0) {
							//printf("Save Error: could not copy backup data to save file\n");
							MessageBox(hwndMain, "Error copying backup file to save file", "FUSE Save Error", MB_OK | MB_ICONERROR);
							exit(1);
						}

						// tell the progress bar that the save is complete
						SendMessage(GetDlgItem(hwndMain, IDC_SAVEPROGRESS), PBM_SETPOS, 100, 0);

						backup = false;

						/*// if nexttime - thistime > 0, sleep for nexttime - thistime
						GetSystemTime(&st);
						SystemTimeToFileTime(&st, &ft);
						memcpy(&li, &ft, sizeof(LARGE_INTEGER));
						// convert relative time from 100 nanosecond units to millisecond units
						dest.QuadPart = (dest.QuadPart - li.QuadPart) / 10000;

						if(dest.QuadPart > 0) {
							Sleep((DWORD)dest.QuadPart);
						}
						else if(saveinterval.QuadPart == 0) {
							Sleep(INFINITE);
						}
						else {
							Sleep(0);
						}*/

						// tell the progress bar that no save is in progress
						SendMessage(GetDlgItem(hwndMain, IDC_SAVEPROGRESS), PBM_SETPOS, 0, 0);

						break;
					}
				}
			}
		}

		// set savemin to 1 so that we perform single saves
		// until it is time for the game thread to awaken
		savemin = 1;

		// exit critical section
		LeaveCriticalSection(&gamedatacsect);

		// if this is the case, we have completed the backup
		if(!backup)
			SaveComplete();
	}

	ExitThread(0);

	return 0;
}
/*
int gamelogin(SOCKET sock, sockaddr_in *ca)
{
	char buf[300], encrypt1[64], encrypt2[64], name[30], pw[30];
	unsigned long key;
	int size, acct, ecpos = 0, count; //locale
	unsigned npcid;
	//dynamic *data;
	//char makedata;
	const PCList *pcs;
	bool logdone;

	// read encryption key
	if((size = ReceiveBytes(sock, &key, 4)) == SOCKET_ERROR) {
		//printf("Error receiving ip address!\n");
		serr = WSAGetLastError();

		return 1;
	}
	else if(size == 0) {
		//printf("Login aborted by client (1).\n");

		return 1;
	}
	else {
		//printf("Encryption key\n");
		//recvoutput(&key, size);
	}

	int i, j, index1, index2, bit1, bit2, bit3, val;

	// attempt
	key = ntohl(key);
	for(i = 0; i < 64; i++) {
		val = 0;
		for(j = i - 7; j <= i; j++) {
			index1 = (j + 64) & 0x3f;
			index2 = (j + 64 + 1) & 0x3f;
			bit1 = (key >> (stuff1[index1] - 1)) & 1;
			bit2 = (key >> (stuff1[index2] - 1)) & 1;
			bit3 = stuff3[index1];
			val = (val >> 1) | ((bit1 ^ bit2 ^ bit3) << 7);
		}

		encrypt1[i] = val;
	}

	val = (unsigned char)key ^ 0xaa;

	for(i = 0; i < 64; i++) {
		encrypt2[i] = val;
		val ^= (unsigned int)encrypt1[(i + 64) & 0x3f];
	}

	// simply take not of this (not necessary, really)
	if(key == 0x1157) {
		//printf("Hmm...  0x1157 login\n");
		MessageBox(hwndMain, "Game login on game socket!!", "FUSE Info...", MB_OK | MB_ICONINFORMATION);
	}

	// instead of executing loop here, put login into loginlist
	// if loginlist is full, send message indicating that

	logdone = false;

	// login info; must be complete for proper login
	acct = -1;

	while(!logdone) {
		// get message type
		if((size = ReceiveBytes(sock, buf, 1)) == SOCKET_ERROR) {
			//printf("Error receiving message type!\n");
			serr = WSAGetLastError();

			return 1;
		}
		else if(size == 0) {
			//printf("Login aborted by client.\n");
			return 1;
		}
		else {
			ClientDecode(buf, size, ecpos, encrypt2);
			//recvoutput(buf, size);
		}

		switch((unsigned char)buf[0]) {
		case 0x80: // login server only
			count = 62;
			if((size = ReceiveBytes(sock, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving user info!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");

				return 1;
			}
			else {
				ClientDecode(buf + 1, size, ecpos, encrypt2);
				//printf("User info\n");
				//recvoutput(buf, size + 1);
			}

			// send message to the client, too
			// to be safe, we may want to set buf[30] and buf[60] to 0
			if((acct = ValidateAccount(buf + 1, buf + 31)) < 0) {
				if(acct == -2) {
					printf("Login declined, invalid credentials.\n");

					buf[0] = (char)0x82;
					buf[1] = 3;

					if((size = send(sock, buf, 2, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					return 0; //1;
				}
				// assume invalid account
				else {
					printf("Login declined, invalid account.\n");

					buf[0] = (char)0x82;
					buf[1] = 0;

					if((size = send(sock, buf, 2, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					return 0; //1;
				}
			}

			// respond to login request
			// eventually, user info would be passed here as well
			loginresponse(sock, acct);

			break;

		case 0xa4: // login server only
			count = 3;
			if((size = ReceiveBytes(sock, buf + 1, count - 1)) == SOCKET_ERROR) {
				printf("Error receiving user info!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				printf("Login aborted by client (2).\n");

				return 1;
			}
			else {
				ClientDecode(buf + 1, size, ecpos, encrypt2);
				printf("Server select info\n");
				recvoutput(buf, size + 1);
			}

			// process server selection
			// eventually, server selection info would be passed here as well
			redirectresponse(sock);

			break;

		case 0xa0: // login server only
			count = 3;
			if((size = ReceiveBytes(sock, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving user info!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");

				return 1;
			}
			else {
				ClientDecode(buf + 1, size, ecpos, encrypt2);
				//printf("Server select info\n");
				//recvoutput(buf, size + 1);
			}

			// process server selection
			// eventually, server selection info would be passed here as well
			// if we chose not to redirect here, simply send the gamelogininfo
			redirectresponse(sock);

			// respond to login request
			// eventually, user info would be passed here as well
			//gameloginresponse(sock, acct);

			break;

		case 0x91:
			count = 65;
			if((size = ReceiveBytes(sock, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving user info (2)!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");

				return 1;
			}
			else {
				ClientDecode(buf + 1, size, ecpos, encrypt2);
				//printf("User info (2)\n");
				//recvoutput(buf, size + 1);
			}

			// send message to the client, too
			// to be safe, we may want to set buf[34] and buf[64] to 0
			if((acct = ValidateAccount(buf + 5, buf + 35)) < 0) {
				if(acct == -2) {
					//printf("Login declined, invalid credentials.\n");

					buf[0] = (char)82;
					buf[1] = 3;

					if((size = send(sock, buf, 2, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					return 0; //1;
				}
				// assume invalid account
				else {
					//printf("Login declined, invalid account.\n");

					buf[0] = (char)82;
					buf[1] = 0;

					if((size = send(sock, buf, 2, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					return 0; //1;
				}
			}

			// respond to login request
			// eventually, user info would be passed here as well
			gameloginresponse(sock, acct);

			break;

		case 0x00: // login server only
			count = 96; // 96 bytes for actual char create message
			if((size = ReceiveBytes(sock, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving character select info!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");

				return 1;
			}
			else {
				ClientDecode(buf + 1, size, ecpos, encrypt2);
				//printf("Character select info\n");
				//recvoutput(buf, size + 1);
			}

			// 4 ip address bytes
			if((size = ReceiveBytes(sock, buf + count, 4)) == SOCKET_ERROR) {
				//printf("Error receiving character select info!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (4).\n");

				return 1;
			}
			else {
				ClientDecode(buf + count, size, ecpos, encrypt2);
			}

			// verify that acct is valid!!
			if(acct < 0) {
				buf[0] = (char)0x82;
				buf[1] = 0;

				if((size = send(sock, buf, 2, 0)) == SOCKET_ERROR) {
					//printf("Error sending login response to client!\n");
					serr = WSAGetLastError();
				}
				else {
					//sendoutput(buf, size);
				}

				// keep socket alive
				return 0; //1;
			}

			// fill in npcdata
			npcid = 0;

			makedata.reset();

			CopyBuf(makedata.name, STRSIZE, buf + 10);
			CopyBuf(makedata.pw, STRSIZE, buf + 40);

			makedata.gender = buf[70];

			if(makedata.gender == 0) {
				makedata.mymodel = makedata.model = 0x0190;
			}
			else {
				makedata.mymodel = makedata.model = 0x0191;
			}

			makedata.hits = makedata.totalhits = makedata.strength = buf[71];
			makedata.stamina = makedata.totalstamina = makedata.dexterity = buf[72];
			makedata.mana = makedata.totalmana = makedata.intelligence = buf[73];

			makedata.acct = acct;

			locale = (unsigned char)(buf[91]);

			// coordinates are determined by starting locale
			if(locale < 0 || locale >= localecount) {
				makedata.x = 0x0723;
				makedata.y = 0x0b80;
				makedata.z = GetMapZ(makedata.x, makedata.y);
			}
			else {
				// how do we determine the selected locale?
				makedata.x = locales[locale] -> x;
				makedata.y = locales[locale] -> y;

				if(locales[locale] -> z == -1)
					makedata.z = GetMapZ(makedata.x, makedata.y);
				else
					makedata.z = locales[locale] -> z;
			}

			makedata.skills[0] = UnBufferByte(buf, 74);
			makedata.skills[1] = UnBufferByte(buf, 75);
			makedata.skills[2] = UnBufferByte(buf, 76);
			makedata.skills[3] = UnBufferByte(buf, 77);
			makedata.skills[4] = UnBufferByte(buf, 78);
			makedata.skills[5] = UnBufferByte(buf, 79);

			makedata.mycolor = makedata.color = UnBufferShort(buf, 80) | 0x8000;
			makedata.hairtype = UnBufferShort(buf, 82);
			makedata.haircolor = UnBufferShort(buf, 84);
			makedata.beardtype = UnBufferShort(buf, 86);
			makedata.beardcolor = UnBufferShort(buf, 88);

			EnterCriticalSection(&gamedatacsect);

			// npcid of 0 means that the npc is being created
			if(addconnection(sock, acct, npcid, ecpos, encrypt2, makedata) == -1) {
				// there should be a message for this
				//printf("Connection attempt aborted: no available connections.\n");
				return 1;
			}

			LeaveCriticalSection(&gamedatacsect);

			logdone = true;

			break;

		case 0x5d: // login server only
			count = 69; // 69 bytes for actual char select message
			if((size = ReceiveBytes(sock, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving character select info!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (3).\n");

				return 1;
			}
			else {
				ClientDecode(buf + 1, size, ecpos, encrypt2);
				//printf("Character select info\n");
				//recvoutput(buf, size + 1);
			}

			// 4 ip address bytes
			if((size = ReceiveBytes(sock, buf + count, 4)) == SOCKET_ERROR) {
				//printf("Error receiving character select info!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (4).\n");

				return 1;
			}
			else {
				ClientDecode(buf + count, size, ecpos, encrypt2);
				//printf("Character select info\n");
				//recvoutput(buf, size + 1);
			}

			// verify that acct is valid!!
			if(acct < 0) {
				buf[0] = (char)0x82;
				buf[1] = 0;

				if((size = send(sock, buf, 2, 0)) == SOCKET_ERROR) {
					//printf("Error sending login response to client!\n");
					serr = WSAGetLastError();
				}
				else {
					//sendoutput(buf, size);
				}

				// keep socket alive
				return 0; //1;
			}

			pcs = GetPlayers(acct);

			for(i = 0; i < pcs -> count; i++) {
				//data = getdata(pcs -> list[i]);
				GetPCNameAndPass(pcs -> list[i], name, pw);

				if(strcmp(name, buf + 5) == 0)
					break;
			}

			if(i < pcs -> count) {
				if(strcmp(pw, buf + 35) == 0) {
					npcid = pcs -> list[i];
				}
				else {
					// invalid char pass
					//printf("Login declined, invalid character password.\n");

					buf[0] = (char)53;
					buf[1] = 0;

					if((size = send(sock, buf, 2, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					return 0; //1;
				}
			}
			else {
				// char does not exist
				//printf("Login declined, invalid character.\n");
				return 1;
			}

			// npcid should already be set

			EnterCriticalSection(&gamedatacsect);

			// npcid of 0 means that the npc is being created
			if(addconnection(sock, acct, npcid, ecpos, encrypt2, NULL) == -1) {
				// there should be a message for this
				//printf("Connection attempt aborted: no available connections.\n");
				return 1;
			}

			LeaveCriticalSection(&gamedatacsect);

			logdone = true;

			break;

		default:
			MessageBox(hwndMain, "Bad message in game socket", "FUSE Info...", MB_OK | MB_ICONINFORMATION);

			count = 300; //1000;
			if((size = recv(sock, buf + 1, count - 1, 0)) == SOCKET_ERROR) {
				//printf("Error receiving unknown message!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");

				return 1;
			}
			else {
				ClientDecode(buf + 1, size, ecpos, encrypt2);
				//printf("Unknown message type\n");
				//recvoutput(buf, size + 1);
			}
		}
	}

	return 0;
}
*/
int login(/*SOCKET sock, sockaddr_in *ca*//*LoginInfo &li*/int index)
{
	char buf[10000], output[10000]/*, encrypt1[64], encrypt2[64]*/, name[30], pw[30];
	//unsigned long key;
	int size, /*acct, ecpos = 0,*/ count, locale, i, offset;
	unsigned npcid;
	//dynamic *data;
	char makedata[100];
	const PCList *pcs;
	LoginInfo &li = logins[index];
	//bool logdone;
	dynamic *data;
/*
	// read encryption key
	if((size = ReceiveBytes(sock, &key, 4)) == SOCKET_ERROR) {
		//printf("Error receiving ip address!\n");
		serr = WSAGetLastError();

		return 1;
	}
	else if(size == 0) {
		//printf("Login aborted by client (1).\n");

		return 1;
	}
	else {
		//printf("Encryption key\n");
		//recvoutput(&key, size);
	}

	int i, j, index1, index2, bit1, bit2, bit3, val;

	// attempt
	key = ntohl(key);
	for(i = 0; i < 64; i++) {
		val = 0;
		for(j = i - 7; j <= i; j++) {
			index1 = (j + 64) & 0x3f;
			index2 = (j + 64 + 1) & 0x3f;
			bit1 = (key >> (stuff1[index1] - 1)) & 1;
			bit2 = (key >> (stuff1[index2] - 1)) & 1;
			bit3 = stuff3[index1];
			val = (val >> 1) | ((bit1 ^ bit2 ^ bit3) << 7);
		}

		encrypt1[i] = val;
	}

	val = (unsigned char)key ^ 0xaa;

	for(i = 0; i < 64; i++) {
		encrypt2[i] = val;
		val ^= (unsigned int)encrypt1[(i + 64) & 0x3f];
	}
*/
	// simply take not of this (not necessary, really)
	/*if(key == 0x1157) {
		//printf("Hmm...  0x1157 login\n");
		//MessageBox(hwndMain, "Game login on login socket", "FUSE Info...", MB_OK | MB_ICONINFORMATION);
	}*/

	// instead of executing loop here, put login into loginlist
	// if loginlist is full, send message indicating that

	//logdone = false;

	// login info; must be complete for proper login
	//acct = -1;

	//while(!logdone) {
		// get message type
		if((size = ReceiveBytes(li.s, buf, 1)) == SOCKET_ERROR) {
			//printf("Error receiving message type!\n");
			serr = WSAGetLastError();
			closesocket(li.s);
			return 1;
		}
		else if(size == 0) {
			//printf("Login aborted by client.\n");
			closesocket(li.s);
			return 1;
		}
		else {
			ClientDecode(buf, size, li.ecpos, li.encrypt);
			//recvoutput(buf, size);
		}

		switch((unsigned char)buf[0]) {
		case 0x00:
			count = 96; // 96 bytes for actual char create message
			if((size = ReceiveBytes(li.s, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving character select info!\n");
				serr = WSAGetLastError();
				closesocket(li.s);
				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");
				closesocket(li.s);
				return 1;
			}
			else {
				ClientDecode(buf + 1, size, li.ecpos, li.encrypt);
				//printf("Character select info\n");
				//recvoutput(buf, size + 1);
			}

			// 4 ip address bytes
			if((size = ReceiveBytes(li.s, buf + count, 4)) == SOCKET_ERROR) {
				//printf("Error receiving character select info!\n");
				serr = WSAGetLastError();
				closesocket(li.s);
				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (4).\n");
				closesocket(li.s);
				return 1;
			}
			else {
				ClientDecode(buf + count, size, li.ecpos, li.encrypt);
			}

			// verify that acct is valid!!
			if(li.acct < 0) {
				buf[0] = (char)0x82;
				buf[1] = 0;

				offset = PutMessage(buf, 2, output);

				if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
					//printf("Error sending login response to client!\n");
					serr = WSAGetLastError();
					closesocket(li.s);
					return 1;
				}
				else {
					//sendoutput(buf, size);
				}

				// keep socket alive
				//return 1; // 0
			}
			else {
				pcs = GetPlayers(li.acct);

				for(offset = 0; offset < pcs -> count; offset++) {
					if(NPCConnection(pcs -> list[offset]) != -1)
						break;
				}

				if(offset < pcs -> count) {
					// another char active
					//printf("Login declined, invalid character password.\n");

					buf[0] = (char)0x53;
					buf[1] = 5;

					offset = PutMessage(buf, 2, output);

					if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
						closesocket(li.s);
						return 1;
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					//return 0; //1;
				}
				else {
					// fill in npcdata
					npcid = 0;

					//makedata.reset();

					BufferShort(makedata, 0, 93);

					CopyBuf(makedata + 2, STRSIZE, buf + 10);
					CopyBuf(makedata + STRSIZE + 2, STRSIZE, buf + 40);

					BufferByte(makedata, 2 * STRSIZE + 2, buf[70]);

					if(buf[70] == 0) {
						BufferShort(makedata, 2 * STRSIZE + 3, 0x0190);
					}
					else {
						BufferShort(makedata, 2 * STRSIZE + 3, 0x0191);
					}

					BufferByte(makedata, 2 * STRSIZE + 5, buf[71]);
					BufferByte(makedata, 2 * STRSIZE + 6, buf[72]);
					BufferByte(makedata, 2 * STRSIZE + 7, buf[73]);

					BufferInt(makedata, 2 * STRSIZE + 8, li.acct);

					locale = (unsigned char)(buf[91]);

					// coordinates are determined by starting locale
					if(locale < 0 || locale >= localecount) {
						BufferShort(makedata, 2 * STRSIZE + 12, 0x0723);
						BufferShort(makedata, 2 * STRSIZE + 14, 0x0b80);
						BufferByte(makedata, 2 * STRSIZE + 16, GetMapZ(0x0723, 0x0b80));
					}
					else {
						// how do we determine the selected locale?
						BufferShort(makedata, 2 * STRSIZE + 12, locales[locale] -> x);
						BufferShort(makedata, 2 * STRSIZE + 14, locales[locale] -> y);

						if(locales[locale] -> z == -1)
							BufferByte(makedata, 2 * STRSIZE + 16, GetMapZ(locales[locale] -> x, locales[locale] -> y));
						else
							BufferByte(makedata, 2 * STRSIZE + 16, locales[locale] -> z);
					}

					BufferByte(makedata, 2 * STRSIZE + 17, UnBufferByte(buf, 74));
					BufferByte(makedata, 2 * STRSIZE + 18, UnBufferByte(buf, 75));
					BufferByte(makedata, 2 * STRSIZE + 19, UnBufferByte(buf, 76));
					BufferByte(makedata, 2 * STRSIZE + 20, UnBufferByte(buf, 77));
					BufferByte(makedata, 2 * STRSIZE + 21, UnBufferByte(buf, 78));
					BufferByte(makedata, 2 * STRSIZE + 22, UnBufferByte(buf, 79));

					BufferShort(makedata, 2 * STRSIZE + 23, UnBufferShort(buf, 80));
					BufferShort(makedata, 2 * STRSIZE + 25, UnBufferShort(buf, 82));
					BufferShort(makedata, 2 * STRSIZE + 27, UnBufferShort(buf, 84));
					BufferShort(makedata, 2 * STRSIZE + 29, UnBufferShort(buf, 86));
					BufferShort(makedata, 2 * STRSIZE + 31, UnBufferShort(buf, 88));

					EnterCriticalSection(&gamedatacsect);

					// npcid of 0 means that the npc is being created
					if(addconnection(li.s, li.acct, npcid, li.ecpos, li.encrypt, makedata) == -1) {
						// server full
						//printf("Login declined, invalid character password.\n");

						buf[0] = (char)0x53;
						buf[1] = 3; // do we have a server full message?

						offset = PutMessage(buf, 2, output);

						if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
							//printf("Error sending login response to client!\n");
							serr = WSAGetLastError();
							closesocket(li.s);
							return 1;
						}
						else {
							//sendoutput(buf, size);
						}

						// keep socket alive
						//return 0; //1;

						LeaveCriticalSection(&gamedatacsect);
					}
					else {
						LeaveCriticalSection(&gamedatacsect);

						//logdone = true;
						return 1;
					}
				}
			}

			break;

		case 0x5d:
			count = 69; // 69 bytes for actual char select message
			if((size = ReceiveBytes(li.s, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving character select info!\n");
				serr = WSAGetLastError();
				closesocket(li.s);
				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (3).\n");
				closesocket(li.s);
				return 1;
			}
			else {
				ClientDecode(buf + 1, size, li.ecpos, li.encrypt);
				//printf("Character select info\n");
				//recvoutput(buf, size + 1);
			}

			// 4 ip address bytes
			if((size = ReceiveBytes(li.s, buf + count, 4)) == SOCKET_ERROR) {
				//printf("Error receiving character select info!\n");
				serr = WSAGetLastError();
				closesocket(li.s);
				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (4).\n");
				closesocket(li.s);
				return 1;
			}
			else {
				ClientDecode(buf + count, size, li.ecpos, li.encrypt);
				//printf("Character select info\n");
				//recvoutput(buf, size + 1);
			}

			// verify that acct is valid!!
			if(li.acct < 0) {
				buf[0] = (char)0x82;
				buf[1] = 0;

				offset = PutMessage(buf, 2, output);

				if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
					//printf("Error sending login response to client!\n");
					serr = WSAGetLastError();
					closesocket(li.s);
					return 1;
				}
				else {
					//sendoutput(buf, size);
				}

				// keep socket alive
				//return 1; // 0
			}
			else {
				pcs = GetPlayers(li.acct);

				// is pc index the entire int at buf + 65?
				// or is it a short or byte? for safety, use byte at buf + 68 for now
				/*for(i = 0; i < pcs -> count; i++) {
					GetPCNameAndPass(pcs -> list[i], name, pw);

					if(strcmp(name, buf + 5) == 0)
						break;
				}*/

				i = (unsigned char)buf[68];

				if(i < pcs -> count) {
					GetPCNameAndPass(pcs -> list[i], name, pw);

					for(offset = 0; offset < pcs -> count; offset++) {
						if(offset != i && NPCConnection(pcs -> list[offset]) != -1)
							break;
					}

					if(strcmp(pw/*data -> pw()*/, buf + 35) != 0) {
						// invalid char pass
						//printf("Login declined, invalid character password.\n");

						buf[0] = (char)0x53;
						buf[1] = 0;

						offset = PutMessage(buf, 2, output);

						if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
							//printf("Error sending login response to client!\n");
							serr = WSAGetLastError();
							closesocket(li.s);
							return 1;
						}
						else {
							//sendoutput(buf, size);
						}

						// keep socket alive
						//return 0; //1;
					}
					else if((data = getdata(pcs -> list[i])) == NULL) {
						// another char active
						//printf("Login declined, invalid character password.\n");

						buf[0] = (char)0x53;
						buf[1] = 1;

						offset = PutMessage(buf, 2, output);

						if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
							//printf("Error sending login response to client!\n");
							serr = WSAGetLastError();
							closesocket(li.s);
							return 1;
						}
						else {
							//sendoutput(buf, size);
						}

						// keep socket alive
						//return 0; //1;
					}
					else if(data -> connection() != -1) {
						// another char active
						//printf("Login declined, invalid character password.\n");

						buf[0] = (char)0x53;
						buf[1] = 2;

						offset = PutMessage(buf, 2, output);

						if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
							//printf("Error sending login response to client!\n");
							serr = WSAGetLastError();
							closesocket(li.s);
							return 1;
						}
						else {
							//sendoutput(buf, size);
						}

						// keep socket alive
						//return 0; //1;
					}
					else if(offset < pcs -> count) {
						// another char active
						//printf("Login declined, invalid character password.\n");

						buf[0] = (char)0x53;
						buf[1] = 5;

						offset = PutMessage(buf, 2, output);

						if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
							//printf("Error sending login response to client!\n");
							serr = WSAGetLastError();
							closesocket(li.s);
							return 1;
						}
						else {
							//sendoutput(buf, size);
						}

						// keep socket alive
						//return 0; //1;
					}
					else {
						npcid = pcs -> list[i];

						// npcid should already be set

						EnterCriticalSection(&gamedatacsect);

						// npcid of 0 means that the npc is being created
						if(addconnection(li.s, li.acct, npcid, li.ecpos, li.encrypt, NULL) == -1) {
							// there should be a message for this
							//printf("Connection attempt aborted: no available connections.\n");
							return 1;
						}

						LeaveCriticalSection(&gamedatacsect);

						//logdone = true;
						return 1;
					}
				}
				else {
						// another char active
						//printf("Login declined, invalid character password.\n");

						buf[0] = (char)0x53;
						buf[1] = 6;

						offset = PutMessage(buf, 2, output);

						if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
							//printf("Error sending login response to client!\n");
							serr = WSAGetLastError();
							closesocket(li.s);
							return 1;
						}
						else {
							//sendoutput(buf, size);
						}

						// keep socket alive
						//return 0; //1;
				}
			}

			break;

		case 0x80:
			count = 62;
			if((size = ReceiveBytes(li.s, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving user info!\n");
				serr = WSAGetLastError();
				closesocket(li.s);
				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");
				closesocket(li.s);
				return 1;
			}
			else {
				ClientDecode(buf + 1, size, li.ecpos, li.encrypt);
				//printf("User info\n");
				//recvoutput(buf, size + 1);
			}

			li.acct = ValidateAccount(buf + 1, buf + 31);

			for(i = 0; i < MAXLOGIN; i++) {
				if(i != index && logins[i].logged && logins[i].acct == li.acct)
					break;
			}

			// send message to the client, too
			// to be safe, we may want to set buf[30] and buf[60] to 0
			if(logincheck && li.acct < 0) {
				if(li.acct == -2) {
					printf("Login declined, invalid credentials.\n");

					buf[0] = (char)0x82;
					buf[1] = 3;

					if((size = send(li.s, buf, 2, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
						closesocket(li.s);
						return 1;
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					//return 0; // 1
				}
				else if(li.acct == -3) {
					printf("Login declined, account blocked.\n");

					buf[0] = (char)0x82;
					buf[1] = 2;

					if((size = send(li.s, buf, 2, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
						closesocket(li.s);
						return 1;
					}
					else {
						//sendoutput(buf, size);
					}

					return 1;
				}
				// assume invalid account
				else {
					printf("Login declined, invalid account.\n");

					buf[0] = (char)0x82;
					buf[1] = 0;

					if((size = send(li.s, buf, 2, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
						closesocket(li.s);
						return 1;
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					//return 0; // 1
				}
			}
			else if(logincheck && i < MAXLOGIN) {
				buf[0] = (char)0x82;
				buf[1] = 1;

				if((size = send(li.s, buf, 2, 0)) == SOCKET_ERROR) {
					//printf("Error sending login response to client!\n");
					serr = WSAGetLastError();
					closesocket(li.s);
					return 1;
				}
				else {
					//sendoutput(buf, size);
				}

				// keep socket alive
				//return 0; // 1
			}
			else {
				// respond to login request
				// eventually, user info would be passed here as well
				loginresponse(li.s, li.acct);
			}

			break;

		case 0x83:
			count = 39;
			if((size = ReceiveBytes(li.s, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving user info!\n");
				serr = WSAGetLastError();
				closesocket(li.s);
				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");
				closesocket(li.s);
				return 1;
			}
			else {
				ClientDecode(buf + 1, size, li.ecpos, li.encrypt);
				//printf("User info\n");
				//recvoutput(buf, size + 1);
			}

			// verify that acct is valid!!
			if(li.acct < 0) {
				buf[0] = (char)0x82;
				buf[1] = 0;

				offset = PutMessage(buf, 2, output);

				if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
					//printf("Error sending login response to client!\n");
					serr = WSAGetLastError();
					closesocket(li.s);
					return 1;
				}
				else {
					//sendoutput(buf, size);
				}

				// keep socket alive
				//return 1; // 0
			}
			else {
				pcs = GetPlayers(li.acct);

				// is pc index the entire int at buf + 65?
				// or is it a short or byte? for safety, use byte at buf + 68 for now
				/*for(i = 0; i < pcs -> count; i++) {
					GetPCNameAndPass(pcs -> list[i], name, pw);

					if(strcmp(name, buf + 5) == 0)
						break;
				}*/

				i = (unsigned char)buf[34];

				if(i < pcs -> count) {
					GetPCNameAndPass(pcs -> list[i], name, pw);

					if(strcmp(pw/*data -> pw()*/, buf + 1) != 0) {
						// invalid char pass
						//printf("Login declined, invalid character password.\n");

						buf[0] = (char)0x85;
						buf[1] = 0;

						offset = PutMessage(buf, 2, output);

						if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
							//printf("Error sending login response to client!\n");
							serr = WSAGetLastError();
							closesocket(li.s);
							return 1;
						}
						else {
							//sendoutput(buf, size);
						}

						// keep socket alive
						//return 0; //1;
					}
					else if((data = getdata(pcs -> list[i])) == NULL) {
						// another char active
						//printf("Login declined, invalid character password.\n");

						buf[0] = (char)0x85;
						buf[1] = 1;

						offset = PutMessage(buf, 2, output);

						if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
							//printf("Error sending login response to client!\n");
							serr = WSAGetLastError();
							closesocket(li.s);
							return 1;
						}
						else {
							//sendoutput(buf, size);
						}

						// keep socket alive
						//return 0; //1;
					}
					else if(data -> connection() != -1) {
						// another char active
						//printf("Login declined, invalid character password.\n");

						buf[0] = (char)0x85;
						buf[1] = 2;

						offset = PutMessage(buf, 2, output);

						if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
							//printf("Error sending login response to client!\n");
							serr = WSAGetLastError();
							closesocket(li.s);
							return 1;
						}
						else {
							//sendoutput(buf, size);
						}

						// keep socket alive
						//return 0; //1;
					}
					else {
						npcid = pcs -> list[i];

						// delete char
						if(DeletePC(li.acct, npcid) == 0) {
							DeletePlayer(li.acct, npcid);

							// show char list again
							BufferByte(buf, 0, (char)0x86);
							BufferShort(buf, 1, 304);

							BufferByte(buf, 3, pcs -> count);

							for(i = 0; i < 5; i++) {
								if(i < pcs -> count) {
									GetPCNameAndPass(pcs -> list[i], name, pw);
									CopyBuf(buf + 4 + i * 60, 30, name);

									if(strcmp(pw, "") != 0) {
										CopyBuf(buf + 4 + i * 60 + 30, 30, "!");
									}
									else {
										memset(buf + 4 + i * 60 + 30, 0, 30);
									}
								}
								else {
									memset(buf + 4 + i * 60, 0, 60);
								}
							}

							offset = PutMessage(buf, 304, output);

							if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
								//printf("Error sending login response to client!\n");
								serr = WSAGetLastError();
								closesocket(li.s);
								return 1;
							}
							else {
								//sendoutput(buf, size);
							}
						}
						else {
							// char delete failed
							buf[0] = (char)0x85;
							buf[1] = 5;

							offset = PutMessage(buf, 2, output);

							if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
								//printf("Error sending login response to client!\n");
								serr = WSAGetLastError();
								closesocket(li.s);
								return 1;
							}
						}
					}
				}
				else {
					// another char active
					//printf("Login declined, invalid character password.\n");

					buf[0] = (char)0x85;
					buf[1] = 5;

					offset = PutMessage(buf, 2, output);

					if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
						closesocket(li.s);
						return 1;
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					//return 0; //1;
				}
			}

			break;

		case 0x91:
			count = 65;
			if((size = ReceiveBytes(li.s, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving user info (2)!\n");
				serr = WSAGetLastError();
				closesocket(li.s);
				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");
				closesocket(li.s);
				return 1;
			}
			else {
				ClientDecode(buf + 1, size, li.ecpos, li.encrypt);
				//printf("User info (2)\n");
				//recvoutput(buf, size + 1);
			}

			li.acct = ValidateAccount(buf + 5, buf + 35);

			for(i = 0; i < MAXLOGIN; i++) {
				if(i != index && logins[i].logged && logins[i].acct == li.acct)
					break;
			}

			// send message to the client, too
			// to be safe, we may want to set buf[34] and buf[64] to 0
			if(li.acct < 0) {
				if(li.acct == -2) {
					//printf("Login declined, invalid credentials.\n");

					buf[0] = (char)0x82;
					buf[1] = 3;

					offset = PutMessage(buf, 2, output);

					if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
						closesocket(li.s);
						return 1;
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					//return 1; // 0
				}
				if(li.acct == -3) {
					//printf("Login declined, account blocked.\n");

					buf[0] = (char)0x82;
					buf[1] = 2;

					offset = PutMessage(buf, 2, output);

					if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
						closesocket(li.s);
						return 1;
					}
					else {
						//sendoutput(buf, size);
					}

					return 1;
				}
				// assume invalid account
				else {
					//printf("Login declined, invalid account.\n");

					buf[0] = (char)0x82;
					buf[1] = 0;

					offset = PutMessage(buf, 2, output);

					if((size = send(li.s, output, offset, 0)) == SOCKET_ERROR) {
						//printf("Error sending login response to client!\n");
						serr = WSAGetLastError();
						closesocket(li.s);
						return 1;
					}
					else {
						//sendoutput(buf, size);
					}

					// keep socket alive
					//return 1; // 0
				}
			}
			else if(i < MAXLOGIN) {
				buf[0] = (char)0x82;
				buf[1] = 1;

				if((size = send(li.s, buf, 2, 0)) == SOCKET_ERROR) {
					//printf("Error sending login response to client!\n");
					serr = WSAGetLastError();
					closesocket(li.s);
					return 1;
				}
				else {
					//sendoutput(buf, size);
				}

				// keep socket alive
				//return 0; // 1
			}
			else {
				// respond to login request
				// eventually, user info would be passed here as well
				gameloginresponse(li.s, li.acct);
			}

			break;

		case 0xa0:
			count = 3;
			if((size = ReceiveBytes(li.s, buf + 1, count - 1)) == SOCKET_ERROR) {
				//printf("Error receiving user info!\n");
				serr = WSAGetLastError();
				closesocket(li.s);
				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");
				closesocket(li.s);
				return 1;
			}
			else {
				ClientDecode(buf + 1, size, li.ecpos, li.encrypt);
				//printf("Server select info\n");
				//recvoutput(buf, size + 1);
			}

			// process server selection
			// eventually, server selection info would be passed here as well
			// if we chose not to redirect here, simply send the gamelogininfo
			redirectresponse(li.s, UnBufferShort(buf, 1));
			/*count = 1; // temp; try setting to 0 if using as a pure login server
			return count;*/

			// respond to login request
			// eventually, user info would be passed here as well
			//gameloginresponse(sock, acct);

			break;

		// client machine profile?
		/*case 0xa4:
			count = 3;
			if((size = ReceiveBytes(sock, buf + 1, count - 1)) == SOCKET_ERROR) {
				printf("Error receiving user info!\n");
				serr = WSAGetLastError();

				return 1;
			}
			else if(size == 0) {
				printf("Login aborted by client (2).\n");

				return 1;
			}
			else {
				ClientDecode(buf + 1, size, ecpos, encrypt2);
				printf("Server select info\n");
				recvoutput(buf, size + 1);
			}

			// process server selection
			// eventually, server selection info would be passed here as well
			redirectresponse(sock);

			break;*/

		default:
			//MessageBox(hwndMain, "Bad message in login socket", "FUSE Info...", MB_OK | MB_ICONINFORMATION);

			count = 300; //1000;
			if((size = recv(li.s, buf + 1, count - 1, 0)) == SOCKET_ERROR) {
				//printf("Error receiving unknown message!\n");
				serr = WSAGetLastError();
				closesocket(li.s);
				return 1;
			}
			else if(size == 0) {
				//printf("Login aborted by client (2).\n");
				closesocket(li.s);
				return 1;
			}
			else {
				ClientDecode(buf + 1, size, li.ecpos, li.encrypt);
				//printf("Unknown message type\n");
				//recvoutput(buf, size + 1);
			}
		}
	//}

	return 0;
}

void loginresponse(SOCKET sock, int acct)
{
	char buf[500];
	int size, offset = 0, i;

	BufferByte(buf, 0, (char)0xa8);

	// ??
	BufferByte(buf, 3, (char)0x05);

	// server count
	BufferShort(buf, 4, servercount + 1);

	offset = 6;

	// server info

	// server id
	BufferShort(buf, offset, 0);
	offset += 2;

	//CopyBuf(buf + offset, 32, "Test Server");
	CopyBuf(buf + offset, 32, servname);
	offset += 32;

	// server %full info
	BufferByte(buf, offset, 0);
	offset += 1;

	// time zone
	BufferByte(buf, offset, 0);
	offset += 1;

	// test addr
	BufferInt(buf, offset, ntohl(servaddr));
	offset += 4;

	for(i = 0; i < servercount; i++) {
		// server id
		BufferShort(buf, offset, i + 1);
		offset += 2;

		CopyBuf(buf + offset, 32, servers[i] -> name);
		offset += 32;

		// server %full info
		BufferByte(buf, offset, 0);
		offset += 1;

		// time zone
		BufferByte(buf, offset, servers[i] -> zone);
		offset += 1;

		// test addr
		BufferInt(buf, offset, ntohl(servers[i] -> testaddr));
		offset += 4;
	}

	size = offset;

	BufferShort(buf, 1, size);

	if((size = send(sock, buf, size, 0)) == SOCKET_ERROR) {
		//printf("Error sending login response to client!\n");
		serr = WSAGetLastError();
	}
	else {
		//sendoutput(buf, size);
	}
}

void redirectresponse(SOCKET sock, int id)
{
	char buf[50];
	int size;
	/*sockaddr_in ga;

	size = sizeof(sockaddr_in);

	if(getsockname(sock, (sockaddr *)&ga, &size) == SOCKET_ERROR) {
		//printf("Error getting game server address!\n");
		serr = WSAGetLastError();
	}*/

	buf[0] = (char)0x8c; // ? redirect command code?
	//*((int *)(buf + 1)) = servaddr; //ga.sin_addr.s_addr; // game server ip addr
	//*((short *)(buf + 5)) = htons(TCP_LOGINPORT); // game server port addr
	buf[7] = 0x00;
	buf[8] = 0x00;
	buf[9] = 0x11;
	buf[10] = 0x57;

	if(id == 0) {
		BufferInt(buf, 1, ntohl(servaddr));
		BufferShort(buf, 5, TCP_LOGINPORT);
	}
	else {
		BufferInt(buf, 1, ntohl(servers[id - 1] -> addr));
		BufferShort(buf, 5, servers[id - 1] -> port);
	}

	if((size = send(sock, buf, 11, 0)) == SOCKET_ERROR) {
		//printf("Error sending login response to client!\n");
		serr = WSAGetLastError();
	}
	else {
		//sendoutput(buf, size);
	}
}

void gameloginresponse(SOCKET sock, int acct)
{
	char buf[5000], output[5000], name[30], pw[30];
	int size, offset = 0, i;
	const PCList *pcs;
	//dynamic *data;

	pcs = GetPlayers(acct);

	// initialize char area just to be safe
	memset(buf + 4, 0, 300);

	BufferByte(buf, 0, (char)0xa9);

	BufferByte(buf, 3, pcs -> count);

	offset = 4;

	for(i = 0; i < (pcs -> count < 5 ? pcs -> count : 5); i++) {
		//data = getdata(pcs -> list[i]);
		GetPCNameAndPass(pcs -> list[i], name, pw);
		CopyBuf(buf + offset, 30, name/*data -> name()*/);
		offset += 30;

		if(strcmp(pw/*data -> pw()*/, "") != 0) {
			CopyBuf(buf + offset, 30, "!");
		}
		else {
			memset(buf + offset, 0, 30);
		}

		offset += 30;
	}

	while(i < 5) {
		offset += 60;
		i += 1;
	}

	if(localecount == 0) {
		// locale count
		BufferByte(buf, offset, 1);
		offset += 1;

		// locale id
		BufferByte(buf, offset, 0);
		offset += 1;

		// locale city
		CopyBuf(buf + offset, 31, "Trinsic Wilderness");
		offset += 31;

		// locale place
		CopyBuf(buf + offset, 31, "Testing Area");
		offset += 31;
	}
	else {
		// locale count
		BufferByte(buf, offset, (char)localecount);
		offset += 1;

		for(i = 0; i < localecount; i++) {
			// locale id
			BufferByte(buf, offset, (char)i);
			offset += 1;

			// locale city
			CopyBuf(buf + offset, 31, locales[i] -> name1);
			offset += 31;

			// locale place
			CopyBuf(buf + offset, 31, locales[i] -> name2);
			offset += 31;
		}
	}

	size = offset;

	BufferShort(buf, 1, size);

	offset = PutMessage(buf, size, output);

	//if((size = send(sock, buf, size, 0)) == SOCKET_ERROR) {
	if((size = send(sock, output, offset, 0)) == SOCKET_ERROR) {
		//printf("Error sending game server login response to client!\n");
		serr = WSAGetLastError();
	}
	else {
		//sendoutput(buf, size);
	}
}
/*
//print out the data received from the client
void recvoutput(void *ptr, int size)
{
	int lcount, i, j, count, done, c;
	char *buf = (char *)ptr;

	lcount = (size + 15) / 16;
	printf("Read size: %d; Read line count: %d\n", size, lcount);
	done = 0;

	for(i = 0; i < lcount; i++) {
		count = (size - done) < 16 ? (size - done) : 16;

		for(j = 0; j < count; j++) {
			if(buf[done + j] < 0x20) {
				printf("%c", '.');
			}
			else {
				printf("%c", buf[done + j]);
			}
		}

		for(j = 0; j < 16 - count; j++) {
			printf(" ");
		}

		for(j = 0; j < count; j++) {
			printf(" ");

			c = (buf[done + j] >> 4) & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');

			c = buf[done + j] & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');
		}

		done += count;
		printf("\n");
	}
}

//print out the data sent from the client
void sendoutput(void *ptr, int size)
{
	int lcount, i, j, count, done, c;
	char *buf = (char *)ptr;

	lcount = (size + 15) / 16;
	printf("Write size: %d; Write line count: %d\n", size, lcount);
	done = 0;

	for(i = 0; i < lcount; i++) {
		count = (size - done) < 16 ? (size - done) : 16;

		for(j = 0; j < count; j++) {
			if(buf[done + j] < 0x20) {
				printf("%c", '.');
			}
			else {
				printf("%c", buf[done + j]);
			}
		}

		for(j = 0; j < 16 - count; j++) {
			printf(" ");
		}

		for(j = 0; j < count; j++) {
			printf(" ");

			c = (buf[done + j] >> 4) & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');

			c = buf[done + j] & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');
		}

		done += count;
		printf("\n");
	}
}
*/
BOOL CALLBACK DlgAboutProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {
		HANDLE_MSG(hwnd, WM_INITDIALOG, OnInitDialogAbout);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommandAbout);
	}

	return FALSE;
}

BOOL OnInitDialogAbout(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
	SetFocus(GetDlgItem(hwnd, IDOK));
	return FALSE;
}

// id - id of menu item, control, or accelerator
// hwndCtl - identifies control sending message (if not control, this is NULL)
// codeNotify - notification code from control (if menu, 0; if accelerator, 1)
void OnCommandAbout(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	switch(id) {
	case IDOK:
	case IDCANCEL:
		EndDialog(hwnd, TRUE);
		break;
	}
}

/*BOOL CALLBACK DlgStatusProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {
		HANDLE_MSG(hwnd, WM_INITDIALOG, OnInitDialogStatus);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommandStatus);
	}

	return FALSE;
}

BOOL OnInitDialogStatus(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
	in_addr sa;
	char buf[100];

	sa.s_addr = servaddr;
	sprintf(buf, "%s,%d", inet_ntoa(sa), TCP_LOGINPORT);
	Edit_SetText(GetDlgItem(hwnd, IDC_SERVADDR), buf);
	SetFocus(GetDlgItem(hwnd, IDC_SERVADDR));
	Edit_SetSel(GetDlgItem(hwnd, IDC_SERVADDR), 0, strlen(inet_ntoa(sa)));

	//SetFocus(GetDlgItem(hwnd, IDOK));
	return FALSE;
}

// id - id of menu item, control, or accelerator
// hwndCtl - identifies control sending message (if not control, this is NULL)
// codeNotify - notification code from control (if menu, 0; if accelerator, 1)
void OnCommandStatus(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	in_addr sa;

	switch(id) {
	case IDOK:
	case IDCANCEL:
		EndDialog(hwnd, TRUE);
		break;

	case IDC_CLIENTRUN:
		// write out the current server settings for fuserun
		sa.s_addr = servaddr;
		WritePrivateProfileString("FuseRun", "ServerIP", inet_ntoa(sa), "fuse.ini");
		WritePrivateProfileString("FuseRun", "ServerPort", "4500", "fuse.ini");

		// run fuserun
		ZeroMemory(&si, sizeof(si));
		si.cb = sizeof(si);

		if(CreateProcess("fuserun.exe", NULL, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi) == 0) {
			MessageBox(NULL, "Could not start fuserun.exe", "FUSERUN Error", MB_OK | MB_ICONERROR);
		}
		else {
			MessageBox(NULL, "Starting the client from FUSE may take a while.", "FUSERUN", MB_OK | MB_ICONINFORMATION);
		}

		EndDialog(hwnd, TRUE);

		break;
	}
}*/

BOOL CALLBACK DlgStartupProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {
		HANDLE_MSG(hwnd, WM_INITDIALOG, OnInitDialogStartup);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommandStartup);
	}

	return FALSE;
}

BOOL OnInitDialogStartup(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
	int n, i, addr;
	in_addr sa;
	char buf[1000];
	HICON hIcon;

	n = 0;

	while(che -> h_addr_list[n] != NULL) {
		addr = *((int *)(che -> h_addr_list[n]));
		sa.s_addr = addr;
		i = ComboBox_AddString(GetDlgItem(hwnd, IDC_IPADDR), inet_ntoa(sa));
		ComboBox_SetItemData(GetDlgItem(hwnd, IDC_IPADDR), i, addr);
		n += 1;
	}

	GetPrivateProfileString("Fuse", "ServerIP", "127.0.0.1", buf, 1000, "fuse.ini");
	i = ComboBox_FindStringExact(GetDlgItem(hwnd, IDC_IPADDR), -1, buf);

	if(i == CB_ERR)
		ComboBox_SetCurSel(GetDlgItem(hwnd, IDC_IPADDR), 0);
	else
		ComboBox_SetCurSel(GetDlgItem(hwnd, IDC_IPADDR), i);

	GetPrivateProfileString("Fuse", "ServerName", "Test Server", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_SERVERNAME), buf);

	GetPrivateProfileString("Fuse", "MapFile", "c:\\uo\\map0.mul", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_MAPFILE), buf);

	GetPrivateProfileString("Fuse", "TiledataFile", "c:\\uo\\tiledata.mul", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_TILEDATAFILE), buf);

	GetPrivateProfileString("Fuse", "StaticsIndexFile", "c:\\uo\\staidx0.mul", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_STATICSINDEXFILE), buf);

	GetPrivateProfileString("Fuse", "StaticsFile", "c:\\uo\\statics0.mul", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_STATICSFILE), buf);

	GetPrivateProfileString("Fuse", "VerdataFile", "c:\\uo\\verdata.mul", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_VERDATAFILE), buf);

	GetPrivateProfileString("Fuse", "MultiIndexFile", "c:\\uo\\multi.idx", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_MULTIINDEXFILE), buf);

	GetPrivateProfileString("Fuse", "MultiFile", "c:\\uo\\multi.mul", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_MULTIFILE), buf);

	GetPrivateProfileString("Fuse", "SaveInterval", "15", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_SAVEINTERVAL), buf);

	GetPrivateProfileString("Fuse", "ClientKey", "1653181528-1417425230 (1.25.35)", buf, 1000, "fuse.ini");
	Edit_SetText(GetDlgItem(hwnd, IDC_CLIENTKEY), buf);

	GetPrivateProfileString("Fuse", "LoginCheck", "true", buf, 1000, "fuse.ini");
	if(stricmp(buf, "false") == 0 || stricmp(buf, "no") == 0 || stricmp(buf, "off") == 0)
		Button_SetCheck(GetDlgItem(hwnd, IDC_LOGINCHECK), BST_UNCHECKED);
	else
		Button_SetCheck(GetDlgItem(hwnd, IDC_LOGINCHECK), BST_CHECKED);

	SetFocus(GetDlgItem(hwnd, IDC_SERVERNAME));
	Edit_SetSel(GetDlgItem(hwnd, IDC_SERVERNAME), 0, -1);

	// get icon handle
	hIcon = LoadIcon(hThisInstance, MAKEINTRESOURCE(IDI_BOMB2));

	// setup dialog icon
	SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LONG)hIcon);
	SendMessage(hwnd, WM_SETICON, ICON_BIG, (LONG)hIcon);

	//SetFocus(GetDlgItem(hwnd, IDOK));
	return FALSE;
}

// id - id of menu item, control, or accelerator
// hwndCtl - identifies control sending message (if not control, this is NULL)
// codeNotify - notification code from control (if menu, 0; if accelerator, 1)
void OnCommandStartup(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	int i;
	char buf[1000];

	switch(id) {
	case IDCANCEL:
		servaddr = 0;
		servname[0] = mapfile[0] = 0;
		saveinterval = -1;
		logincheck = true;
		EndDialog(hwnd, 1);
		break;

	case IDOK:
		i = ComboBox_GetCurSel(GetDlgItem(hwnd, IDC_IPADDR));
		servaddr = ComboBox_GetItemData(GetDlgItem(hwnd, IDC_IPADDR), i);
		ComboBox_GetText(GetDlgItem(hwnd, IDC_IPADDR), buf, 1000);
		WritePrivateProfileString("Fuse", "ServerIP", buf, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_SERVERNAME), servname, SERVNAMESIZE);
		WritePrivateProfileString("Fuse", "ServerName", servname, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_MAPFILE), mapfile, MAPFILESIZE);
		WritePrivateProfileString("Fuse", "MapFile", mapfile, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_TILEDATAFILE), tilefile, MAPFILESIZE);
		WritePrivateProfileString("Fuse", "TiledataFile", tilefile, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_STATICSINDEXFILE), staidxfile, MAPFILESIZE);
		WritePrivateProfileString("Fuse", "StaticsIndexFile", staidxfile, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_STATICSFILE), staticsfile, MAPFILESIZE);
		WritePrivateProfileString("Fuse", "StaticsFile", staticsfile, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_VERDATAFILE), verfile, MAPFILESIZE);
		WritePrivateProfileString("Fuse", "VerdataFile", verfile, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_MULTIINDEXFILE), multiidxfile, MAPFILESIZE);
		WritePrivateProfileString("Fuse", "MultiIndexFile", multiidxfile, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_MULTIFILE), multifile, MAPFILESIZE);
		WritePrivateProfileString("Fuse", "MultiFile", multifile, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_SAVEINTERVAL), buf, 1000);
		saveinterval = atoi(buf);
		WritePrivateProfileString("Fuse", "SaveInterval", buf, "fuse.ini");

		Edit_GetText(GetDlgItem(hwnd, IDC_CLIENTKEY), buf, 1000);
		sscanf(buf, "%lu-%lu", &ck1, &ck2);
		WritePrivateProfileString("Fuse", "ClientKey", buf, "fuse.ini");

		if(Button_GetCheck(GetDlgItem(hwnd, IDC_LOGINCHECK)) == BST_UNCHECKED) {
			logincheck = false;
			WritePrivateProfileString("Fuse", "LoginCheck", "false", "fuse.ini");
		}
		else {
			logincheck = true;
			WritePrivateProfileString("Fuse", "LoginCheck", "true", "fuse.ini");
		}

		EndDialog(hwnd, 0);
		break;
	}
}

/*struct Locale {
	char name1[31], name2[31];
	int x, y, z;
} ;

Locale **locales = NULL;
int localecount = 0;*/
void GetLocales(const char *filename)
{
	FILE *ifil;
	char name1[31], name2[31], buf[1000], *bad;
	int x, y, z, n, i;
	Locale **nlocales, **olocales;

	// enter crit sect

	if(localecount > 0) {
		for(i = 0; i < localecount; i++) {
			delete locales[i];
		}

		delete []locales;
		localecount = 0;
	}

	ifil = fopen(filename, "rt");

	if(ifil == NULL)
		return;

	while(fgets(buf, 1000, ifil) != NULL) {
		bad = strrchr(buf, 0x0a);

		if(bad != NULL)
			*bad = 0;

		bad = strrchr(buf, 0x0d);

		if(bad != NULL)
			*bad = 0;

		if(strlen(buf) == 0)
			continue;

		CopyBuf(name1, 31, buf);

		fgets(buf, 1000, ifil);

		bad = strrchr(buf, 0x0a);

		if(bad != NULL)
			*bad = 0;

		bad = strrchr(buf, 0x0d);

		if(bad != NULL)
			*bad = 0;

		CopyBuf(name2, 31, buf);

		fgets(buf, 1000, ifil);
		n = sscanf(buf, "%x %x %x", &x, &y, &z);

		if(n < 3) {
			z = -1;
		}

		nlocales = new Locale *[localecount + 1];

		for(i = 0; i < localecount; i++) {
			nlocales[i] = locales[i];
		}

		nlocales[localecount] = new Locale(name1, name2, x, y, z);

		olocales = locales;
		locales = nlocales;
		localecount += 1;

		if(olocales != NULL) {
			delete []olocales;
		}
	}
}

void GetServers(const char *filename)
{
	FILE *ifil;
	char name[100], addr1[100], addr2[100], buf[1000];
	int port, zone, iaddr1, iaddr2, i;
	serverdef **nservers, **oservers;
	struct hostent *he;

	ifil = fopen(filename, "rt");

	if(ifil == NULL)
		return;

	while(fgets(buf, 1000, ifil) != NULL) {
		if(sscanf(buf, "%s %s %d %s %d", name, addr1, &port, addr2, &zone) == 5) {
			nservers = new serverdef *[servercount + 1];

			for(i = 0; i < servercount; i++) {
				nservers[i] = servers[i];
			}

			if((iaddr1 = inet_addr(addr1)) == INADDR_NONE) {
				if((he = gethostbyname(addr1)) != NULL) {
					iaddr1 = *((int *)(he -> h_addr_list[0]));
				}
				else {
					iaddr1 = INADDR_NONE;
				}
			}

			if((iaddr2 = inet_addr(addr2)) == INADDR_NONE) {
				if((he = gethostbyname(addr2)) != NULL) {
					iaddr2 = *((int *)(he -> h_addr_list[0]));
				}
				else {
					iaddr2 = INADDR_NONE;
				}
			}

			nservers[servercount] = new serverdef(name, iaddr1, port, iaddr2, zone);

			oservers = servers;
			servers = nservers;
			servercount += 1;

			if(oservers != NULL) {
				delete []oservers;
			}
		}
	}
}

BOOL CALLBACK DlgSaveIntervalProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {
		HANDLE_MSG(hwnd, WM_INITDIALOG, OnInitDialogSaveInterval);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommandSaveInterval);
	}

	return FALSE;
}

BOOL OnInitDialogSaveInterval(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
	char buf[1000];

	sprintf(buf, "%d", saveinterval);

	Edit_SetText(GetDlgItem(hwnd, IDC_CURRENTSI), buf);
	Edit_SetText(GetDlgItem(hwnd, IDC_NEWSI), buf);

	SetFocus(GetDlgItem(hwnd, IDC_NEWSI));
	Edit_SetSel(GetDlgItem(hwnd, IDC_NEWSI), 0, -1);

	//SetFocus(GetDlgItem(hwnd, IDOK));
	return FALSE;
}

// id - id of menu item, control, or accelerator
// hwndCtl - identifies control sending message (if not control, this is NULL)
// codeNotify - notification code from control (if menu, 0; if accelerator, 1)
void OnCommandSaveInterval(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	char buf[1000];

	switch(id) {
	case IDCANCEL:
		EndDialog(hwnd, 1);
		break;

	case IDOK:
		Edit_GetText(GetDlgItem(hwnd, IDC_NEWSI), buf, 1000);
		SetSaveInterval(atoi(buf));
		EndDialog(hwnd, 0);
		break;
	}
}

void DisplaySystemTime(void)
{
	SYSTEMTIME st;
	TIME_ZONE_INFORMATION tzi;
	char buf[1000], buf1[1000];
	DWORD result;

	GetLocalTime(&st);

	if(st.wHour < 12) {
		if(st.wHour == 0) {
			sprintf(buf1, "12:%02u AM", st.wMinute);
		}
		else {
			sprintf(buf1, "%u:%02u AM", st.wHour, st.wMinute);
		}
	}
	else {
		if(st.wHour == 12) {
			sprintf(buf1, "12:%02u PM", st.wMinute);
		}
		else {
			sprintf(buf1, "%u:%02u PM", st.wHour - 12, st.wMinute);
		}
	}

	result = GetTimeZoneInformation(&tzi);

	 // for now, force no display of time zone info
	result = TIME_ZONE_ID_UNKNOWN;

	if(result == TIME_ZONE_ID_STANDARD) {
		sprintf(buf, "%s %S", buf1, tzi.StandardName);
	}
	else if(result == TIME_ZONE_ID_DAYLIGHT) {
		sprintf(buf, "%s %S", buf1, tzi.DaylightName);
	}
	else {
		sprintf(buf, "%s", buf1);
	}

	Static_SetText(GetDlgItem(hwndMain, IDC_SYSTEMTIME), buf);
}

// for now, guarantee that the time will be no larger than 30 chars
// this is true by virute of the way it is currently displayed
FUSELIBFNDEF void GetFUSETime(char *buf)
{
	Static_GetText(GetDlgItem(hwndMain, IDC_SYSTEMTIME), buf, 30);
}

// for now, guarantee that the time will be no larger than 30 chars
// this is true by virute of the way it is currently displayed
FUSELIBFNDEF void GetSaveTime(char *buf)
{
	Static_GetText(GetDlgItem(hwndMain, IDC_LASTSAVE), buf, 30);
}

FUSELIBFNDEF int GetSaveInterval(void)
{
	return saveinterval;
}

FUSELIBFNDEF void SetLoginCheck(int check)
{
	//logincheck = check;
	if(check == 0)
		logincheck = false;
	else
		logincheck = true;
}

FUSELIBFNDEF int GetLoginCheck(void)
{
	//return logincheck;
	int state;

	if(logincheck)
		state = 1;
	else
		state = 0;

	return state;
}

FUSELIBFNDEF int RefreshAccounts(void)
{
	GetAccounts("accounts.txt");
	return 0;
}

FUSELIBFNDEF void RefreshLocales(void)
{
	GetLocales("locales.txt");
}

FUSELIBFNDEF void RefreshServerList(void)
{
	GetServers("servers.txt");
}
