// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

#ifndef _FUSELIB_
#define _FUSELIB_

// exports
#define FUSELIBDEF __declspec(dllexport)
#define FUSELIBFNDEF extern "C" FUSELIBDEF

// imports
#define FUSELIB __declspec(dllimport)
#define FUSELIBFN extern "C" FUSELIB

// property types
#define TYPE_INT 1
#define TYPE_SHORT 2
#define TYPE_BYTE 3
#define TYPE_STRING 4
#define TYPE_ASHORT 5
#define TYPE_IDLIST 6

#define STRSIZE 30
#define MAPX 6144
#define MAPY 4096
#define MAPVIEWRANGE 18

// tile flags
#define TILE_WATER      0x80000000
#define TILE_BLOCKABLE  0x40000000
#define TILE_FLOORTILE  0x01000000
#define TILE_SINGULARV  0x00800000
#define TILE_SINGULARC  0x00400000
#define TILE_STACKABLE  0x00080000
#define TILE_CLIMABLE   0x00040000
#define TILE_STANDABLE  0x00020000
#define TILE_LIGHTABLE  0x00008000
#define TILE_EQUIPPABLE 0x00004000
#define TILE_CONTAINER  0x00002000
#define TILE_FADABLE    0x00000200
#define TILE_PASSABLE   0x00000100
#define TILE_DIRECTION  0x000000C0
#define TILE_DIRNORTH   0x00000000
#define TILE_DIREAST    0x00000040
#define TILE_DIRWEST    0x00000080
#define TILE_DIRSOUTH   0x000000C0
#define TILE_DOORWAY    0x00000020
#define TILE_ANIMATED   0x00000001

// tile data offsets
#define TILE_WEIGHT 0
#define TILE_EQUIP  1
#define TILE_HEIGHT 12

// server -> client messages
#define FUSEMSG_PLAYERSTATUS1	0x00010001
#define FUSEMSG_NPCSTATUS1		0x00010002
#define FUSEMSG_ITEMPLACE1		0x00020001
#define FUSEMSG_PLAYERLOGIN1	0x00030001
#define FUSEMSG_CLIENTBARK1		0x00040001
#define FUSEMSG_OBJECTREMOVE1	0x00050001
#define FUSEMSG_PLAYERPLACE1	0x00060001
#define FUSEMSG_PLAYERRESET1	0x00070001
#define FUSEMSG_PLAYERSET1		0x00080001
#define FUSEMSG_ITEMDRAG1		0x00090001
#define FUSEMSG_ITEMGUMP1		0x000a0001
#define FUSEMSG_ITEMCONTAIN1	0x000b0001
#define FUSEMSG_PICKUPCANCEL1	0x000c0001
#define FUSEMSG_DEATHOPTION1	0x000d0001
#define FUSEMSG_ITEMEQUIP1		0x000e0001
#define FUSEMSG_CLIENTFREEZE1	0x000f0001
#define FUSEMSG_PLAYERSKILLS1	0x00100001
#define FUSEMSG_CONTAINERSETUP1	0x00110001
#define FUSEMSG_NPCLIGHT1		0x00120001
#define FUSEMSG_GLOBALLIGHT1	0x00130001
#define FUSEMSG_SOUND1			0x00140001
#define FUSEMSG_TARGET1			0x00150001
#define FUSEMSG_MUSIC1			0x00160001
#define FUSEMSG_NPCANIMATE1		0x00170001
#define FUSEMSG_EFFECT1			0x00180001
#define FUSEMSG_PLAYERMODE1		0x00190001
#define FUSEMSG_PLAYERMODE2		0x00190002
#define FUSEMSG_NPCSET1			0x001a0001
#define FUSEMSG_NPCSETUP1		0x001b0001
#define FUSEMSG_NPCPAPERDOLL1	0x001c0001
#define FUSEMSG_CORPSESETUP1	0x001d0001
#define FUSEMSG_DYEOPTION1		0x001e0001
#define FUSEMSG_TIPS1			0x001f0001
#define FUSEMSG_UPDATEHP1		0x00200001
#define FUSEMSG_UPDATEMANA1		0x00210001
#define FUSEMSG_UPDATESTAMINA1	0x00220001
#define FUSEMSG_PLAYERKICK1		0x00230001
#define FUSEMSG_BOOKPAGES1		0x00240001
#define FUSEMSG_BOOKCOVER1		0x00250001
#define FUSEMSG_DIALOGBOX1		0x00260001
#define FUSEMSG_WEATHER1		0x00270001
#define FUSEMSG_BROWSER1		0x00280001
#define FUSEMSG_IDLEWARN1		0x00290001
#define FUSEMSG_BUYWINDOW1		0x002a0001
#define FUSEMSG_SELLWINDOW1		0x002b0001
#define FUSEMSG_CPROMPT1		0x002c0001
#define FUSEMSG_SECURETRADE1	0x002d0001
#define FUSEMSG_UNKNOWN1		0x10000001

// client -> server messages
#define FUSECMSG_IDLE1			0x01
#define FUSECMSG_MOVE1			0x02
#define FUSECMSG_SET1			0x03
#define FUSECMSG_MODE1			0x04
#define FUSECMSG_CIN1			0x05
#define FUSECMSG_DATA1			0x06
#define FUSECMSG_DYESELECT1		0x07
#define FUSECMSG_TARGET1		0x08
#define FUSECMSG_PICKUP1		0x09
#define FUSECMSG_DROP1			0x0a
#define FUSECMSG_DBLCLICK1		0x0b
#define FUSECMSG_EQUIP1			0x0c
#define FUSECMSG_CLICK1			0x0d
#define FUSECMSG_ATTACK1		0x0e
#define FUSECMSG_SKILLUSE1		0x0f
#define FUSECMSG_SPELLUSE1		0x10
#define FUSECMSG_ACTION1		0x11
#define FUSECMSG_TIPS1			0x12
#define FUSECMSG_NPCRENAME1		0x13
#define FUSECMSG_GMPAGE1		0x14
#define FUSECMSG_DIALOGBOX1		0x15
#define FUSECMSG_BOOKPAGES1		0x16
#define FUSECMSG_BOOKGETPAGE1	0x17
#define FUSECMSG_PURCHASE1		0x18
#define FUSECMSG_SPELLUSE2		0x19
#define FUSECMSG_SALE1			0x1a
#define FUSECMSG_CPROMPT1		0x1b
#define FUSECMSG_SECURETRADE1	0x1c

inline int isitemid(int id)
{
	return (id & 0xc0000000) == 0x40000000;
}

inline int isnpcid(int id)
{
	return (id & 0xc0000000) == 0x00000000;
}

inline int itemindextoid(int index)
{
	if(index == -1) {
		return -1;
	}
	else {
		return (index + 1) | 0x40000000;
	}
}

inline int npcindextoid(int index)
{
	if(index == -1) {
		return -1;
	}
	else {
		return (index + 1);
	}
}

inline int itemidtoindex(int id)
{
	if(id == -1 || !isitemid(id)) {
		return -1;
	}
	else {
		return (id & 0x3fffffff) - 1;
	}
}

inline int npcidtoindex(int id)
{
	if(id == -1 || !isnpcid(id)) {
		return -1;
	}
	else {
		return (id & 0x3fffffff) - 1;
	}
}

// max valid size is 4095
inline short maketype(int type, int size = 0)
{
	switch(type) {
	case TYPE_INT:
	case TYPE_SHORT:
	case TYPE_BYTE:
		return type;

	case TYPE_STRING:
	case TYPE_ASHORT:
	case TYPE_IDLIST:
		if(size > 0)
			return (size << 4) | (type & 0x000f);
		else
			return 0;

	default:
		return 0;
	}
}

#ifdef _FUSE_ // (EXE)

struct FUSELIBDEF posinfo {
	int x, y, z, topparentid;
} ;

struct FUSELIBDEF fieldlist {
	int count, fields[9];
} ;

struct FUSELIBDEF CommandItem {
	unsigned int type;
	unsigned short word1, word2;
	unsigned char byte1, byte2;
	unsigned long dword1, dword2;
} ;

struct FUSELIBDEF propinfo {
	int offset, type, size;

	propinfo(void)
	{
		offset = type = size = 0;
	}

	propinfo(propinfo &info)
	{
		offset = info.offset;
		type = info.type;
		size = info.size;
	}
} ;

#pragma pack(push, 1)
struct FUSELIBDEF StaticBlock {
	unsigned short model;
	unsigned short x;
	unsigned short y;
	char z;
	unsigned short color;
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIBDEF StaticDescriptor {
	unsigned count;
	StaticBlock items[1];
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIBDEF TileData {
	unsigned flags;
	char data[13];
	char name[20];
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIBDEF LandTileData {
	unsigned flags;
	char data[2];
	char name[20];
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIBDEF MultiData {
	unsigned short model;
	short x;
	short y;
	short z;
	unsigned flag;
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIBDEF MultiDescriptor {
	unsigned count;
	MultiData items[1];
} ;
#pragma pack(pop)

// connect.cpp
FUSELIBFNDEF void SetDPAInput(int);
FUSELIBFNDEF int DPAPrepareMessage(char *, int &, const char *, int);
FUSELIBFNDEF int PrepareMessage(unsigned, char *, int &, const char *);
FUSELIBFNDEF void SendData(int, char *, int);
FUSELIBFNDEF void Globalcast(int, char *, int);
FUSELIBFNDEF void ConnectionPool(int, int, short);
FUSELIBFNDEF void ConnectionUnpool(int);
FUSELIBFNDEF int ConnectionPoolID(int);
FUSELIBFNDEF int ConnectionPoolCount(int);
FUSELIBFNDEF int GetConnectionNPCID(int);
FUSELIBFNDEF int GetMaxConnectionCount(void);

// utils.cpp
FUSELIBFNDEF void CopyBuf(char *, int, const char *);
FUSELIBFNDEF void BufferByte(char *, int, char);
FUSELIBFNDEF void BufferShort(char *, int, short);
FUSELIBFNDEF void BufferInt(char *, int, int);
FUSELIBFNDEF void BufferString(char *, int, const char *, int);
FUSELIBFNDEF char UnBufferByte(const char *, int);
FUSELIBFNDEF short UnBufferShort(const char *, int);
FUSELIBFNDEF int UnBufferInt(const char *, int);
FUSELIBFNDEF void UnBufferString(const char *, int, char *, int);
FUSELIBFNDEF int GetProperty(int, const char *, char *, int, int = -1);
FUSELIBFNDEF int SetProperty(int, const char *, char *, int, int = -1);
FUSELIBFNDEF void GetItemProp(const char *, propinfo &);
FUSELIBFNDEF void GetNPCProp(const char *, propinfo &);
FUSELIBFNDEF void AddItemProp(const char *prop, int type);
FUSELIBFNDEF void AddNPCProp(const char *prop, int type);
FUSELIBFNDEF short propshortarrayget(int id, const propinfo &array, int index);
FUSELIBFNDEF void propshortarrayset(int id, const propinfo &array, int index, short val);
FUSELIBFNDEF void propshortarrayinit(int id, const propinfo &array);
FUSELIBFNDEF int propintget(int id, const propinfo &info);
FUSELIBFNDEF void propintset(int id, const propinfo &info, int val);
FUSELIBFNDEF short propshortget(int id, const propinfo &info);
FUSELIBFNDEF void propshortset(int id, const propinfo &info, short val);
FUSELIBFNDEF char propcharget(int id, const propinfo &info);
FUSELIBFNDEF void propcharset(int id, const propinfo &info, char val);
FUSELIBFNDEF int propboolget(int id, const propinfo &info);
FUSELIBFNDEF void propboolset(int id, const propinfo &info, int val);
FUSELIBFNDEF const char *propstringget(int id, const propinfo &info);
FUSELIBFNDEF void propstringset(int id, const propinfo &info, const char *val);
FUSELIBFNDEF void proplistinit(int id, const propinfo &info);
FUSELIBFNDEF int proplistcount(int id, const propinfo &info);
FUSELIBFNDEF int proplistid(int id, const propinfo &info, int index);
FUSELIBFNDEF int proplistplus(int id, const propinfo &info, int iid);
FUSELIBFNDEF int proplistminus(int id, const propinfo &info, int iid);
FUSELIBFNDEF int proplistfull(int id, const propinfo &info);

// object.cpp
FUSELIBFNDEF int CreateItem(void);
FUSELIBFNDEF int CreateNPC(void);
FUSELIBFNDEF int MoveField(int, const posinfo &, const posinfo &);
FUSELIBFNDEF void DestroyID(int);
FUSELIBFNDEF void GlobalIDPlus(int, const posinfo &);
FUSELIBFNDEF void GlobalIDMinus(int, const posinfo &);
FUSELIBFNDEF int GetField(const posinfo &);
FUSELIBFNDEF void FieldNeighbors(int, fieldlist &);
FUSELIBFNDEF void FieldLists(int, int, fieldlist &, fieldlist &, fieldlist &);
FUSELIBFNDEF void Broadcast(int, char *, int, const posinfo &, int);
FUSELIBFNDEF int NPCCount(void);
FUSELIBFNDEF int ItemCount(void);
FUSELIBFNDEF int CommandAvailable(int);
FUSELIBFNDEF void GetCommand(int, CommandItem *);
FUSELIBFNDEF void AddCommand(int, const CommandItem *);
FUSELIBFNDEF void ClearCommands(int);
FUSELIBFNDEF void ReadyConnection(int);
FUSELIBFNDEF const int *GetFieldIDList(int);
FUSELIBFNDEF const int *GetFieldPCList(int);
FUSELIBFNDEF int GetFieldIDCount(int);
FUSELIBFNDEF int GetFieldPCCount(int);
FUSELIBFNDEF void CommandCycle(int);
FUSELIBFNDEF int IDActive(int);
FUSELIBFNDEF int NPCConnection(int);
FUSELIBFNDEF void SetAllocationSize(int);

// mapcache.cpp
FUSELIBFNDEF char GetMapZ(unsigned short, unsigned short);
FUSELIBFNDEF unsigned short GetMapTile(unsigned short, unsigned short);
FUSELIBFNDEF void GetTileName(unsigned short, char *);
FUSELIBFNDEF unsigned GetTileFlags(unsigned short);
FUSELIBFNDEF unsigned char GetTileWeight(unsigned short);
FUSELIBFNDEF unsigned char GetTileEquip(unsigned short);
FUSELIBFNDEF unsigned char GetTileHeight(unsigned short);
FUSELIBFNDEF void GetTileData(unsigned short, TileData *);
FUSELIBFNDEF void GetLandTileName(unsigned short, char *);
FUSELIBFNDEF unsigned GetLandTileFlags(unsigned short);
FUSELIBFNDEF void GetLandTileData(unsigned short, LandTileData *);
FUSELIBFNDEF void LockStatics(void);
FUSELIBFNDEF void UnlockStatics(void);
FUSELIBFNDEF const StaticDescriptor *GetStaticDescriptor(const posinfo &);
FUSELIBFNDEF void LockMultis(void);
FUSELIBFNDEF void UnlockMultis(void);
FUSELIBFNDEF const MultiDescriptor *GetMultiDescriptor(unsigned short);

// entry.cpp
FUSELIBFNDEF int FUSEInterfaceVersion(void);
FUSELIBFNDEF void StartSaveNow(void);
FUSELIBFNDEF void GetFUSETime(char *);
FUSELIBFNDEF void GetSaveTime(char *);
FUSELIBFNDEF void SetSaveInterval(int);
FUSELIBFNDEF int GetSaveInterval(void);
FUSELIBFNDEF void SetLoginCheck(int);
FUSELIBFNDEF int GetLoginCheck(void);
FUSELIBFNDEF int RefreshAccounts(void);
FUSELIBFNDEF void RefreshLocales(void);
FUSELIBFNDEF void RefreshServerList(void);

#else // _FUSE_ (DLL)

struct FUSELIB posinfo {
	int x, y, z, topparentid;
} ;

struct FUSELIB fieldlist {
	int count, fields[9];
} ;

struct FUSELIB CommandItem {
	unsigned char type;
	unsigned short word1, word2;
	unsigned char byte1, byte2;
	unsigned dword1, dword2;
} ;

struct FUSELIB propinfo {
	int offset, type, size;

	propinfo(void)
	{
		offset = type = size = 0;
	}

	propinfo(propinfo &info)
	{
		offset = info.offset;
		type = info.type;
		size = info.size;
	}
} ;

#pragma pack(push, 1)
struct FUSELIB StaticBlock {
	unsigned short model;
	unsigned short x;
	unsigned short y;
	char z;
	unsigned short color;
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIB StaticDescriptor {
	unsigned count;
	StaticBlock items[1];
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIB TileData {
	unsigned flags;
	char data[13];
	char name[20];
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIB LandTileData {
	unsigned flags;
	char data[2];
	char name[20];
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIB MultiData {
	unsigned short model;
	short x;
	short y;
	short z;
	unsigned flag;
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct FUSELIB MultiDescriptor {
	unsigned count;
	MultiData items[1];
} ;
#pragma pack(pop)

// connect.cpp
FUSELIBFN void SetDPAInput(int);
FUSELIBFN int DPAPrepareMessage(char *, int &, const char *, int);
FUSELIBFN int PrepareMessage(unsigned, char *, int &, const char *);
FUSELIBFN void SendData(int, char *, int);
FUSELIBFN void Globalcast(int, char *, int);
FUSELIBFN void ConnectionPool(int, int, short);
FUSELIBFN void ConnectionUnpool(int);
FUSELIBFN int ConnectionPoolID(int);
FUSELIBFN int ConnectionPoolCount(int);
FUSELIBFN int GetConnectionNPCID(int);
FUSELIBFN int GetMaxConnectionCount(void);

// utils.cpp
FUSELIBFN void CopyBuf(char *, int, const char *);
FUSELIBFN void BufferByte(char *, int, char);
FUSELIBFN void BufferShort(char *, int, short);
FUSELIBFN void BufferInt(char *, int, int);
FUSELIBFN void BufferString(char *, int, const char *, int);
FUSELIBFN char UnBufferByte(const char *, int);
FUSELIBFN short UnBufferShort(const char *, int);
FUSELIBFN int UnBufferInt(const char *, int);
FUSELIBFN void UnBufferString(const char *, int, char *, int);
FUSELIBFN int GetProperty(int, const char *, char *, int, int = -1);
FUSELIBFN int SetProperty(int, const char *, char *, int, int = -1);
FUSELIBFN void GetItemProp(const char *, propinfo &);
FUSELIBFN void GetNPCProp(const char *, propinfo &);
FUSELIBFN void AddItemProp(const char *prop, int type);
FUSELIBFN void AddNPCProp(const char *prop, int type);
FUSELIBFN short propshortarrayget(int id, const propinfo &array, int index);
FUSELIBFN void propshortarrayset(int id, const propinfo &array, int index, short val);
FUSELIBFN void propshortarrayinit(int id, const propinfo &array);
FUSELIBFN int propintget(int id, const propinfo &info);
FUSELIBFN void propintset(int id, const propinfo &info, int val);
FUSELIBFN short propshortget(int id, const propinfo &info);
FUSELIBFN void propshortset(int id, const propinfo &info, short val);
FUSELIBFN char propcharget(int id, const propinfo &info);
FUSELIBFN void propcharset(int id, const propinfo &info, char val);
FUSELIBFN int propboolget(int id, const propinfo &info);
FUSELIBFN void propboolset(int id, const propinfo &info, int val);
FUSELIBFN const char *propstringget(int id, const propinfo &info);
FUSELIBFN void propstringset(int id, const propinfo &info, const char *val);
FUSELIBFN void proplistinit(int id, const propinfo &info);
FUSELIBFN int proplistcount(int id, const propinfo &info);
FUSELIBFN int proplistid(int id, const propinfo &info, int index);
FUSELIBFN int proplistplus(int id, const propinfo &info, int iid);
FUSELIBFN int proplistminus(int id, const propinfo &info, int iid);
FUSELIBFN int proplistfull(int id, const propinfo &info);

// object.cpp
FUSELIBFN int CreateItem(void);
FUSELIBFN int CreateNPC(void);
FUSELIBFN int MoveField(int, const posinfo &, const posinfo &);
FUSELIBFN void DestroyID(int);
FUSELIBFN void GlobalIDPlus(int, const posinfo &);
FUSELIBFN void GlobalIDMinus(int, const posinfo &);
FUSELIBFN int GetField(const posinfo &);
FUSELIBFN void FieldNeighbors(int, fieldlist &);
FUSELIBFN void FieldLists(int, int, fieldlist &, fieldlist &, fieldlist &);
FUSELIBFN void Broadcast(int, char *, int, const posinfo &, int);
FUSELIBFN int NPCCount(void);
FUSELIBFN int ItemCount(void);
FUSELIBFN int CommandAvailable(int);
FUSELIBFN void GetCommand(int, CommandItem *);
FUSELIBFN void AddCommand(int, const CommandItem *);
FUSELIBFN void ClearCommands(int);
FUSELIBFN void ReadyConnection(int);
FUSELIBFN const int *GetFieldIDList(int);
FUSELIBFN const int *GetFieldPCList(int);
FUSELIBFN int GetFieldIDCount(int);
FUSELIBFN int GetFieldPCCount(int);
FUSELIBFN void CommandCycle(int);
FUSELIBFN int IDActive(int);
FUSELIBFN int NPCConnection(int);
FUSELIBFN void SetAllocationSize(int);

// mapcache.cpp
FUSELIBFN char GetMapZ(unsigned short, unsigned short);
FUSELIBFN unsigned short GetMapTile(unsigned short, unsigned short);
FUSELIBFN void GetTileName(unsigned short, char *);
FUSELIBFN unsigned GetTileFlags(unsigned short);
FUSELIBFN unsigned char GetTileWeight(unsigned short);
FUSELIBFN unsigned char GetTileEquip(unsigned short);
FUSELIBFN unsigned char GetTileHeight(unsigned short);
FUSELIBFN void GetTileData(unsigned short, TileData *);
FUSELIBFN void GetLandTileName(unsigned short, char *);
FUSELIBFN unsigned GetLandTileFlags(unsigned short);
FUSELIBFN void GetLandTileData(unsigned short, LandTileData *);
FUSELIBFN void LockStatics(void);
FUSELIBFN void UnlockStatics(void);
FUSELIBFN const StaticDescriptor *GetStaticDescriptor(const posinfo &);
FUSELIBFN void LockMultis(void);
FUSELIBFN void UnlockMultis(void);
FUSELIBFN const MultiDescriptor *GetMultiDescriptor(unsigned short);

// entry.cpp
FUSELIBFN int FUSEInterfaceVersion(void);
FUSELIBFN void StartSaveNow(void);
FUSELIBFN void GetFUSETime(char *);
FUSELIBFN void GetSaveTime(char *);
FUSELIBFN void SetSaveInterval(int);
FUSELIBFN int GetSaveInterval(void);
FUSELIBFN void SetLoginCheck(int);
FUSELIBFN int GetLoginCheck(void);
FUSELIBFN int RefreshAccounts(void);
FUSELIBFN void RefreshLocales(void);
FUSELIBFN void RefreshServerList(void);

// dll
FUSELIBFNDEF void cycle(void);
FUSELIBFNDEF void CreatePlayer(int, const char *);
FUSELIBFNDEF void GetIDPoint(int, posinfo &);
FUSELIBFNDEF void NPCLoaded(int);
FUSELIBFNDEF void ItemLoaded(int);
FUSELIBFNDEF void GetPCNameAndPass(int, char *, char *);
FUSELIBFNDEF void PCLogin(int);
FUSELIBFNDEF void PCLogout(int);
FUSELIBFNDEF void DPAProcessInput(int, unsigned short, const char *);
FUSELIBFNDEF void ProcessInput(int, unsigned short, const char *);
FUSELIBFNDEF int DeletePC(int, int);

#endif // _FUSE_

#endif // _FUSELIB_
