#include "pcheader.h"

// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

#include "server.h"

//#define MAPFIELDDIM MAPBLOCKDIM
#define MAPFIELDDIM 64

#define GLOBAL_X (MAPX / MAPFIELDDIM)
#define GLOBAL_Y (MAPY / MAPFIELDDIM)

// access these arrays through functions so
// that the programming interface gives access
// to them, and so that we can make sure that
// values out of range are dealth with properly
idlist globals[GLOBAL_X][GLOBAL_Y];

// this will be a list of connected pcs
// these npcs must also appear iin the globals lists
idlist pcs[GLOBAL_X][GLOBAL_Y];

dynamiclist<itemunit> items;
dynamiclist<npcunit> npcs;

// default allocation size is 5
// default allocation max is 5
int allocsize = 5, allocmax = 5;

FUSELIBFNDEF void SetAllocationSize(int size)
{
	allocationsize(size);
}

void allocationsize(int size)
{
	allocsize = size;
}

int allocationsize(void)
{
	return allocsize;
}

void allocationmax(int max)
{
	allocmax = max;
}

int allocationmax(void)
{
	return allocmax;
}

void inittemplates(void)
{
	//propsetup();
	items.init(allocationsize(), allocationmax());
	npcs.init(allocationsize(), allocationmax());
}

/*void makepoint(int id, posinfo &pos)
{
	int pid = id, thisid;

	do {
		thisid = pid;
		pid = parentid(thisid);
	} while(pid != 0);

	pos.topparentid = thisid;

	pos.x = x(thisid);
	pos.y = y(thisid);
	pos.z = z(thisid);
}*/

/*bool idinrange(int id, const posinfo &p, int range)
{
	int dx, dy;

	dx = x(id) - p.x;
	dy = y(id) - p.y;

	if(dx < 0)
		dx = -dx;

	if(dy < 0)
		dy = -dy;

	if(dx < dy)
		return dy <= range;
	else
		return dx <= range;
}*/
bool PointInRange(const posinfo &p1, const posinfo &p2, int range)
{
	int dx, dy;

	dx = p1.x - p2.x;
	dy = p1.y - p2.y;

	if(dx < 0)
		dx = -dx;

	if(dy < 0)
		dy = -dy;

	if(dx < dy)
		return dy <= range;
	else
		return dx <= range;
}

// external functions
FUSELIBFNDEF int NPCCount(void)
{
	return npcs.count();
}

FUSELIBFNDEF int ItemCount(void)
{
	return items.count();
}

FUSELIBFNDEF int CommandAvailable(int id)
{
	dynamic *data;

	data = getdata(id);

	if(data != NULL && data -> commandavailable())
		return true;
	else
		return false;
}

FUSELIBFNDEF void GetCommand(int id, CommandItem *ci)
{
	dynamic *data;

	data = getdata(id);

	if(data != NULL && data -> commandavailable())
		data -> getcommand(ci);
	else
		ci -> type = -1; //0xff;
}

FUSELIBFNDEF void AddCommand(int id, const CommandItem *ci)
{
	dynamic *data;

	data = getdata(id);

	if(data != NULL)
		data -> addcommand(ci);
}

FUSELIBFNDEF void ClearCommands(int id)
{
	dynamic *data;

	data = getdata(id);

	if(data != NULL)
		data -> clearcommands();
}

FUSELIBFNDEF void ReadyConnection(int conn)
{
	readyconnection(conn, true);
}

FUSELIBFNDEF const int *GetFieldIDList(int field)
{
	short dx, dy;

	dx = MakeFieldX(field);
	dy = MakeFieldY(field);

	if(dx == -1 || dy == -1)
		return NULL;
	else
		return globals[dx][dy].getids();
}

FUSELIBFNDEF const int *GetFieldPCList(int field)
{
	short dx, dy;

	dx = MakeFieldX(field);
	dy = MakeFieldY(field);

	if(dx == -1 || dy == -1)
		return NULL;
	else
		return pcs[dx][dy].getids();
}

FUSELIBFNDEF int GetFieldIDCount(int field)
{
	short dx, dy;

	dx = MakeFieldX(field);
	dy = MakeFieldY(field);

	if(dx == -1 || dy == -1)
		return 0;
	else
		return globals[dx][dy].getcount();
}

FUSELIBFNDEF int GetFieldPCCount(int field)
{
	short dx, dy;

	dx = MakeFieldX(field);
	dy = MakeFieldY(field);

	if(dx == -1 || dy == -1)
		return 0;
	else
		return pcs[dx][dy].getcount();
}

FUSELIBFNDEF void CommandCycle(int id)
{
	dynamic *data;

	data = getdata(id);

	if(data != NULL)
		data -> commandcycle();
}

FUSELIBFNDEF int IDActive(int id)
{
	dynamic *data;

	data = getdata(id);

	if(data != NULL)
		return data -> used();
	else
		return false;
}

FUSELIBFNDEF int NPCConnection(int npcid)
{
	dynamic *data;
	int conn;

	if(isnpcid(npcid)) {
		data = getdata(npcid);

		if(data == NULL)
			conn = -1;
		else
			conn = data -> connection();
	}
	else {
		conn = -1;
	}

	return conn;
}

// make default dynamic to return instead of NULL
// default item should be flagged as *unused*
//dynamic defitem;

dynamic *getdata(int id)
{
	dynamic *data;

	if(isnpcid(id)) {
		data = npcs.getdata(npcidtoindex(id));

		if(!data -> used())
			data = NULL;
	}
	else if(isitemid(id)) {
		data = items.getdata(itemidtoindex(id));

		if(!data -> used())
			data = NULL;
	}
	else {
		data = NULL;
	}

	// later we can return default data for
	// non-existent items and npcs
	return data;
}

FUSELIBFNDEF int CreateItem(void)
{
	int id, index;
	dynamic *data;

	index = items.create();
	id = itemindextoid(index);

	data = items.getdata(index);
	data -> used(true, getbackupstate(id));

	return id;
}

/*int newitem(const createdata *cdata)
{
	int id;

	id = CreateItem();

	x(id, cdata -> x);
	y(id, cdata -> y);
	model(id, cdata -> model);
	color(id, cdata -> color);
	z(id, cdata -> z);
	name(id, cdata -> name);
	childlistinit(id);
	parentid(id, 0);
	pooled(id, true);

	equip(id, cdata -> equip);
	flag(id, cdata -> flag);
	extra(id, cdata -> extra);
	flag2(id, 0);
	flag3(id, 0);

	SetProperty(id, "container", "0", 2, -1);

	return id;
}*/

FUSELIBFNDEF int CreateNPC(void)
{
	int id, index;
	dynamic *data;

	index = npcs.create();
	id = npcindextoid(index);

	data = npcs.getdata(index);
	data -> used(true, getbackupstate(id));

	return id;
}

/*int newnpc(const createdata *cdata)
{
	int id, i;

	id = CreateNPC();

	x(id, cdata -> x);
	y(id, cdata -> y);
	model(id, cdata -> model);
	color(id, cdata -> color);
	z(id, cdata -> z);
	name(id, cdata -> name);
	childlistinit(id);
	parentid(id, 0);
	pooled(id, true);

	acct(id, cdata -> acct);
	mymodel(id, cdata -> model);
	mycolor(id, cdata -> color);
	gmcolor(id, 0x21);
	skillsinit(id);
	for(i = 0; i < CREATESKILLS; i++) {
		skills(id, cdata -> skills[i << 1] + 1, (int)(cdata -> skills[(i << 1) + 1]) * 10);
	}
	gender(id, cdata -> gender);
	facing(id, 0);
	show(id, 0);
	pw(id, cdata -> pw);
	rep(id, 0);
	gold(id, 0);
	strength(id, cdata -> strength);
	dexterity(id, cdata -> dexterity);
	intelligence(id, cdata -> intelligence);
	armor(id, 0);
	weight(id, 0);
	totalhits(id, cdata -> totalhits);
	hits(id, cdata -> hits);
	totalstamina(id, cdata -> totalstamina);
	stamina(id, cdata -> stamina);
	totalmana(id, cdata -> totalmana);
	mana(id, cdata -> mana);

	if(cdata -> gender == 0)
		SetProperty(id, "deathcry1", "15a", 2, -1);
	else
		SetProperty(id, "deathcry1", "150", 2, -1);

	return id;
}*/

/*void CreatePlayer(int id, const char *cdata)
{
	int i, childid;
	createdata hair;

	// size of data in cdata is stored as a short at cdata + 0

	// size = UnBufferShort(cdata, 0);

	x(id, UnBufferShort(cdata, 72));
	y(id, UnBufferShort(cdata, 74));
	model(id, UnBufferShort(cdata, 63));
	color(id, UnBufferShort(cdata, 83) | 0x8000);
	z(id, UnBufferByte(cdata, 76));
	name(id, cdata + 2);
	childlistinit(id);
	parentid(id, 0);
	pooled(id, true);

	acct(id, UnBufferInt(cdata, 68));
	mymodel(id, UnBufferShort(cdata, 63));
	mycolor(id, UnBufferShort(cdata, 83) | 0x8000);
	gmcolor(id, (short)0x8021);

	skillsinit(id);
	for(i = 0; i < CREATESKILLS; i++) {
		skills(id, UnBufferByte(cdata, 77 + (i << 1)) + 1, (int)(UnBufferByte(cdata, 78 + (i << 1))) * 10);
	}

	gender(id, UnBufferByte(cdata, 62));
	facing(id, 0);
	show(id, 0);
	pw(id, cdata + 32);
	rep(id, 0);
	gold(id, 0);
	strength(id, UnBufferByte(cdata, 65));
	dexterity(id, UnBufferByte(cdata, 66));
	intelligence(id, UnBufferByte(cdata, 67));
	armor(id, 0);
	weight(id, 0);
	totalhits(id, UnBufferByte(cdata, 65));
	hits(id, UnBufferByte(cdata, 65));
	totalstamina(id, UnBufferByte(cdata, 66));
	stamina(id, UnBufferByte(cdata, 66));
	totalmana(id, UnBufferByte(cdata, 67));
	mana(id, UnBufferByte(cdata, 67));

	if(UnBufferByte(cdata, 62)== 0)
		SetProperty(id, "deathcry1", "15a", 2, -1);
	else
		SetProperty(id, "deathcry1", "150", 2, -1);

	// add basic items here
	if(UnBufferShort(cdata, 85) != 0) {
		hair.reset();
		hair.x = hair.y = 0xffff;
		hair.model = UnBufferShort(cdata, 85);
		hair.color = UnBufferShort(cdata, 87);
		hair.equip = 0x0b;
		childid = newitem(&hair);
		AddChild(id, childid);
	}

	if(UnBufferShort(cdata, 89) != 0) {
		hair.reset();
		hair.x = hair.y = 0xffff;
		hair.model = UnBufferShort(cdata, 89);
		hair.color = UnBufferShort(cdata, 91);
		hair.equip = 0x10;
		childid = newitem(&hair);
		AddChild(id, childid);
	}

	// umm... hehe, use same createdata struct
	hair.reset();
	hair.x = hair.y = 0xffff;
	hair.model = 0x0f51;
	hair.color = 0x0000;
	hair.equip = 0x01;
	CopyBuf(hair.name, STRSIZE, "a dagger");
	childid = newitem(&hair);
	AddChild(id, childid);

	hair.reset();
	hair.x = hair.y = 0xffff;
	hair.model = 0x170f;
	hair.color = 0x0000;
	hair.equip = 0x03;
	CopyBuf(hair.name, STRSIZE, "shoes");
	childid = newitem(&hair);
	AddChild(id, childid);

	hair.reset();
	hair.x = hair.y = 0xffff;
	hair.model = 0x0e75;
	hair.color = 0x0000;
	hair.equip = 0x15;
	CopyBuf(hair.name, STRSIZE, "a backpack");
	childid = newitem(&hair);
	SetProperty(childid, "container", "3c", 3, -1);
	AddChild(id, childid);

	if(gender(id) == 0x00) { // male
		hair.reset();
		hair.x = hair.y = 0xffff;
		hair.model = 0x1517;
		hair.color = 0x0000;
		hair.equip = 0x05;
		CopyBuf(hair.name, STRSIZE, "a shirt");
		childid = newitem(&hair);
		AddChild(id, childid);

		hair.reset();
		hair.x = hair.y = 0xffff;
		hair.model = 0x1539;
		hair.color = 0x0000;
		hair.equip = 0x04;
		CopyBuf(hair.name, STRSIZE, "long pants");
		childid = newitem(&hair);
		AddChild(id, childid);
	}
	else { // the other thing
		hair.reset();
		hair.x = hair.y = 0xffff;
		hair.model = 0x1efd;
		hair.color = 0x0000;
		hair.equip = 0x05;
		CopyBuf(hair.name, STRSIZE, "a fancy shirt");
		childid = newitem(&hair);
		AddChild(id, childid);

		hair.reset();
		hair.x = hair.y = 0xffff;
		hair.model = 0x1516;
		hair.color = 0x0000;
		hair.equip = 0x17;
		CopyBuf(hair.name, STRSIZE, "a skirt");
		childid = newitem(&hair);
		AddChild(id, childid);
	}
}*/

/*void destroy(int id)
{
	int i;
	dynamic *data;

	// recursively destroy all children!
	data = getdata(id);

	for(i = 0; i < data -> childlistcount(); i++) {
		destroy(data -> childlistid(i));
	}

	// DestroyID(id);

	data -> used(false, getbackupstate(id));

	if(isnpcid(id)) {
		npcs.destroy(npcidtoindex(id));
	}
	else if(isitemid(id)) {
		items.destroy(itemidtoindex(id));
	}
}*/

FUSELIBFNDEF void DestroyID(int id)
{
	dynamic *data = getdata(id);

	if(data != NULL) {
		data -> clearcommands();

		data -> used(false, getbackupstate(id));

		if(isnpcid(id)) {
			npcs.destroy(npcidtoindex(id));
		}
		else if(isitemid(id)) {
			items.destroy(itemidtoindex(id));
		}
	}
}

FUSELIBFNDEF void GlobalIDPlus(int id, const posinfo &p)
{
	short dx, dy;
	int field;
	dynamic *data;

	data = getdata(id);

	if(data != NULL) {
		field = GetField(p);

		dx = MakeFieldX(field);
		dy = MakeFieldY(field);

		globals[dx][dy] += id;

		if(data -> connection() != -1) {
			pcs[dx][dy] += id;
		}
	}
}

FUSELIBFNDEF void GlobalIDMinus(int id, const posinfo &p)
{
	short dx, dy;
	int field;
	dynamic *data;

	data = getdata(id);

	if(data != NULL) {
		field = GetField(p);

		dx = MakeFieldX(field);
		dy = MakeFieldY(field);

		globals[dx][dy] -= id;

		if(data -> connection() != -1) {
			pcs[dx][dy] -= id;
		}
	}
}

void npcconnect(int id, int connection)
{
	dynamic *data;

	data = getdata(id);
	data -> connection(connection);
}

/*void npccommand(int id, const CommandItem *cmd)
{
	dynamic *data;

	data = getdata(id);
	data -> addcommand(cmd);
}*/

FUSELIBFNDEF int MoveField(int id, const posinfo &p1, const posinfo &p2)
{
	int field1, field2;
	short dx, dy;
	dynamic *data;

	data = getdata(id);

	if(p1.topparentid == id && p2.topparentid == id) {
		field1 = GetField(p1);
		field2 = GetField(p2);

		if(field1 != field2) {
			dx = MakeFieldX(field1);
			dy = MakeFieldY(field1);

			globals[dx][dy] -= id;

			if(data -> connection() != -1) {
				pcs[dx][dy] -= id;
			}

			dx = MakeFieldX(field2);
			dy = MakeFieldY(field2);

			globals[dx][dy] += id;

			if(data -> connection() != -1) {
				pcs[dx][dy] += id;
			}
		}

		return true;
	}
	else {
		return false;
	}
}
/*
FUSELIBFNDEF bool AddChild(int iparentid, int itemid)
{
	dynamic *data1, *data2;
	short x, y;
	bool val;

	data1 = getdata(itemid);

	// item already has a parent
	if(!data1 -> pooled() && data1 -> parentid() != -1) {
		if(data1 -> parentid() != 0) {
			data2 = getdata(data1 -> parentid());
			childlistminus(data1 -> parentid(), itemid);
		}
		// processing for top-level items
		else {
			x = data1 -> x() / MAPFIELDDIM;
			y = data1 -> y() / MAPFIELDDIM;
			globals[x][y] -= itemid;

			// if the item is connected...
			if(data1 -> connection() != -1) {
				pcs[x][y] -= itemid;
			}
		}
	}

	parentid(itemid, iparentid);
	pooled(itemid, false);

	if(data1 -> parentid() != 0) {
		data2 = getdata(data1 -> parentid());
		val = childlistplus(data1 -> parentid(), itemid);
	}
	// processing for top-level items
	else {
		x = data1 -> x() / MAPFIELDDIM;
		y = data1 -> y() / MAPFIELDDIM;
		globals[x][y] += itemid;

		// if the item is connected...
		if(data1 -> connection() != -1) {
			pcs[x][y] += itemid;
		}

		val = true;
	}

	return val;
}

FUSELIBFNDEF void PoolItem(int itemid)
{
	dynamic *data1, *data2;
	short x, y;

	data1 = getdata(itemid);

	// we cannot repool an item, nor
	// can we pool an item w/o a parent
	if(!data1 -> pooled() && data1 -> parentid() != -1) {
		if(data1 -> parentid() != 0) {
			data2 = getdata(data1 -> parentid());
			childlistminus(data1 -> parentid(), itemid);
		}
		// processing for top-level items
		else {
			x = data1 -> x() / MAPFIELDDIM;
			y = data1 -> y() / MAPFIELDDIM;
			globals[x][y] -= itemid;

			// if the item is connected...
			if(data1 -> connection() != -1) {
				pcs[x][y] -= itemid;
			}
		}

		pooled(itemid, true);
	}
}

FUSELIBFNDEF bool UnpoolItem(int itemid)
{
	dynamic *data1, *data2;
	short x, y;
	bool val;

	data1 = getdata(itemid);

	// item must already be pooled; if it
	// is pooled, it must have a parent
	if(data1 -> pooled()) {
		if(data1 -> parentid() != 0) {
			data2 = getdata(data1 -> parentid());
			val = childlistplus(data1 -> parentid(), itemid);
		}
		// processing for top-level items
		else {
			x = data1 -> x() / MAPFIELDDIM;
			y = data1 -> y() / MAPFIELDDIM;
			globals[x][y] += itemid;

			// if the item is connected...
			if(data1 -> connection() != -1) {
				pcs[x][y] += itemid;
			}

			val = true;
		}

		pooled(itemid, false);
	}
	else {
		val = false;
	}

	return val;
}

FUSELIBFNDEF void UnlinkItem(int id)
{
	int iparentid;
	dynamic *data;
	short x, y;

	data = getdata(id);

	if(data != NULL) {
		iparentid = data -> parentid();

		if(iparentid == 0) {
			x = data -> x() / MAPFIELDDIM;
			y = data -> y() / MAPFIELDDIM;
			globals[x][y] -= id;

			// if the item is connected...
			if(data -> connection() != -1) {
				pcs[x][y] -= id;
			}
		}
		else {
			data = getdata(iparentid);

			if(data != NULL) {
				childlistminus(iparentid, id);
			}
		}

		destroy(id);
	}
}
*/
FUSELIBFNDEF int GetField(const posinfo &p)
{
	return getfield(p.x, p.y);
}

int getfield(int x, int y)
{
	return MakeField(x / MAPFIELDDIM, y / MAPFIELDDIM);
}

int MakeField(int x, int y)
{
	return ((x & 0x0000ffff) << 16) | (y & 0x0000ffff);
}

int MakeFieldX(int field)
{
	return (field >> 16) & 0x0000ffff;
}

int MakeFieldY(int field)
{
	return field & 0x0000ffff;
}

FUSELIBFNDEF void FieldNeighbors(int field, fieldlist &list)
{
	int index;
	short x, y, i, j;

	index = 0;
	x = MakeFieldX(field);
	y = MakeFieldY(field);

	for(i = x - 1; i <= x + 1; i++) {
		if(i >= 0 && i < GLOBAL_X) {
			for(j = y - 1; j <= y + 1; j++) {
				if(j >= 0 && j < GLOBAL_Y) {
					list.fields[index++] = MakeField(i, j);
				}
			}
		}
	}

	list.count = index;
}

bool FieldInList(int field, fieldlist &list)
{
	int i;
	bool val = false;

	for(i = 0; i < list.count; i++) {
		if(field == list.fields[i]) {
			val = true;
			break;
		}
	}

	return val;
}

FUSELIBFNDEF void FieldLists(int oldfield, int newfield, fieldlist &same, fieldlist &add, fieldlist &remove)
{
	fieldlist oldlist, newlist;
	int i;

	FieldNeighbors(oldfield, oldlist);
	FieldNeighbors(newfield, newlist);

	same.count = add.count = remove.count = 0;

	for(i = 0; i < oldlist.count; i++) {
		if(FieldInList(oldlist.fields[i], newlist)) {
			same.fields[same.count++] = oldlist.fields[i];
		}
		else {
			remove.fields[remove.count++] = oldlist.fields[i];
		}
	}

	for(i = 0; i < newlist.count; i++) {
		if(FieldInList(newlist.fields[i], oldlist)) {
			// should already be here
			//same.fields[same.count++] = newlist.fields[i];
		}
		else {
			add.fields[add.count++] = newlist.fields[i];
		}
	}
}

int bitemcount = 0, bnpccount = 0;
int bindex = -1;

int backnpccount(void)
{
	return bnpccount;
}

int backitemcount(void)
{
	return bitemcount;
}

void getbackupdata(backupdata &bdata)
{
	dynamic *data;

	if(bindex == -1) {
		bdata.use = false;
		bdata.size = 0;
		bdata.data = NULL;
	}
	else {
		if((bindex & 0x40000000) == 0x00000000) {
			data = npcs.getdata(bindex & 0x3fffffff);
			data -> getbackupdata(bdata);
		}
		else {
			data = items.getdata(bindex & 0x3fffffff);
			data -> getbackupdata(bdata);
		}
	}
}

// returns true if it is ok to backup changes
// returns false if it is not...  any modifications
// that take place in this instance must set the
// dirty flag for the object
//
// also, we need special stuff for when objects are
// created; objects outside of the range will be backed
// up upon initialization; we need a way to let the
// creation routines know when to do this; perhaps
// use another routine getcreatebackupstate or something
// actually, it is easy to account for new items as well
// as old items. return true if id < currentbackupid
// or >= maxbackupid (for items/npcs)
bool getbackupstate(int id)
{
	int index;

	// always backup if not performing a save
	if(bindex == -1) {
		return true;
	}
	else {
		if(isnpcid(id)) {
			if((bindex & 0x40000000) == 0x00000000) {
				index = npcidtoindex(id);
				if(index < (bindex & 0x3fffffff) || index >= bnpccount)
					return true;
				else
					return false;
			}
			// automatically backup if we have moved on to items
			else {
				return true;
			}
		}
		else if(isitemid(id)) {
			// automatically fail if we haven't moved on to items
			if((bindex & 0x40000000) == 0x00000000) {
				return false;
			}
			else {
				index = itemidtoindex(id);
				if(index < (bindex & 0x3fffffff) || index >= bitemcount)
					return true;
				else
					return false;
			}
		}
		else {
			// umm...
			return false;
		}
	}
}

// this is a backup id (item/npc index or'd 0x40000000 for items)
int nextbackup(void)
{
	dynamic *data;

	// this should cleanup the object at bindex
	// since it is backed up (if dirty flag is
	// set, copy props to bprops, etc)

	// return the id of the next item to backup
	// or -1 if backup is complete
	if(bindex != -1) {
		if((bindex & 0x40000000) == 0x00000000) {
			data = npcs.getdata(bindex & 0x3fffffff);
			data -> refreshbackupdata();

			bindex = nextbackupid(bindex);
		}
		else {
			data = items.getdata(bindex & 0x3fffffff);
			data -> refreshbackupdata();

			bindex = nextbackupid(bindex);
		}
	}

	return bindex;
}

int nextbackupid(int bid)
{
	if(bid == -1) {
		bid = 0x00000000;

		if((bid & 0x3fffffff) >= bnpccount) {
			bid = 0x40000000;

			if((bid & 0x3fffffff) >= bitemcount) {
				bid = -1;
			}
		}
	}
	else {
		if((bid & 0x40000000) == 0x00000000) {
			bid += 1;

			if((bid & 0x3fffffff) >= bnpccount) {
				bid = 0x40000000;

				if((bid & 0x3fffffff) >= bitemcount) {
					bid = -1;
				}
			}
		}
		else {
			bid += 1;

			if((bid & 0x3fffffff) >= bitemcount) {
				bid = -1;
			}
		}
	}

	return bid;
}

bool initbackup(void)
{
	// sets up backup state (takes note of
	// current item/npc counts for backup)
	// since any backup in progress must be
	// completed to prevent data corruption,
	// return false if a backup is already
	// in progress
	if(bindex == -1) {
		bitemcount = items.count();

#ifdef WORLD_BUILD
		bnpccount = 0;
#else
		bnpccount = npcs.count();
#endif

		bindex = nextbackupid(bindex);

		return true;
	}
	else {
		return false;
	}
}

// load npcs/items
bool LoadGameData(FILE *sfil)
{
	int temp, i, count1, count2;
	dynamic *data;
	char c, *stuff;

	if(fread(&temp, 4, 1, sfil) != 1) {
		return false;
	}

	count1 = ntohl(temp);

	if(fread(&temp, 4, 1, sfil) != 1) {
		return false;
	}

	count2 = ntohl(temp);

	if(!npcs.alloc(count1))
		return false;

	for(i = 0; i < count1; i++) {
		data = npcs.getdata(i);

		if(fread(&c, 1, 1, sfil) != 1) {
			return false;
		}

		if(c == 0) {
			data -> used(false, false);
			npcs.destroy(i);
		}
		else {
			data -> used(true, false);
			stuff = data -> propptr();

			if(fread(stuff, npcpropsize(), 1, sfil) != 1) {
				return false;
			}

			data -> refreshbackupdata();

			NPCLoaded(i);

			/*// automatically pool players on startup
			if(data -> acct() != -1) {
				PoolItem(npcindextoid(i));
			}
			// pooled npcs are fine, because they represent logged out players
			else if(data -> parentid() == 0 && !data -> pooled()) {
				AddChild(0, npcindextoid(i));
			}*/
		}
	}

	if(!items.alloc(count2))
		return false;

	for(i = 0; i < count2; i++) {
		data = items.getdata(i);

		if(fread(&c, 1, 1, sfil) != 1) {
			return false;
		}

		if(c == 0) {
			data -> used(false, false);
			items.destroy(i);
		}
		else {
			data -> used(true, false);
			stuff = data -> propptr();

			if(fread(stuff, itempropsize(), 1, sfil) != 1) {
				return false;
			}

			data -> refreshbackupdata();

			ItemLoaded(i);

			/*// for pooled items, we must unpool them
			if(data -> pooled()) {
				UnpoolItem(itemindextoid(i));
			}
			else if(data -> parentid() == 0) {
				AddChild(0, itemindextoid(i));
			}*/
		}
	}

	return true;
}

FUSELIBFNDEF void Broadcast(int id, char *buf, int size, const posinfo &p, int range)
{
	int dx, dy, j, k, field;
	const int *ids;
	fieldlist list;
	dynamic *data;
	posinfo idp;

	field = GetField(p);
	FieldNeighbors(field, list);

	for(j = 0; j < list.count; j++) {
		dx = MakeFieldX(list.fields[j]);
		dy = MakeFieldY(list.fields[j]);

		ids = pcs[dx][dy].getids();

		for(k = 0; k < pcs[dx][dy].getcount(); k++) {
			data = getdata(ids[k]);

			GetIDPoint(ids[k], idp);

			// don't show the npc to its own connection
			if(id != data -> connection() && PointInRange(idp, p, range)) {
				SendData(data -> connection(), buf, size);
			}
		}
	}
}
