// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

class item : public dynamic {
public:
	item(void)
	{
		pt = itemproptemplate();
	}
} ;

class npc : public dynamic {
public:
	npc(void)
	{
		pt = npcproptemplate();
	}
} ;

struct itemunit {
	item *units;
	char *props;
	char *bprops;
	int *freelist;

public:
	itemunit(void)
	{
		units = NULL;
		bprops = props = NULL;
		freelist = NULL;
	}

	itemunit(const itemunit &unit)
	{
		units = unit.units;
		props = unit.props;
		bprops = unit.bprops;
		freelist = unit.freelist;
	}

	void allocate(int size)
	{
		int i;

		units = new item[size];
		props = new char[itempropsize() * size];
		bprops = new char[itempropsize() * size];
		freelist = new int[size];

		for(i = 0; i < size; i++) {
			units[i].setprops(props + itempropsize() * i, bprops + itempropsize() * i);
		}
	}

	void deallocate(void)
	{
		delete []units;
		delete []props;
		delete []bprops;
		delete []freelist;
	}
} ;

struct npcunit {
	npc *units;
	char *props;
	char *bprops;
	int *freelist;

public:
	npcunit(void)
	{
		units = NULL;
		bprops = props = NULL;
		freelist = NULL;
	}

	npcunit(const npcunit &unit)
	{
		units = unit.units;
		props = unit.props;
		bprops = unit.props;
		freelist = unit.freelist;
	}

	void allocate(int size)
	{
		int i;

		units = new npc[size];
		props = new char[size * npcpropsize()];
		bprops = new char[size * npcpropsize()];
		freelist = new int[size];

		for(i = 0; i < size; i++) {
			units[i].setprops(props + npcpropsize() * i, bprops + npcpropsize() * i);
		}
	}

	void deallocate(void)
	{
		delete []units;
		delete []props;
		delete []bprops;
		delete []freelist;
	}
} ;

// at the list level, have default dynamiclists
// for items and npcs; if non-existent dynamiclists
// are referenced, return pointers to the default
// values
template <class unit> class dynamiclist {
	unit *units;
	// size is the exponent of a power of 2
	int unitcount, unitsize, unitmax, unitmask;
	int freelist; // list is linked through next field

public:
	dynamiclist(void)
	{
		units = NULL;
		unitcount = 0;
		unitsize = 0;
		unitmax = 0;
		unitmask = 0;
		freelist = -1;
	}

	void init(int size, int max)
	{
		// we *really* don't want to change anything
		// if we have started using the system
		if(unitcount == 0) {
			unitsize = size;
			unitmax = max;
			unitmask = (1 << unitsize) - 1;
		}
	}

	int count(void) const
	{
		return unitcount << unitsize;
	}

	// necessary? probably
	dynamic *getdata(int index)
	{
		int unitindex = index >> unitsize;
		int itemindex = index & unitmask;

		if(unitindex < unitcount) {
			return units[unitindex].units + itemindex;
		}
		else {
			return NULL;
		}
	}

	int create(void)
	{
		int index, i, first;
		unit *newunits;

		if(freelist == -1) {
			if(1/*unitcount < unitmax*/) {
				newunits = new unit[unitcount + 1];

				for(i = 0; i < unitcount; i++) {
					newunits[i] = units[i];
				}

				units = newunits;

				units[unitcount].allocate(1 << unitsize);

				first = unitcount << unitsize;

				for(i = 0; i < (1 << unitsize); i++) {
					if(i == unitmask /*(1 << unitsize) - 1*/) {
						units[unitcount].freelist[i] = freelist;
					}
					else {
						units[unitcount].freelist[i] = first + i + 1;
					}
				}

				freelist = first;

				unitcount += 1;
			}
		}

		if(freelist != -1) {
			index = freelist;
			freelist = units[index >> unitsize].freelist[index & unitmask];
			units[index >> unitsize].freelist[index & unitmask] = -1;

			return index;
		}
		else {
			return -1;
		}
	}

	// this can only be called once (at the beginning)
	bool alloc(int num)
	{
		int i, j, extra, first;

		if(unitcount == 0) {
			unitcount = (num + unitmask) >> unitsize;
			extra = (unitcount << unitsize) - num;

			if(1/*unitcount < unitmax*/) {
				units = new unit[unitcount];

				for(i = 0; i < (unitcount /*<< unitsize*/); i++) {
					units[i].allocate(1 << unitsize);

					for(j = 0; j < (1 << unitsize); j++) {
						units[i].freelist[j] = -1;
					}
				}

				if(extra > 0) {
					first = (1 << unitsize) - extra;
					freelist = ((unitcount - 1) << unitsize) + first;

					for(i = first; i < first + extra; i++) {
						if(i == extra - 1) {
							units[unitcount - 1].freelist[i] = -1;
						}
						else {
							units[unitcount - 1].freelist[i] = ((unitcount - 1) << unitsize) + i + 1;
 						}
					}
				}
				else {
					freelist = -1;
				}
			}
			else {
				return false;
			}

			return true;
		}
		else {
			return false;
		}
	}

	void destroy(int index)
	{
		units[index >> unitsize].freelist[index & unitmask] = freelist;
		freelist = index;
	}
} ;

//void makepoint(int, posinfo &); // dll

void allocationsize(int);
int allocationsize(void);
void inittemplates(void);
dynamic *getdata(int);

//int newitem(const createdata *); // dll
//int newnpc(const createdata *); // dll
//void destroy(int);

void npcconnect(int, int);
//void npccommand(int, const CommandItem *);

int getfield(int, int);
int MakeField(int, int);
int MakeFieldX(int);
int MakeFieldY(int);
bool FieldInList(int, fieldlist &);
int backnpccount(void);
int backitemcount(void);
void getbackupdata(backupdata &);
bool getbackupstate(int);
int nextbackup(void);
int nextbackupid(int);
bool initbackup(void);
bool LoadGameData(FILE *);
bool PointInRange(const posinfo &, const posinfo &, int);
