// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

#define _FUSE_

#define IDLISTSIZE 32
#define COMMANDQUEUE 50

// remember, when using world build, make all characters
// with no hair/beard
//#define WORLD_BUILD

// currently, there are 46 skills
//#define SKILL_COUNT 46

#include "fuselib.h"

// empty: first == last
// full: last + 1 == cycle
// reset item: 0xff
class CommandQueue {
	CommandItem commands[COMMANDQUEUE];
	int first, last, cycle;

public:
	CommandQueue(void)
	{
		first = last = cycle = 0;
	}

	// if the list is full, set the last item to the reset item
	void AddCommand(const CommandItem *cmd)
	{
		if(((last + 1) % COMMANDQUEUE) == cycle) {
			commands[last].type = -1;
		}
		else {
			commands[last] = *cmd;
			last = (last + 1) % COMMANDQUEUE;
		}
	}

	void ClearCommands(void)
	{
		first = last = cycle = 0;
	}

	// if the list is empty, return the reset item
	void GetCommand(CommandItem *cmd)
	{
		if(first != last) {
			*cmd = commands[first];
			first = (first + 1) % COMMANDQUEUE;
		}
		else {
			cmd -> type = -1;
		}
	}

	bool HasCommand(void) const
	{
		return first != last;
	}

	void Cycle(void)
	{
		cycle = first;
	}
} ;

// map constants
#define MAPBLOCKDIM (8)
#define MAPPOSBYTES (3)
#define MAPBLOCKHDR (4)
#define MAPBLOCK (MAPBLOCKHDR + MAPBLOCKDIM * MAPBLOCKDIM * MAPPOSBYTES)
#define MAPYSTRIP ((MAPY / MAPBLOCKDIM) * MAPBLOCK)

// mapcache.cpp
bool MapSetup(const char *, const char *, const char *, const char *, const char *, const char *, const char *);
void MapRelease(void);

// encode.cpp
int PutMessage(const char *, int, char *);
int GetMessage(const char *, int, int, char *, int &);

// connect.cpp
bool GetDPAInput(void);

#include "utils.h"

#include "account.h"
#include "connect.h"
#include "object.h"

// dll functions
typedef void (*CYCLE)(void);
typedef void (*CREATEPLAYER)(int, const char *);
typedef void (*GETIDPOINT)(int, posinfo &);
typedef void (*OBJECTLOADED)(int);
typedef void (*GETPCNAMEANDPASS)(int, char *, char *);
typedef void (*PCLOG)(int);
typedef void (*PROCESSINPUT)(int, unsigned short, const char *);
typedef int (*DELETEPC)(int, int);

void cycle(void);
void CreatePlayer(int, const char *);
void GetIDPoint(int, posinfo &);
void NPCLoaded(int);
void ItemLoaded(int);
void GetPCNameAndPass(int, char *, char *);
void PCLogin(int);
void PCLogout(int);
void DPAProcessInput(int, unsigned short, const char *);
void ProcessInput(int, unsigned short, const char *);
int DeletePC(int, int);
int UsesVerdata(void);
