#include "pcheader.h"

// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

#include "server.h"

proptemplate itemprops, npcprops;

// load ptemp and itemp from sfil
bool LoadGameTemplateInfo(FILE *sfil, proptemplate &ptemp, proptemplate &itemp)
{
	unsigned short type;
	int i, temp;
	char name[STRSIZE];

	if(fread(&temp, 4, 1, sfil) != 1) {
		return false;
	}

	temp = ntohl(temp);

	// load ptemp
	for(i = 0; i < temp; i++) {
		if(fread(name, STRSIZE, 1, sfil) != 1) {
			return false;
		}

		if(fread(&type, 2, 1, sfil) != 1) {
			return false;
		}

		type = ntohs(type);

		ptemp.add(name, type);
	}

	if(fread(&temp, 4, 1, sfil) != 1) {
		return false;
	}

	temp = ntohl(temp);

	// load itemp
	for(i = 0; i < temp; i++) {
		if(fread(name, STRSIZE, 1, sfil) != 1) {
			return false;
		}

		if(fread(&type, 2, 1, sfil) != 1) {
			return false;
		}

		type = ntohs(type);

		itemp.add(name, type);
	}

	return true;
}

// make sure ptemp == npcprops and itemp == itemprops
bool TemplateMatch(const proptemplate &ptemp, const proptemplate &itemp)
{
	int i, count;

	count = npcprops.getcount();

	if(count != ptemp.getcount())
		return false;

	for(i = 0; i < count; i++) {
		if(*(npcprops.prop(i)) != *(ptemp.prop(i)))
			return false;
	}

	count = itemprops.getcount();

	if(count != itemp.getcount())
		return false;

	for(i = 0; i < count; i++) {
		if(*(itemprops.prop(i)) != *(itemp.prop(i)))
			return false;
	}

	return true;
}

bool SaveGameTemplateInfo(FILE *sfil)
{
	unsigned short type;
	int i, temp;
	const property *prop;

	temp = htonl(npcprops.getcount());

	if(fwrite(&temp, 4, 1, sfil) != 1) {
		return false;
	}

	// save npcprops
	for(i = 0; i < npcprops.getcount(); i++) {
		prop = npcprops.prop(i);

		if(fwrite(prop -> name, STRSIZE, 1, sfil) != 1) {
			return false;
		}

		type = htons(prop -> type);

		if(fwrite(&type, 2, 1, sfil) != 1) {
			return false;
		}
	}

	temp = htonl(itemprops.getcount());

	if(fwrite(&temp, 4, 1, sfil) != 1) {
		return false;
	}

	// save itemprops
	for(i = 0; i < itemprops.getcount(); i++) {
		prop = itemprops.prop(i);

		if(fwrite(prop -> name, STRSIZE, 1, sfil) != 1) {
			return false;
		}

		type = htons(prop -> type);

		if(fwrite(&type, 2, 1, sfil) != 1) {
			return false;
		}
	}

	return true;
}

int typesize(int type)
{
	int btype = type & 0x000f;
	int size = (type >> 4) & 0x0fff;

	switch(btype) {
	case TYPE_INT:
		return 4;

	case TYPE_SHORT:
		return 2;

	case TYPE_BYTE:
		return 1;

	case TYPE_STRING:
		if(size > 0) {
			return size;
		}
		else {
			return 0;
		}

	case TYPE_ASHORT:
		if(size > 0) {
			return size * 2;
		}
		else {
			return 0;
		}

	case TYPE_IDLIST:
		if(size > 0)
			return size * 4;
		else
			return 0;

	default:
		return 0;
	}
}

const proptemplate *itemproptemplate(void)
{
	return &itemprops;
}

const proptemplate *npcproptemplate(void)
{
	return &npcprops;
}
/*
void propsetup(void)
{
	// item general
	itemprops.add("x", maketype(TYPE_SHORT));
	itemprops.add("y", maketype(TYPE_SHORT));
	itemprops.add("model", maketype(TYPE_SHORT));
	itemprops.add("color", maketype(TYPE_SHORT));
	itemprops.add("z", maketype(TYPE_BYTE));
	itemprops.add("name", maketype(TYPE_STRING, 30));
	itemprops.add("childlist", maketype(TYPE_IDLIST, 32));
	itemprops.add("parentid", maketype(TYPE_INT));
	itemprops.add("pooled", maketype(TYPE_BYTE));

	// item specific
	itemprops.add("equip", maketype(TYPE_BYTE));
	itemprops.add("flag", maketype(TYPE_BYTE));
	itemprops.add("extra", maketype(TYPE_SHORT));
	itemprops.add("flag2", maketype(TYPE_BYTE));
	itemprops.add("flag3", maketype(TYPE_BYTE));

	itemprops.add("container", maketype(TYPE_SHORT));

	// item general
	itemtemplate("x", itemprops.ix);
	itemtemplate("y", itemprops.iy);
	itemtemplate("model", itemprops.imodel);
	itemtemplate("color", itemprops.icolor);
	itemtemplate("z", itemprops.iz);
	itemtemplate("name", itemprops.iname);
	itemtemplate("childlist", itemprops.ichildlist);
	itemtemplate("parentid", itemprops.iparentid);
	itemtemplate("pooled", itemprops.ipooled);

	// item specific
	itemtemplate("equip", itemprops.iequip);
	itemtemplate("flag", itemprops.iflag);
	itemtemplate("extra", itemprops.iextra);
	itemtemplate("flag2", itemprops.iflag2);
	itemtemplate("flag3", itemprops.iflag3);

	// npc general
	npcprops.add("x", maketype(TYPE_SHORT));
	npcprops.add("y", maketype(TYPE_SHORT));
	npcprops.add("model", maketype(TYPE_SHORT));
	npcprops.add("color", maketype(TYPE_SHORT));
	npcprops.add("z", maketype(TYPE_BYTE));
	npcprops.add("name", maketype(TYPE_STRING, 30));
	npcprops.add("childlist", maketype(TYPE_IDLIST, 32));
	npcprops.add("parentid", maketype(TYPE_INT));
	npcprops.add("pooled", maketype(TYPE_BYTE));

	// npc specific
	npcprops.add("acct", maketype(TYPE_INT));
	npcprops.add("mymodel", maketype(TYPE_SHORT));
	npcprops.add("mycolor", maketype(TYPE_SHORT));
	npcprops.add("gmcolor", maketype(TYPE_SHORT));
	npcprops.add("skills", maketype(TYPE_ASHORT, 46));
	npcprops.add("gender", maketype(TYPE_BYTE));
	npcprops.add("facing", maketype(TYPE_BYTE));
	npcprops.add("show", maketype(TYPE_BYTE));
	npcprops.add("pw", maketype(TYPE_STRING, 30));
	npcprops.add("rep", maketype(TYPE_BYTE));
	npcprops.add("gold", maketype(TYPE_INT));
	npcprops.add("strength", maketype(TYPE_SHORT));
	npcprops.add("dexterity", maketype(TYPE_SHORT));
	npcprops.add("intelligence", maketype(TYPE_SHORT));
	npcprops.add("armor", maketype(TYPE_SHORT));
	npcprops.add("weight", maketype(TYPE_SHORT));
	npcprops.add("totalhits", maketype(TYPE_SHORT));
	npcprops.add("hits", maketype(TYPE_SHORT));
	npcprops.add("totalstamina", maketype(TYPE_SHORT));
	npcprops.add("stamina", maketype(TYPE_SHORT));
	npcprops.add("totalmana", maketype(TYPE_SHORT));
	npcprops.add("mana", maketype(TYPE_SHORT));

	npcprops.add("deathcry1", maketype(TYPE_SHORT));

	// npc general
	npctemplate("x", npcprops.ix);
	npctemplate("y", npcprops.iy);
	npctemplate("model", npcprops.imodel);
	npctemplate("color", npcprops.icolor);
	npctemplate("z", npcprops.iz);
	npctemplate("name", npcprops.iname);
	npctemplate("childlist", npcprops.ichildlist);
	npctemplate("parentid", npcprops.iparentid);
	npctemplate("pooled", npcprops.ipooled);

	// npc specific
	npctemplate("acct", npcprops.iacct);
	npctemplate("mymodel", npcprops.imymodel);
	npctemplate("mycolor", npcprops.imycolor);
	npctemplate("gmcolor", npcprops.igmcolor);
	npctemplate("skills", npcprops.iskills);
	npctemplate("gender", npcprops.igender);
	npctemplate("facing", npcprops.ifacing);
	npctemplate("show", npcprops.ishow);
	npctemplate("pw", npcprops.ipw);
	npctemplate("rep", npcprops.irep);
	npctemplate("gold", npcprops.igold);
	npctemplate("strength", npcprops.istrength);
	npctemplate("dexterity", npcprops.idexterity);
	npctemplate("intelligence", npcprops.iintelligence);
	npctemplate("armor", npcprops.iarmor);
	npctemplate("weight", npcprops.iweight);
	npctemplate("totalhits", npcprops.itotalhits);
	npctemplate("hits", npcprops.ihits);
	npctemplate("totalstamina", npcprops.itotalstamina);
	npctemplate("stamina", npcprops.istamina);
	npctemplate("totalmana", npcprops.itotalmana);
	npctemplate("mana", npcprops.imana);
}
*/
/*void propsetup(void)
{
	// item general
	AddItemProp("x", maketype(TYPE_SHORT));
	AddItemProp("y", maketype(TYPE_SHORT));
	AddItemProp("model", maketype(TYPE_SHORT));
	AddItemProp("color", maketype(TYPE_SHORT));
	AddItemProp("z", maketype(TYPE_BYTE));
	AddItemProp("name", maketype(TYPE_STRING, 30));
	AddItemProp("childlist", maketype(TYPE_IDLIST, 32));
	AddItemProp("parentid", maketype(TYPE_INT));
	AddItemProp("pooled", maketype(TYPE_BYTE));

	// item specific
	AddItemProp("equip", maketype(TYPE_BYTE));
	AddItemProp("flag", maketype(TYPE_BYTE));
	AddItemProp("extra", maketype(TYPE_SHORT));
	AddItemProp("flag2", maketype(TYPE_BYTE));
	AddItemProp("flag3", maketype(TYPE_BYTE));

	AddItemProp("container", maketype(TYPE_SHORT));

	// item general
	GetItemProp("x", itemprops.ix);
	GetItemProp("y", itemprops.iy);
	GetItemProp("model", itemprops.imodel);
	GetItemProp("color", itemprops.icolor);
	GetItemProp("z", itemprops.iz);
	GetItemProp("name", itemprops.iname);
	GetItemProp("childlist", itemprops.ichildlist);
	GetItemProp("parentid", itemprops.iparentid);
	GetItemProp("pooled", itemprops.ipooled);

	// item specific
	GetItemProp("equip", itemprops.iequip);
	GetItemProp("flag", itemprops.iflag);
	GetItemProp("extra", itemprops.iextra);
	GetItemProp("flag2", itemprops.iflag2);
	GetItemProp("flag3", itemprops.iflag3);

	// npc general
	AddNPCProp("x", maketype(TYPE_SHORT));
	AddNPCProp("y", maketype(TYPE_SHORT));
	AddNPCProp("model", maketype(TYPE_SHORT));
	AddNPCProp("color", maketype(TYPE_SHORT));
	AddNPCProp("z", maketype(TYPE_BYTE));
	AddNPCProp("name", maketype(TYPE_STRING, 30));
	AddNPCProp("childlist", maketype(TYPE_IDLIST, 32));
	AddNPCProp("parentid", maketype(TYPE_INT));
	AddNPCProp("pooled", maketype(TYPE_BYTE));

	// npc specific
	AddNPCProp("acct", maketype(TYPE_INT));
	AddNPCProp("mymodel", maketype(TYPE_SHORT));
	AddNPCProp("mycolor", maketype(TYPE_SHORT));
	AddNPCProp("gmcolor", maketype(TYPE_SHORT));
	AddNPCProp("skills", maketype(TYPE_ASHORT, 46));
	AddNPCProp("gender", maketype(TYPE_BYTE));
	AddNPCProp("facing", maketype(TYPE_BYTE));
	AddNPCProp("show", maketype(TYPE_BYTE));
	AddNPCProp("pw", maketype(TYPE_STRING, 30));
	AddNPCProp("rep", maketype(TYPE_BYTE));
	AddNPCProp("gold", maketype(TYPE_INT));
	AddNPCProp("strength", maketype(TYPE_SHORT));
	AddNPCProp("dexterity", maketype(TYPE_SHORT));
	AddNPCProp("intelligence", maketype(TYPE_SHORT));
	AddNPCProp("armor", maketype(TYPE_SHORT));
	AddNPCProp("weight", maketype(TYPE_SHORT));
	AddNPCProp("totalhits", maketype(TYPE_SHORT));
	AddNPCProp("hits", maketype(TYPE_SHORT));
	AddNPCProp("totalstamina", maketype(TYPE_SHORT));
	AddNPCProp("stamina", maketype(TYPE_SHORT));
	AddNPCProp("totalmana", maketype(TYPE_SHORT));
	AddNPCProp("mana", maketype(TYPE_SHORT));

	AddNPCProp("deathcry1", maketype(TYPE_SHORT));

	// npc general
	GetNPCProp("x", npcprops.ix);
	GetNPCProp("y", npcprops.iy);
	GetNPCProp("model", npcprops.imodel);
	GetNPCProp("color", npcprops.icolor);
	GetNPCProp("z", npcprops.iz);
	GetNPCProp("name", npcprops.iname);
	GetNPCProp("childlist", npcprops.ichildlist);
	GetNPCProp("parentid", npcprops.iparentid);
	GetNPCProp("pooled", npcprops.ipooled);

	// npc specific
	GetNPCProp("acct", npcprops.iacct);
	GetNPCProp("mymodel", npcprops.imymodel);
	GetNPCProp("mycolor", npcprops.imycolor);
	GetNPCProp("gmcolor", npcprops.igmcolor);
	GetNPCProp("skills", npcprops.iskills);
	GetNPCProp("gender", npcprops.igender);
	GetNPCProp("facing", npcprops.ifacing);
	GetNPCProp("show", npcprops.ishow);
	GetNPCProp("pw", npcprops.ipw);
	GetNPCProp("rep", npcprops.irep);
	GetNPCProp("gold", npcprops.igold);
	GetNPCProp("strength", npcprops.istrength);
	GetNPCProp("dexterity", npcprops.idexterity);
	GetNPCProp("intelligence", npcprops.iintelligence);
	GetNPCProp("armor", npcprops.iarmor);
	GetNPCProp("weight", npcprops.iweight);
	GetNPCProp("totalhits", npcprops.itotalhits);
	GetNPCProp("hits", npcprops.ihits);
	GetNPCProp("totalstamina", npcprops.itotalstamina);
	GetNPCProp("stamina", npcprops.istamina);
	GetNPCProp("totalmana", npcprops.itotalmana);
	GetNPCProp("mana", npcprops.imana);
}*/

FUSELIBFNDEF short propshortarrayget(int id, const propinfo &array, int index)
{
	dynamic *data = getdata(id);

	return data -> propshortarray(array, index);
}

// should be prepared to backup
FUSELIBFNDEF void propshortarrayset(int id, const propinfo &array, int index, short val)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id);

	data -> propshortarray(array, index, val, backup);
}

// should be prepared to backup
FUSELIBFNDEF void propshortarrayinit(int id, const propinfo &array)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id);

	data -> propshortarrayinit(array, backup);
}

FUSELIBFNDEF int propintget(int id, const propinfo &info)
{
	dynamic *data = getdata(id);

	return data -> propint(info);
}

// should be prepared to backup
FUSELIBFNDEF void propintset(int id, const propinfo &info, int val)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id);

	data -> propint(info, val, backup);
}

FUSELIBFNDEF short propshortget(int id, const propinfo &info)
{
	dynamic *data = getdata(id);

	return data -> propshort(info);
}

// should be prepared to backup
FUSELIBFNDEF void propshortset(int id, const propinfo &info, short val)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id);

	data -> propshort(info, val, backup);
}

FUSELIBFNDEF char propcharget(int id, const propinfo &info)
{
	dynamic *data = getdata(id);

	return data -> propchar(info);
}

// should be prepared to backup
FUSELIBFNDEF void propcharset(int id, const propinfo &info, char val)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id);

	data -> propchar(info, val, backup);
}

FUSELIBFNDEF int propboolget(int id, const propinfo &info)
{
	dynamic *data = getdata(id);

	return data -> propbool(info);
}

// should be prepared to backup
FUSELIBFNDEF void propboolset(int id, const propinfo &info, int val)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id), state;

	if(val == 0)
		state = false;
	else
		state  = true;

	data -> propbool(info, state, backup);
}

FUSELIBFNDEF const char *propstringget(int id, const propinfo &info)
{
	dynamic *data = getdata(id);

	return data -> propstring(info);
}

// should be prepared to backup
FUSELIBFNDEF void propstringset(int id, const propinfo &info, const char *val)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id);

	data -> propstring(info, val, backup);
}

// should be prepared to backup
FUSELIBFNDEF void proplistinit(int id, const propinfo &info)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id);

	data -> proplistinit(info, backup);
}

FUSELIBFNDEF int proplistcount(int id, const propinfo &info)
{
	dynamic *data = getdata(id);

	return data -> proplistcount(info);
}

FUSELIBFNDEF int proplistid(int id, const propinfo &info, int index)
{
	dynamic *data = getdata(id);

	return data -> proplistid(info, index);
}

// should be prepared to backup
FUSELIBFNDEF int proplistplus(int id, const propinfo &info, int iid)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id);

	return data -> proplistplus(info, iid, backup);
}

// should be prepared to backup
FUSELIBFNDEF int proplistminus(int id, const propinfo &info, int iid)
{
	dynamic *data = getdata(id);
	bool backup = getbackupstate(id);

	return data -> proplistminus(info, iid, backup);
}

FUSELIBFNDEF int proplistfull(int id, const propinfo &info)
{
	dynamic *data = getdata(id);

	return data -> proplistfull(info);
}

/*void x(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ix;
	else if(isitemid(id))
		info = itemprops.ix;

	propshortset(id, info, val);
}

short x(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ix;
	else if(isitemid(id))
		info = itemprops.ix;

	return propshortget(id, info);
}

void y(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iy;
	else if(isitemid(id))
		info = itemprops.iy;

	propshortset(id, info, val);
}

short y(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iy;
	else if(isitemid(id))
		info = itemprops.iy;

	return propshortget(id, info);
}

void model(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.imodel;
	else if(isitemid(id))
		info = itemprops.imodel;

	propshortset(id, info, val);
}

short model(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.imodel;
	else if(isitemid(id))
		info = itemprops.imodel;

	return propshortget(id, info);
}

void color(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.icolor;
	else if(isitemid(id))
		info = itemprops.icolor;

	propshortset(id, info, val);
}

short color(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.icolor;
	else if(isitemid(id))
		info = itemprops.icolor;

	return propshortget(id, info);
}

void z(int id, char val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iz;
	else if(isitemid(id))
		info = itemprops.iz;

	propcharset(id, info, val);
}

char z(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iz;
	else if(isitemid(id))
		info = itemprops.iz;

	return propcharget(id, info);
}

void name(int id, const char *val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iname;
	else if(isitemid(id))
		info = itemprops.iname;

	propstringset(id, info, val);
}

const char *name(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iname;
	else if(isitemid(id))
		info = itemprops.iname;

	return propstringget(id, info);
}

void childlistinit(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ichildlist;
	else if(isitemid(id))
		info = itemprops.ichildlist;

	proplistinit(id, info);
}

int childlistcount(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ichildlist;
	else if(isitemid(id))
		info = itemprops.ichildlist;

	return proplistcount(id, info);
}

int childlistid(int id, int index)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ichildlist;
	else if(isitemid(id))
		info = itemprops.ichildlist;

	return proplistid(id, info, index);
}

bool childlistplus(int id, int iid)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ichildlist;
	else if(isitemid(id))
		info = itemprops.ichildlist;

	return proplistplus(id, info, iid);
}

bool childlistminus(int id, int iid)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ichildlist;
	else if(isitemid(id))
		info = itemprops.ichildlist;

	return proplistminus(id, info, iid);
}

bool childlistfull(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ichildlist;
	else if(isitemid(id))
		info = itemprops.ichildlist;

	return proplistfull(id, info);
}

void parentid(int id, int val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iparentid;
	else if(isitemid(id))
		info = itemprops.iparentid;

	propintset(id, info, val);
}

int parentid(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iparentid;
	else if(isitemid(id))
		info = itemprops.iparentid;

	return propintget(id, info);
}

void pooled(int id, bool val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ipooled;
	else if(isitemid(id))
		info = itemprops.ipooled;

	propboolset(id, info, val);
}

bool pooled(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ipooled;
	else if(isitemid(id))
		info = itemprops.ipooled;

	return propboolget(id, info);
}

// item-specific prop access
void equip(int id, char val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iequip;
	else if(isitemid(id))
		info = itemprops.iequip;

	propcharset(id, info, val);
}

char equip(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iequip;
	else if(isitemid(id))
		info = itemprops.iequip;

	return propcharget(id, info);
}

void flag(int id, char val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iflag;
	else if(isitemid(id))
		info = itemprops.iflag;

	propcharset(id, info, val);
}

char flag(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iflag;
	else if(isitemid(id))
		info = itemprops.iflag;

	return propcharget(id, info);
}

void extra(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iextra;
	else if(isitemid(id))
		info = itemprops.iextra;

	propshortset(id, info, val);
}

short extra(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iextra;
	else if(isitemid(id))
		info = itemprops.iextra;

	return propshortget(id, info);
}

void flag2(int id, char val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iflag2;
	else if(isitemid(id))
		info = itemprops.iflag2;

	propcharset(id, info, val);
}

char flag2(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iflag2;
	else if(isitemid(id))
		info = itemprops.iflag2;

	return propcharget(id, info);
}

void flag3(int id, char val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iflag3;
	else if(isitemid(id))
		info = itemprops.iflag3;

	propcharset(id, info, val);
}

char flag3(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iflag3;
	else if(isitemid(id))
		info = itemprops.iflag3;

	return propcharget(id, info);
}

// npc-specific prop access
void acct(int id, int val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iacct;
	else if(isitemid(id))
		info = itemprops.iacct;

	propintset(id, info, val);
}

void mymodel(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.imymodel;
	else if(isitemid(id))
		info = itemprops.imymodel;

	propshortset(id, info, val);
}

short mymodel(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.imymodel;
	else if(isitemid(id))
		info = itemprops.imycolor;

	return propshortget(id, info);
}

void mycolor(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.imycolor;
	else if(isitemid(id))
		info = itemprops.imycolor;

	propshortset(id, info, val);
}

short mycolor(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.imycolor;
	else if(isitemid(id))
		info = itemprops.imycolor;

	return propshortget(id, info);
}

void gmcolor(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.igmcolor;
	else if(isitemid(id))
		info = itemprops.igmcolor;

	propshortset(id, info, val);
}

void gender(int id, char val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.igender;
	else if(isitemid(id))
		info = itemprops.igender;

	propcharset(id, info, val);
}

char gender(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.igender;
	else if(isitemid(id))
		info = itemprops.igender;

	return propcharget(id, info);
}

void facing(int id, char val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ifacing;
	else if(isitemid(id))
		info = itemprops.ifacing;

	propcharset(id, info, val);
}

char facing(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ifacing;
	else if(isitemid(id))
		info = itemprops.ifacing;

	return propcharget(id, info);
}

void show(int id, char val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ishow;
	else if(isitemid(id))
		info = itemprops.ishow;

	propcharset(id, info, val);
}

char show(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ishow;
	else if(isitemid(id))
		info = itemprops.ishow;

	return propcharget(id, info);
}

void pw(int id, const char *val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ipw;
	else if(isitemid(id))
		info = itemprops.ipw;

	propstringset(id, info, val);
}

void rep(int id, char val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.irep;
	else if(isitemid(id))
		info = itemprops.irep;

	propcharset(id, info, val);
}

char rep(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.irep;
	else if(isitemid(id))
		info = itemprops.irep;

	return propcharget(id, info);
}

void gold(int id, int val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.igold;
	else if(isitemid(id))
		info = itemprops.igold;

	propintset(id, info, val);
}

int gold(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.igold;
	else if(isitemid(id))
		info = itemprops.igold;

	return propintget(id, info);
}

void strength(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.istrength;
	else if(isitemid(id))
		info = itemprops.istrength;

	propshortset(id, info, val);
}

short strength(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.istrength;
	else if(isitemid(id))
		info = itemprops.istrength;

	return propshortget(id, info);
}

void dexterity(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.idexterity;
	else if(isitemid(id))
		info = itemprops.idexterity;

	propshortset(id, info, val);
}

short dexterity(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.idexterity;
	else if(isitemid(id))
		info = itemprops.idexterity;

	return propshortget(id, info);
}

void intelligence(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iintelligence;
	else if(isitemid(id))
		info = itemprops.iintelligence;

	propshortset(id, info, val);
}

short intelligence(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iintelligence;
	else if(isitemid(id))
		info = itemprops.iintelligence;

	return propshortget(id, info);
}

void armor(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iarmor;
	else if(isitemid(id))
		info = itemprops.iarmor;

	propshortset(id, info, val);
}

short armor(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iarmor;
	else if(isitemid(id))
		info = itemprops.iarmor;

	return propshortget(id, info);
}

void weight(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iweight;
	else if(isitemid(id))
		info = itemprops.iweight;

	propshortset(id, info, val);
}

short weight(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iweight;
	else if(isitemid(id))
		info = itemprops.iweight;

	return propshortget(id, info);
}

void totalhits(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.itotalhits;
	else if(isitemid(id))
		info = itemprops.itotalhits;

	propshortset(id, info, val);
}

short totalhits(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.itotalhits;
	else if(isitemid(id))
		info = itemprops.itotalhits;

	return propshortget(id, info);
}

void hits(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ihits;
	else if(isitemid(id))
		info = itemprops.ihits;

	propshortset(id, info, val);
}

short hits(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.ihits;
	else if(isitemid(id))
		info = itemprops.ihits;

	return propshortget(id, info);
}

void totalstamina(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.itotalstamina;
	else if(isitemid(id))
		info = itemprops.itotalstamina;

	propshortset(id, info, val);
}

short totalstamina(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.itotalstamina;
	else if(isitemid(id))
		info = itemprops.itotalstamina;

	return propshortget(id, info);
}

void stamina(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.istamina;
	else if(isitemid(id))
		info = itemprops.istamina;

	propshortset(id, info, val);
}

short stamina(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.istamina;
	else if(isitemid(id))
		info = itemprops.istamina;

	return propshortget(id, info);
}

void totalmana(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.itotalmana;
	else if(isitemid(id))
		info = itemprops.itotalmana;

	propshortset(id, info, val);
}

short totalmana(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.itotalmana;
	else if(isitemid(id))
		info = itemprops.itotalmana;

	return propshortget(id, info);
}

void mana(int id, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.imana;
	else if(isitemid(id))
		info = itemprops.imana;

	propshortset(id, info, val);
}

short mana(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.imana;
	else if(isitemid(id))
		info = itemprops.imana;

	return propshortget(id, info);
}

short skills(int id, int iid)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iskills;
	else if(isitemid(id))
		info = itemprops.iskills;

	return propshortarrayget(id, info, iid - 1);
}

void skills(int id, int iid, short val)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iskills;
	else if(isitemid(id))
		info = itemprops.iskills;

	propshortarrayset(id, info, iid - 1, val);
}

void skillsinit(int id)
{
	propinfo info;

	if(isnpcid(id))
		info = npcprops.iskills;
	else if(isitemid(id))
		info = itemprops.iskills;

	propshortarrayinit(id, info);
}*/

// we should try not to write more than size bytes to buf
FUSELIBFNDEF int GetProperty(int id, const char *prop, char *buf, int size, int index)
{
	propinfo info;
	int btype, typesize, ret = 0;
	unsigned temp;
	//dynamic *data;

	buf[0] = 0;

	if(isitemid(id)) {
		GetItemProp(prop, info);
	}
	else if(isnpcid(id)) {
		GetNPCProp(prop, info);
	}

	//data = getdata(id);

	btype = info.type & 0x000f;
	typesize = (info.type >> 4) & 0x0fff;

	//if(data != NULL) {
		switch(btype) {
			case TYPE_INT:
				temp = (unsigned)propintget(id, info);
				sprintf(buf, "%x", temp);
				break;

			case TYPE_SHORT:
				temp = (unsigned short)propshortget(id, info);
				sprintf(buf, "%x", temp);
				break;

			case TYPE_BYTE:
				temp = (unsigned char)propcharget(id, info);
				sprintf(buf, "%x", temp);
				break;

			case TYPE_STRING:
				strcpy(buf, propstringget(id, info));
				break;

			case TYPE_ASHORT:
				if(index >= 0 && index < typesize) {
					temp = (unsigned short)propshortarrayget(id, info, index);
					sprintf(buf, "%x", temp);
				}
				else {
					// for now, don't display the whole array
					ret = 3;
				}

				break;

			case TYPE_IDLIST:
				if(index >= 0 && index < typesize) {
					temp = proplistid(id, info, index);
					sprintf(buf, "%x", temp);
				}
				else {
					// for now, don't display the whole list
					ret = 4;
				}

				break;

			default:
				ret = 2; // prop doesn't exist
		}
	/*}
	else {
		ret = 1; // item doesn't exist
	}*/

	return ret;
}

// buf is no larger than size bytes
FUSELIBFNDEF int SetProperty(int id, const char *prop, char *buf, int size, int index)
{
	propinfo info;
	int btype, typesize, ret = 0;
	unsigned temp;
	dynamic *data;

	if(isitemid(id)) {
		GetItemProp(prop, info);
	}
	else if(isnpcid(id)) {
		GetNPCProp(prop, info);
	}

	data = getdata(id);

	btype = info.type & 0x000f;
	typesize = (info.type >> 4) & 0x0fff;

	if(data != NULL) {
		switch(btype) {
			case TYPE_INT:
				sscanf(buf, "%x", &temp);
				propintset(id, info, temp);
				break;

			case TYPE_SHORT:
				sscanf(buf, "%x", &temp);
				propshortset(id, info, temp);
				break;

			case TYPE_BYTE:
				sscanf(buf, "%x", &temp);
				propcharset(id, info, temp);
				break;

			case TYPE_STRING:
				propstringset(id, info, buf);
				break;

			case TYPE_ASHORT:
				if(index >= 0 && index < typesize) {
					sscanf(buf, "%x", &temp);
					propshortarrayset(id, info, index, temp);
				}
				else {
					// for now, don't set the whole array
					ret = 3;
				}

				break;

			case TYPE_IDLIST:
				// manually editing lists like this may produce unpredictable
				// results (as would be the case with childlist)
				// use with caution
				if(index > 0) {
					// add id
					sscanf(buf, "%x", &temp);
					proplistplus(id, info, temp);
				}
				else if(index < 0) {
					// for now, don't set the whole list
					ret = 4;
				}
				else {
					// remove id
					sscanf(buf, "%x", &temp);
					proplistminus(id, info, temp);
				}

				break;

			default:
				ret = 2; // prop doesn't exist
		}
	}
	else {
		ret = 1; // item doesn't exist
	}

	return ret;
}

FUSELIBFNDEF void GetItemProp(const char *name, propinfo &info)
{
	itemprops.info(name, info);
}

FUSELIBFNDEF void GetNPCProp(const char *name, propinfo &info)
{
	npcprops.info(name, info);
}

FUSELIBFNDEF void AddItemProp(const char *prop, int type)
{
	itemprops.add(prop, type);
}

FUSELIBFNDEF void AddNPCProp(const char *prop, int type)
{
	npcprops.add(prop, type);
}

int itempropsize(void)
{
	return itemprops.size();
}

int npcpropsize(void)
{
	return npcprops.size();
}

FUSELIBFNDEF void CopyBuf(char *dest, int dsize, const char *src)
{
	int i, ssize;

	ssize = strlen(src);

	if(ssize + 1 > dsize)
		ssize = dsize - 1;

	for(i = 0; i < ssize; i++)
		dest[i] = src[i];

	for(i = ssize; i < dsize; i++)
		dest[i] = 0;

	// fail-safe
	dest[dsize - 1] = 0;
}

//decode data received from client
void ClientDecode(void *ptr, int size, int &ecthis, const char *encrypt)
{
	int i, temp;
	char *buf = (char *)ptr;

	for(i = 0; i < size; i++) {
		if(ecthis < 0x40)
			temp = encrypt[ecthis & 0x3f] ^ buf[i];
		else
			temp = encrypt[ecthis & 0x3f] ^ buf[i] ^ 0xff;

		buf[i] = temp;

		ecthis = (ecthis + 1) & 0x7f;
	}
}

int SendBytes(SOCKET sock, void *buf, int count)
{
	int size = 0, val;

	while(size < count) {
		if((val = send(sock, (char *)buf, count - size, 0)) == SOCKET_ERROR) {
			return val;
		}
		else if(val == 0) {
			return val;
		}
		else {
			size += val;
		}
	}

	return size;
}

int ReceiveBytes(SOCKET sock, void *buf, int count)
{
	int size = 0, val;

	while(size < count) {
		if((val = recv(sock, (char *)buf, count - size, 0)) == SOCKET_ERROR) {
			return val;
		}
		else if(val == 0) {
			return val;
		}
		else {
			size += val;
		}
	}

	return size;
}
/*
//print out the data sent by the server
void serveroutput(void *ptr, int size)
{
	int lcount, i, j, count, done, c;
	char *buf = (char *)ptr;

	lcount = (size + 15) / 16;
	printf("Outgoing message(s) size: %d; Line count: %d\n", size, lcount);
	done = 0;

	for(i = 0; i < lcount; i++) {
		count = (size - done) < 16 ? (size - done) : 16;

		for(j = 0; j < count; j++) {
			if(buf[done + j] < 0x20) {
				printf("%c", '.');
			}
			else {
				printf("%c", buf[done + j]);
			}
		}

		for(j = 0; j < 16 - count; j++) {
			printf(" ");
		}

		for(j = 0; j < count; j++) {
			printf(" ");

			c = (buf[done + j] >> 4) & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');

			c = buf[done + j] & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');
		}

		done += count;
		printf("\n");
	}
}

//print out the data received from the client
void clientoutput(void *ptr, int size)
{
	int lcount, i, j, count, done, c;
	char *buf = (char *)ptr;

	lcount = (size + 15) / 16;
	printf("Incoming client message(s) size: %d; Line count: %d\n", size, lcount);
	done = 0;

	for(i = 0; i < lcount; i++) {
		count = (size - done) < 16 ? (size - done) : 16;

		for(j = 0; j < count; j++) {
			if(buf[done + j] < 0x20) {
				printf("%c", '.');
			}
			else {
				printf("%c", buf[done + j]);
			}
		}

		for(j = 0; j < 16 - count; j++) {
			printf(" ");
		}

		for(j = 0; j < count; j++) {
			printf(" ");

			c = (buf[done + j] >> 4) & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');

			c = buf[done + j] & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');
		}

		done += count;
		printf("\n");
	}
}

//print out the data received from the client
void unknownoutput(void *ptr, int size)
{
	int lcount, i, j, count, done, c;
	char *buf = (char *)ptr;

	lcount = (size + 15) / 16;
	printf("Unknown client message(s) size: %d; Line count: %d\n", size, lcount);
	done = 0;

	for(i = 0; i < lcount; i++) {
		count = (size - done) < 16 ? (size - done) : 16;

		for(j = 0; j < count; j++) {
			if(buf[done + j] < 0x20) {
				printf("%c", '.');
			}
			else {
				printf("%c", buf[done + j]);
			}
		}

		for(j = 0; j < 16 - count; j++) {
			printf(" ");
		}

		for(j = 0; j < count; j++) {
			printf(" ");

			c = (buf[done + j] >> 4) & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');

			c = buf[done + j] & 0x0000000F;
			if(c < 0xa)
				printf("%c", c + '0');
			else
				printf("%c", c - 0xa + 'a');
		}

		done += count;
		printf("\n");
	}
}
*/
FUSELIBFNDEF void BufferByte(char *buf, int offset, char data)
{
	buf[offset] = data;
}

FUSELIBFNDEF void BufferShort(char *buf, int offset, short data)
{
	*((short *)(buf + offset)) = htons(data);
}

FUSELIBFNDEF void BufferInt(char *buf, int offset, int data)
{
	*((int *)(buf + offset)) = htonl(data);
}

FUSELIBFNDEF void BufferString(char *buf, int offset, const char *data, int size)
{
	int i;

	for(i = 0; i < size; i++) {
		buf[offset + i] = data[i];
	}
}

FUSELIBFNDEF char UnBufferByte(const char *buf, int offset)
{
	return buf[offset];
}

FUSELIBFNDEF short UnBufferShort(const char *buf, int offset)
{
	return ntohs(*((short *)(buf + offset)));
}

FUSELIBFNDEF int UnBufferInt(const char *buf, int offset)
{
	return ntohl(*((int *)(buf + offset)));
}

FUSELIBFNDEF void UnBufferString(const char *buf, int offset, char *data, int size)
{
	int i;

	for(i = 0; i < size; i++) {
		data[i] = buf[offset + i];
	}
}
