// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

class idlist {
	int *ids;
	int count;

public:
	idlist(void)
	{
		ids = NULL;
		count = 0;
	}

	~idlist(void)
	{
		clear();
	}

	void clear(void)
	{
		if(ids != NULL) {
			delete []ids;
			ids = NULL;
		}

		count = 0;
	}

	const idlist &operator+=(int id)
	{
		int i;
		int *ids2;

		for(i = 0; i < count; i++) {
			if(ids[i] == id)
				break;
		}

		if(i == count) {
			ids2 = new int[count + 1];

			for(i = 0; i < count; i++)
				ids2[i] = ids[i];

			ids2[i] = id;

			delete []ids;
			ids = ids2;
			count += 1;
		}

		return *this;
	}

	const idlist &operator-=(int id)
	{
		int i, j;
		int *ids2;

		for(i = 0; i < count; i++) {
			if(ids[i] == id)
				break;
		}

		if(i < count) {
			ids2 = new int[count - 1];

			for(j = 0; j < i; j++)
				ids2[j] = ids[j];

			for(j = i + 1; j < count; j++)
				ids2[j - 1] = ids[j];

			delete []ids;
			ids = ids2;
			count -= 1;
		}

		return *this;
	}

	const int *getids(void) const
	{
		return ids;
	}

	int getcount(void) const
	{
		return count;
	}
} ;

class staticidlist {
	int *ids, *bids;
	int count, max;

public:
	staticidlist(void)
	{
		bids = ids = NULL;
		count = max = 0;
	}

	void setids(int *idptr, int size, int *bidptr)
	{
		int i;

		ids = idptr;
		bids = bidptr;
		max = size / sizeof(int);

		count = 0;

		for(i = 0; i < max; i++) {
			if(ntohl(ids[i]) == -1)
				break;
			else
				count += 1;
		}
	}

	void init(void)
	{
		int i;

		count = 0;

		if(bids == NULL) {
			for(i = 0; i < max; i++) {
				ids[i] = htonl(-1);
			}
		}
		else {
			for(i = 0; i < max; i++) {
				bids[i] = ids[i] = htonl(-1);
			}
		}
	}

	bool operator+=(int id)
	{
		int i;

		if(count < max) {
			for(i = 0; i < count; i++) {
				if((int)ntohl(ids[i]) == id)
					break;
			}

			if(i == count) {
				if(bids == NULL)
					ids[count] = htonl(id);
				else
					bids[count] = ids[count] = htonl(id);
				count += 1;
				return true;
			}
			else {
				return false;
			}
		}
		else {
			return false;
		}
	}

	bool operator-=(int id)
	{
		int i;

		for(i = 0; i < count; i++) {
			if((int)ntohl(ids[i]) == id)
				break;
		}

		if(i < count) {
			count -= 1;

			if(count >= 0) {
				if(bids == NULL) {
					ids[i] = ids[count];
					ids[count] = htonl(-1);
				}
				else {
					bids[i] = ids[i] = ids[count];
					bids[count] = ids[count] = htonl(-1);
				}
			}

			return true;
		}
		else {
			return false;
		}
	}

	int id(int index) const
	{
		if(index >= 0 && index < count)
			return ntohl(ids[index]);
		else
			return -1;
	}

	bool full(void) const
	{
		return count < max;
	}

	int getcount(void) const
	{
		return count;
	}
} ;

void ClientDecode(void *, int, int &, const char *);
int SendBytes(SOCKET, void *, int);
int ReceiveBytes(SOCKET, void *, int);
/*void serveroutput(void *, int);
void clientoutput(void *, int);
void unknownoutput(void *, int);*/

/*#define CREATESKILLS 3
struct createdata {
	unsigned short x, y, model, color;
	unsigned char equip;
	char z;
	char name[STRSIZE];
	int acct;
	unsigned short mymodel, mycolor;
	unsigned char skills[CREATESKILLS * 2];
	unsigned char gender;
	char pw[STRSIZE];
	unsigned char flag;
	unsigned short extra;
	unsigned short strength, dexterity, intelligence;
	unsigned short totalhits, totalstamina, totalmana, hits, stamina, mana;
	unsigned short hairtype, haircolor, beardtype, beardcolor;

	void reset(void)
	{
		x = y = model = color = 0;
		equip = 0;
		z = 0;
		memset(name, 0, STRSIZE);
		acct = -1;
		mymodel = mycolor = 0;
		memset(skills, 0, CREATESKILLS * 2 * sizeof(unsigned char));
		gender = 0;
		memset(pw, 0, STRSIZE);
		flag = 0;
		extra = 0;
		strength = dexterity = intelligence = 0;
		totalhits = totalstamina = totalmana = hits = stamina = mana = 0;
		hairtype = haircolor = beardtype = beardcolor = 0;
	}
} ;*/

/*void itemtemplate(const char *, propinfo &);
void npctemplate(const char *, propinfo &);*/

//short maketype(int, int = 0);
int typesize(int);

struct property {
	char name[STRSIZE];
	int offset, type, size;

	property(const char *prop, int off, int typ)
	{
		CopyBuf(name, STRSIZE, prop);
		offset = off;
		type = typ;
		size = typesize(type);
	}

	bool operator==(const property &prop) const
	{
		bool val = false;

		if(stricmp(name, prop.name) == 0) {
			if(offset == prop.offset && type == prop.type && size == prop.size)
				val = true;
		}

		return val;
	}

	bool operator!=(const property &prop) const
	{
		return !(*this == prop);
	}
} ;

class proptemplate {
	property **props;
	int count, total;

public:
	// well-known props

	/*// general props
	propinfo ix, iy, imodel, icolor, iz, iname, ichildlist, iparentid, ipooled;

	// item-specific
	propinfo iequip, iflag, iextra, iflag2, iflag3;

	// npc-specific
	propinfo iacct, imymodel, imycolor, igmcolor, iskills, igender, ifacing, ishow, ipw, irep, igold;
	propinfo istrength, idexterity, iintelligence, iarmor, iweight;
	propinfo itotalhits, ihits, itotalstamina, istamina, itotalmana, imana;*/

	proptemplate(void)
	{
		props = NULL;
		count = 0;
		total = 0;
	}

	void add(const char *prop, int type)
	{
		property **iprops;
		int i;

		iprops = new property *[count + 1];

		for(i = 0; i < count; i++) {
			iprops[i] = props[i];
		}

		if(props != NULL)
			delete []props;

		props = iprops;

		iprops[count] = new property(prop, total, type);
		count += 1;
		total += typesize(type);
	}

	int size(void) const
	{
		return total;
	}

	void info(const char *name, propinfo &prop) const
	{
		int i;

		for(i = 0; i < count; i++) {
			if(stricmp(name, props[i] -> name) == 0)
				break;
		}

		if(i < count) {
			prop.offset = props[i] -> offset;
			prop.type = props[i] -> type;
			prop.size = props[i] -> size;
		}
		else {
			prop.offset = 0;
			prop.type = 0;
			prop.size = 0;
		}
	}

	int getcount(void) const
	{
		return count;
	}

	const property *prop(int i) const
	{
		if(i >= 0 && i < count)
			return props[i];
		else
			return NULL;
	}
} ;

struct backupdata {
	bool use;
	int size;
	char *data;
} ;

// no destructor; do not attempt to delete props
//
// for now, for access to props that are not well-known,
// we must check id type, explicitly look-up prop in
// appropriate template, and pass propinfo -- we can also
// make access to well-known props operate in a similar
// manner so we won't need to store ix, iy, etc in each object
//
// anything that is saved must also be backed up (props and use)
class dynamic {
	char *props, *bprops;
	bool use, buse;
	int connectionid; // for pcs
	CommandQueue commands; // commands
	bool dirty;

protected:
	// properties
	char *getprops(void)
	{
		return props;
	}

	// well-known props
	const proptemplate *pt;

public:
	dynamic(void)
	{
		bprops = props = NULL;
		buse = use = false;
		connectionid = -1;
		//pt = NULL;
		// dirty is false, because whenever new space is allocated,
		// the data is invalid and is not in use
		dirty = false;
	}

	/*// general prop access
	short x(void)
	{
		return propshort(pt -> ix);
	}

	short y(void)
	{
		return propshort(pt -> iy);
	}

	short model(void)
	{
		return propshort(pt -> imodel);
	}

	short color(void)
	{
		return propshort(pt -> icolor);
	}

	char z(void)
	{
		return propchar(pt -> iz);
	}

	const char *name(void)
	{
		return propstring(pt -> iname);
	}

	int childlistcount(void) const
	{
		return proplistcount(pt -> ichildlist);
	}

	int childlistid(int index) const
	{
		return proplistid(pt -> ichildlist, index);
	}

	int parentid(void)
	{
		return propint(pt -> iparentid);
	}

	bool pooled(void)
	{
		return propbool(pt -> ipooled);
	}

	// item-specific prop access
	char equip(void)
	{
		return propchar(pt -> iequip);
	}

	char flag(void)
	{
		return propchar(pt -> iflag);
	}

	short extra(void)
	{
		return propshort(pt -> iextra);
	}

	char flag2(void)
	{
		return propchar(pt -> iflag2);
	}

	char flag3(void)
	{
		return propchar(pt -> iflag3);
	}

	// npc-specific prop access
	int acct(void)
	{
		return propint(pt -> iacct);
	}

	short mymodel(void)
	{
		return propshort(pt -> imymodel);
	}

	short mycolor(void)
	{
		return propshort(pt -> imycolor);
	}

	short gmcolor(void)
	{
		return propshort(pt -> igmcolor);
	}

	char gender(void)
	{
		return propchar(pt -> igender);
	}

	char facing(void)
	{
		return propchar(pt -> ifacing);
	}

	char show(void)
	{
		return propchar(pt -> ishow);
	}

	const char *pw(void)
	{
		return propstring(pt -> ipw);
	}

	char rep(void)
	{
		return propchar(pt -> irep);
	}

	int gold(void)
	{
		return propint(pt -> igold);
	}

	short strength(void)
	{
		return propshort(pt -> istrength);
	}

	short dexterity(void)
	{
		return propshort(pt -> idexterity);
	}

	short intelligence(void)
	{
		return propshort(pt -> iintelligence);
	}

	short armor(void)
	{
		return propshort(pt -> iarmor);
	}

	short weight(void)
	{
		return propshort(pt -> iweight);
	}

	short totalhits(void)
	{
		return propshort(pt -> itotalhits);
	}

	short hits(void)
	{
		return propshort(pt -> ihits);
	}

	short totalstamina(void)
	{
		return propshort(pt -> itotalstamina);
	}

	short stamina(void)
	{
		return propshort(pt -> istamina);
	}

	short totalmana(void)
	{
		return propshort(pt -> itotalmana);
	}

	short mana(void)
	{
		return propshort(pt -> imana);
	}

	short skills(int id)
	{
		return propshortarray(pt -> iskills, id - 1);
	}*/

	// sign extend everything
	short propshortarray(const propinfo &array, int index)
	{
		propinfo info;

		info = array;

		if(index >= 0 && index < (array.size >> 1)) {
			info.offset += index << 1;
			info.size = 2; // should probably set to sizeof short prop type

			return propshort(info);
		}
		else {
			return 0;
		}
	}

	void propshortarray(const propinfo &array, int index, short val, bool backup)
	{
		propinfo info;

		info = array;

		if(index >= 0 && index < (array.size >> 1)) {
			info.offset += index << 1;
			info.size = 2; // should probably set to sizeof short prop type

			propshort(info, val, backup);
		}
	}

	void propshortarrayinit(const propinfo &array, bool backup)
	{
		memset(props + array.offset, 0, array.size);

		if(backup) {
			memset(bprops + array.offset, 0, array.size);
		}
		else {
			dirty = true;
		}
	}

	int propint(const propinfo &info)
	{
		return UnBufferInt(props, info.offset);
	}

	void propint(const propinfo &info, int val, bool backup)
	{
		BufferInt(props, info.offset, val);

		if(backup) {
			BufferInt(bprops, info.offset, val);
		}
		else {
			dirty = true;
		}
	}

	short propshort(const propinfo &info)
	{
		return UnBufferShort(props, info.offset);
	}

	void propshort(const propinfo &info, short val, bool backup)
	{
		BufferShort(props, info.offset, val);

		if(backup) {
			BufferShort(bprops, info.offset, val);
		}
		else {
			dirty = true;
		}
	}

	char propchar(const propinfo &info)
	{
		return UnBufferByte(props, info.offset);
	}

	void propchar(const propinfo &info, char val, bool backup)
	{
		BufferByte(props, info.offset, val);

		if(backup) {
			BufferByte(bprops, info.offset, val);
		}
		else {
			dirty = true;
		}
	}

	bool propbool(const propinfo &info)
	{
		if(UnBufferByte(props, info.offset) == 0)
			return false;
		else
			return true;
	}

	void propbool(const propinfo &info, bool val, bool backup)
	{
		char byte;

		if(val) {
			byte = 1;
		}
		else {
			byte = 0;
		}

		BufferByte(props, info.offset, byte);

		if(backup) {
			BufferByte(bprops, info.offset, byte);
		}
		else {
			dirty = true;
		}
	}

	const char *propstring(const propinfo &info)
	{
		return props + info.offset;
	}

	void propstring(const propinfo &info, const char *val, bool backup)
	{
		CopyBuf(props + info.offset, info.size, val);

		if(backup) {
			CopyBuf(bprops + info.offset, info.size, val);
		}
		else {
			dirty = true;
		}
	}

	void proplistinit(const propinfo &info, bool backup)
	{
		staticidlist idlist;
		int *back;

		if(backup) {
			back = (int *)(bprops + info.offset);
		}
		else {
			back = NULL;
			dirty = true;
		}

		idlist.setids((int *)(props + info.offset), info.size, back);
		idlist.init();
	}

	int proplistcount(const propinfo &info) const
	{
		staticidlist idlist;

		idlist.setids((int *)(props + info.offset), info.size, NULL);
		return idlist.getcount();
	}

	int proplistid(const propinfo &info, int index) const
	{
		staticidlist idlist;

		idlist.setids((int *)(props + info.offset), info.size, NULL);
		return idlist.id(index);
	}

	bool proplistplus(const propinfo &info, int id, bool backup)
	{
		staticidlist idlist;
		int *back;
		bool result;

		if(backup) {
			back = (int *)(bprops + info.offset);
		}
		else {
			back = NULL;
		}

		idlist.setids((int *)(props + info.offset), info.size, back);
		result = (idlist += id);

		if(!backup && result)
			dirty = true;

		return result;
	}

	bool proplistminus(const propinfo &info, int id, bool backup)
	{
		staticidlist idlist;
		int *back;
		bool result;

		if(backup) {
			back = (int *)(bprops + info.offset);
		}
		else {
			back = NULL;
		}

		idlist.setids((int *)(props + info.offset), info.size, back);
		result = (idlist -= id);

		if(!backup && result)
			dirty = true;

		return result;
	}

	bool proplistfull(const propinfo &info) const
	{
		staticidlist idlist;

		idlist.setids((int *)(props + info.offset), info.size, NULL);
		return idlist.full();
	}

	void setprops(char *iprops, char *ibprops)
	{
		props = iprops;
		bprops = ibprops;
	}

	void used(bool flag, bool backup)
	{
		use = flag;

		if(backup) {
			buse = flag;
		}
		else {
			dirty = true;
		}
	}

	bool used(void)
	{
		return use;
	}

	int connection(void)
	{
		return connectionid;
	}

	void connection(int newconn)
	{
		connectionid = newconn;
	}

	void addcommand(const CommandItem *cmd)
	{
		commands.AddCommand(cmd);
	}

	void clearcommands(void)
	{
		commands.ClearCommands();
	}

	void getcommand(CommandItem *cmd)
	{
		commands.GetCommand(cmd);
	}

	bool commandavailable(void) const
	{
		return commands.HasCommand();
	}

	void commandcycle(void)
	{
		commands.Cycle();
	}

	// ack!  must only be called at the beginning to load data
	char *propptr(void)
	{
		return props;
	}

	void getbackupdata(backupdata &bdata)
	{
		bdata.use = buse;
		bdata.size = pt -> size();
		bdata.data = bprops;
	}

	// basically, if the use flag is clear, all we
	// need to do is copy the use flag (making sure
	// it is also clear in the backup) it doesn't
	// matter if the prop data has changed (even
	// if dirty is set) since the object has been
	// destroyed. if the use flag is set, then still
	// copy the use flag (just for peace of mind)
	// but only backup the prop data if dirty is set
	//
	// although this seems strange; it makes more sense
	// either not to use the precaution (copy nothing
	// if dirty is not set, everything if dirty is set)
	// or to not use dirty and to simply copy everything
	// after a successful backup; for now though, maybe
	// try dirty w/o the precaution
	void refreshbackupdata(void)
	{
		if(dirty) {
			buse = use;
			memcpy(bprops, props, pt -> size());
			dirty = false;
		}
	}
} ;

// general prop access
// we only need functions to set
// the properties.  accessing the
// properties doesn't need to worry
// about data backup.  also, once
// the script is in place, we won't
// need functions to set the props
/*void x(int id, short val);
short x(int);
void y(int id, short val);
short y(int);
void model(int id, short val);
short model(int);
void color(int id, short val);
short color(int);
void z(int id, char val);
char z(int);
void name(int id, const char *val);
const char *name(int);
void childlistinit(int id);
bool childlistplus(int id, int iid);
bool childlistminus(int id, int iid);
int childlistcount(int);
int childlistid(int, int);
bool childlistfull(int);
void parentid(int id, int val);
int parentid(int id);
void pooled(int id, bool val);
bool pooled(int);
void equip(int id, char val);
char equip(int);
void flag(int id, char val);
char flag(int id);
void extra(int id, short val);
short extra(int);
void flag2(int id, char val);
char flag2(int id);
void flag3(int id, char val);
char flag3(int id);
void acct(int id, int val);
void mymodel(int id, short val);
short mymodel(int);
void mycolor(int id, short val);
short mycolor(int);
void gmcolor(int id, short val);
void gender(int id, char val);
char gender(int);
void facing(int id, char val);
char facing(int);
void show(int id, char val);
char show(int);
void pw(int id, const char *val);
void rep(int id, char val);
char rep(int);
void gold(int id, int val);
int gold(int id);
void strength(int id, short val);
short strength(int id);
void dexterity(int id, short val);
short dexterity(int id);
void intelligence(int id, short val);
short intelligence(int id);
void armor(int id, short val);
short armor(int id);
void weight(int id, short val);
short weight(int id);
void totalhits(int id, short val);
short totalhits(int id);
void hits(int id, short val);
short hits(int id);
void totalstamina(int id, short val);
short totalstamina(int id);
void stamina(int id, short val);
short stamina(int id);
void totalmana(int id, short val);
short totalmana(int id);
void mana(int id, short val);
short mana(int id);
void skills(int id, int iid, short val);
short skills(int id, int iid);
void skillsinit(int id);*/

const proptemplate *itemproptemplate(void);
const proptemplate *npcproptemplate(void);

//void propsetup(void);
int itempropsize(void);
int npcpropsize(void);

bool LoadGameTemplateInfo(FILE *, proptemplate &, proptemplate &);
bool TemplateMatch(const proptemplate &, const proptemplate &);
bool SaveGameTemplateInfo(FILE *);
